%if style == newcode
\begin{code}
module BasicRewrites(rewrite) where
import VirtualMachine
\end{code}
%endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Basic rewrites rules} \label{chapter:basic-rules}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%{
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Expression statement and basic backward steps
\section{Blocks}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Blocks are sequences of statements. Python programs are represented by blocks. Compound statements, such as the while statement and module definitions, contain blocks. 

%In other language's semantics, blocks of statements usually delimit a scope for variables. In these semantics however, blocks do not influence variable scoping, although some compound statements, do have scoping rules for the blocks they contain. For example the body of a function is a scoped block (see Section \ref{section:function-execution}).
%CPython's documentation is different in this respect.


\rulesection{Executing a non-empty block statement}
\label{fw-rewrite-ConsBlock}
A block is executed statement by statement: the first statement |s| of the block |ConsBlock s b| is executed, while the remaining statements in block |b| are put onto the stack for later execution.

\begin{code}
rewrite  (State  heap  envs  ( stack                  )  ( FwBlock  (ConsBlock s b) )) =
         (state  heap  envs  ( stack :|: BlockFrame b )  ( FwStmt   s               ))
\end{code}

\rulesection{Executing an empty block}
\label{fw-rewrite-EmptyBlock}
An empty block is simply popped from the stack, returning the value |None|.

\begin{code}
rewrite  (State  heap  envs  stack  ( FwBlock  EmptyBlock )) =
         (state  heap  envs  stack  ( BwResult a_None     ))
\end{code}

Rather than creating a new object on the heap, the rule returns a predefined address pointing to the built-in value |None|.


\rulesection{Continuing execution of a block}
\label{bw-rewrite-BlockFrame}
When the execution the previous statement has completed with result |a|, the block |b| on the stack is executed if it is not empty. If |b| is empty, the block frame is popped and we return |a|. Thus the result of a block's final statement is the result of the complete block.

\begin{code}
rewrite           (State  heap  envs  ( stack :|: BlockFrame b )  ( BwResult  a )) = case b of
  EmptyBlock  ->  (state  heap  envs  ( stack                  )  ( BwResult  a ))
  otherwise   ->  (state  heap  envs  ( stack                  )  ( FwBlock   b ))
\end{code}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Pass and expression statements}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\rulesection{Pass statements}
\label{fw-rewrite-PassStmt}
The pass statement affects neither the heap, the environment nor the stack and returns |None|.

\begin{code}
rewrite  (State  heap  envs  stack  ( FwStmt PassStmt  )) =
         (state  heap  envs  stack  ( BwResult  a_None ))
\end{code}


\rulesection{Expression statements}
\label{fw-rewrite-ExprStmt}
Expressions also serve as statements. Syntactically these expression statements are indistinguishable from normal expressions, therefore they are annotated with a suffix |;|

\begin{code}
rewrite  (State  heap  envs  stack  ( FwStmt  (ExprStmt e) )) =
         (state  heap  envs  stack  ( FwExpr  e            ))
\end{code}

%if hide
\rulesection{Instruction frames}

%format instr = "\iota"
\begin{code}
rewrite  (State  heap  envs  ( stack :|: InstrFrame instr )  ( BwResult a )) =
         (state  heap  envs  ( stack                      )  ( instr      ))
\end{code}


\rulesection{Address expressions}
\begin{code}
rewrite  (State  heap  envs  stack  ( FwExpr (AddrExpr a) )) =
         (state  heap  envs  stack  ( BwResult (Addr a)   ))
\end{code}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%endif

%if style==newcode
\begin{code}
rewrite _ = skipState
\end{code}
%endif
%}
