%if style == newcode
\begin{code}
module ControlStructures(rewrite) where
import VirtualMachine
import ObjectTheory
\end{code}
%endif
%{

%format b_if
%format b_else
%format concatBlocks x y = x";~"y
%format concatBlocksAround x y z = x";~"y";~"z
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Control structures} \label{chapter:control-structures}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Python includes three basic control structures: if, while and for statements. All three are very common in imperative programming languages and have similar semantics in Python. The if statement executes one of its two blocks of statements, depending on the value of a test expression. The while expression repeats its body as long as its test expression is nonzero. The for statement iterates over a number of values, executing its body once for every value.

Unlike many other languages, the control structures of Python are non-scoping compound statements, i.e., statements that contain blocks of statements but do not define a lexical scope (see Chapter~\ref{section:statements-AST}). This means that the variables assigned inside an if, while or for statement are still bound after the execution of the statements.

The break and continue statements, respectively stop or interrupt a loop. They can only occur in the body of a for or while loop.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The if statement}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
An if statement~|IfStmt e b_if b_else| consists of the test expression~|e|, the blocks |b_if| and |b_else|. If |e| evaluates to a nonzero value, |b_if| is executed, else |b_else| is executed. We consider a value to be nonzero, if its class function~|__nonzero__| returns |True|.


\rulesection{Evaluating the test expression of an if statement}
\label{fw-rewrite-IfStmt}
First, the test expression |e| is evaluated, while the frame~|IfFrame1 b_if b_else| indicates that we still need to call |__nonzero__| and execute either |b_if| or |b_else|.

\begin{code}
rewrite (State  heap  envs  ( stack                          )  ( FwStmt (IfStmt e b_if b_else) )) =
        (state  heap  envs  ( stack :|: IfFrame1 b_if b_else )  ( FwExpr e                      ))
\end{code}


\rulesection{Testing the test value of an if statement}
\label{bw-rewrite-IfFrame1}
Next, we call the class function~|__nonzero__| to determine which block should be executed. The stack frame~|IfFrame2 b_if b_else| indicates that we are evaluating |a.__nonzero__()| and contains the two blocks. If the test value~|a| has no class attribute~|__nonzero__|, we raise a |TypeError|.

\pagebreak
%format a_TE = "a_{\mathtt{TE}}"
\begin{code}
rewrite                                       (State  heap  envs  ( stack :|: IfFrame1 b_if b_else )  ( BwResult a    ))
  | __nonzero__ `isin` getInhAttrs heap a  =  (state  heap  envs  ( stack :|: IfFrame2 b_if b_else )  ( FwApp a_f [a] ))
  | otherwise                              =  (state  heap  envs  ( stack                          )  ( raise a_TE [] ))
   where  a_f = get (getInhAttrs heap a) __nonzero__
          a_TE = a_TypeError
\end{code}


\rulesection{Executing the body of an if statement}
\label{bw-rewrite-IfFrame2}
Depending on the value of the test expression we execute either the |if| or the |else| block:

\begin{enumerate}
\item If |__nonzero__| returned a boolean, and it is |True|, we execute |b_if|.
\item If |__nonzero__| returned a boolean, and it is |False|, we execute |b_else|.
\item Otherwise, the call to |__nonzero__| returned a different class, so we return a |TypeError|.
\end{enumerate}

\begin{code}
rewrite                                                     (State  heap  envs  ( stack :|: IfFrame2 b_if b_else )  ( BwResult a           ))
  | isInstOf a a_bool heap && getBoolValue heap a        =  (state  heap  envs  ( stack                          )  ( FwBlock b_if         ))
  | isInstOf a a_bool heap && not (getBoolValue heap a)  =  (state  heap  envs  ( stack                          )  ( FwBlock b_else       ))
  | otherwise                                            =  (state  heap  envs  ( stack                          )  ( raise a_TypeError [] ))
\end{code}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The while statement}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The while statement~|WhileStmt e b| consists of a test expression~|e| and a body~|b|. Like the if statement, the while statement depends on the value of the test expression: if it evaluates to a nonzero value, the body is executed and the while statement (including the test) is executed again, otherwise the while-statement ends.


\rulesection{Evaluating the test expression of a while statement}
\label{fw-rewrite-WhileStmt}
First, we evaluate the test expression~|e| and place the while statement on the stack. The frame~|WhileFrame1 e b| indicates that we still have to call |__nonzero__| and contains both the body and test expression.

\begin{code}
rewrite (State  heap  envs  ( stack                     )  ( FwStmt (WhileStmt e b) )) =
        (state  heap  envs  ( stack :|: WhileFrame1 e b )  ( FwExpr e               ))
\end{code}


\rulesection{Testing the test value of a while statement}
\label{bw-rewrite-WhileFrame1}
Next, we call the class function~|__nonzero__| of the test value~|a|, or, if it has no such attribute, raise a |TypeError|.

\begin{code}
rewrite                                       (State  heap  envs  ( stack :|: WhileFrame1 e b )  ( BwResult a           ))
  | __nonzero__ `isin` getInhAttrs heap a  =  (state  heap  envs  ( stack :|: WhileFrame2 e b )  ( FwApp a_f [a]        ))
  | otherwise                              =  (state  heap  envs  ( stack                     )  ( raise a_TypeError [] ))
   where a_f = get (getInhAttrs heap a) __nonzero__
\end{code}


\rulesection{Executing the body of a while statement}
\label{bw-rewrite-WhileFrame2}
We execute the body |b| and rerun the while loop, if the test expression evaluated to a nonzero value. If not, we stop the loop:

\begin{enumerate}
\item If |__nonzero__| returned a boolean, and it is |True|, we execute the body, and place a while marker~|WhileMarker e b| on the stack.
\item If |__nonzero__| returned a boolean, and it is |False|, we stop of the while statement and return |None|.
\item Otherwise, the call to |__nonzero__| returned a different class, so we return a |TypeError|.
\end{enumerate}

\begin{code}
rewrite                                                     (State  heap  envs  ( stack :|: WhileFrame2 e b )  ( BwResult a           ))
  | isInstOf a a_bool heap && getBoolValue heap a        =  (state  heap  envs  ( stack :|: WhileMarker e b )  ( FwBlock b            ))
  | isInstOf a a_bool heap && not (getBoolValue heap a)  =  (state  heap  envs  ( stack                     )  ( BwResult a_None      ))
  | otherwise                                            =  (state  heap  envs  ( stack                     )  ( raise a_TypeError [] ))
\end{code}


\rulesection{Repeating a while statement}
\label{bw-rewrite-WhileMarker}
After the execution of the while body, the while marker on top of the stack is executed again. 

\begin{code}
rewrite  (State  heap  envs  ( stack :|: WhileMarker e b )  ( BwResult a             )) =
         (state  heap  envs  ( stack                     )  ( FwStmt (WhileStmt e b) ))
\end{code}

One might think that the while marker can be replaced by a simple while statement on the stack. However, the rules that handle continue and break statements depend on the difference (see Sections~\ref{section:break} and \ref{section:continue}).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The for statement}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
A for statement~|ForStmt x e b| consists of a variable name~|x| to which the values produced by the iterable~|e| are bound, and the body~|b|. 

An iterable is any object with a class function~|__iter__| that produces an iterator: an object with a class function~|__next__| that returns the next value if there is one, and raises a |StopIteration| exception when not. In particular, every generator is an iterator and every generator function is an iterable.

A for statement first creates an iterator by calling the class function~|__iter__| of the iterable~|e|. Then it executes the body~|b| repeatedly, calling the |__next__| function of the iterator. Binding the variable~|x| to the value returned by |__next__|. Once the |__next__| function raises a |StopIteration| exception, the for statement stops.


\rulesection{Executing the for statement}
\label{fw-rewrite-ForStmt}
First, the iterable~|e| is evaluated and the rest of the for statement is put on the stack. The stack frame~|ForFrame1 x b| indicates that we need to call |e.__iter__|.

\begin{code}
rewrite (State  heap  envs  ( stack                   )  ( FwStmt (ForStmt x e b) )) =
        (state  heap  envs  ( stack :|: ForFrame1 x b )  ( FwExpr e               ))
\end{code}


\rulesection{Get the iterator of the iterable}
\label{bw-rewrite-ForFrame1}
Next, if the iterable has an |__iter__| attribute, we call it, otherwise we raise a |TypeError|.

\begin{code}
rewrite                                    (State  heap  envs  ( stack :|: ForFrame1 x b ) ( BwResult a           ))
  | __iter__ `isin` getInhAttrs heap a  =  (state  heap  envs  ( stack :|: ForFrame2 x b ) ( FwApp a_f [a]        ))
  | otherwise                           =  (state  heap  envs  ( stack                   ) ( raise a_TypeError [] ))
   where a_f = get (getInhAttrs heap a) __iter__
\end{code}


\rulesection{Start the for loop with the iterator}
\label{bw-rewrite-ForFrame2}
Then, we start the for loop with the iterator~|a_i|, that has been returned by |__iter__|, and push a for marker~|ForMarker x a_i b| on the stack to start an iteration.

\begin{code}
rewrite  (State  heap  envs  ( stack :|: ForFrame2 x     b )  ( BwResult a_i    )) =
         (state  heap  envs  ( stack :|: ForMarker x a_i b )  ( BwResult a_None ))
\end{code}

The notation of the for marker suggests that the body is being executed. This is clearly not the case, instead, we pretend to just have executed the body in order to trigger Rule~\ref{bw-rewrite-ForMarker}.


\rulesection{Starting an iteration of the for loop}
\label{bw-rewrite-ForMarker}
An iteration starts with a call to the |__next__| function attribute of the iterator |a_i|. If no such attribute is available, then |a_i| is no iterator, so a |TypeError| is raised.

\begin{code}
rewrite                                      (State  heap  envs  ( stack :|: ForMarker x a_i b )  ( BwResult a           ))
  | __next__ `isin` getInhAttrs heap a_i  =  (state  heap  envs  ( stack :|: ForFrame3 x a_i b )  ( FwApp a_f [a_i]      ))
  | otherwise                             =  (state  heap  envs  ( stack                       )  ( raise a_TypeError [] ))
   where a_f = get (getInhAttrs heap a_i) __next__
\end{code}


\rulesection{Executing the body of a for loop}
\label{bw-rewrite-ForFrame3}
The value~|a| returned by the iterator, is bound to |x| and the body~|b| is executed. We schedule the next iteration by pushing a for marker~|ForMarker x a_i b| on the stack.

\begin{code}
rewrite (State  ( heap                   )  ( envs :|: env_1 )  ( stack :|: ForFrame3 x a_i b )  ( BwResult a )) = 
        (state  ( heap <+> [env_1 |-> m] )  ( envs :|: env_1 )  ( stack :|: ForMarker x a_i b )  ( FwBlock b  ))
  where m = get heap env_1 <+> [x |-> a]
\end{code}


\rulesection{Stopping the for loop}
\label{raise-rewrite-ForFrame3}
The for loop stops, when the iterator's  |__next__| function attribute raises a |StopIteration|: 
If we encounter a stack frame~|ForFrame3 x a_i b| while unwinding the stack with an exception |a|, we check whether the exception |a| is a |StopIteration|. If it is, normal execution resumes and we return |None|. Otherwise, the frame is popped and unwinding continues.

\begin{code}
rewrite                                 (State  heap  envs  ( stack :|: ForFrame3 x a_i b )  ( Raise a         ))
  | isInstOf a a_StopIteration heap  =  (state  heap  envs  ( stack                       )  ( BwResult a_None ))
  | otherwise                        =  (state  heap  envs  ( stack                       )  ( Raise a         ))
\end{code}



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The break statement} \label{section:break}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

A break statement immediately terminates execution of the current block and continues execution of the block that encapsulates the while or for statement. In other words, a break statement jumps directly to the end of the loop.

%A break statement that is not contained in a while statement is considered a syntactical error. Like the return statement, this restriction is checked at the parsing stage, and therefore need not be checked here in the semantics.

%Break statements, like raise and return statements, are executed by rewinding the stack until a frame of interest is encountered. The break statement rewinds the stack up to the first while marker where execution continues.


\rulesection{Rewinding the stack for a break statement}
\label{fw-rewrite-BreakStmt}
Break statements unwind the stack until a frame indicating the end of a while or for loop is encountered. If we encounter a finally clause~|FinallyMarker b| (see Section~\ref{fw-rewrite-FinallyStmt}), we execute its body~|b| and put the break statement on the stack. When |b| has been executed, unwinding will continue with the execution of the break statement.

\begin{code}
rewrite                  (State  heap  envs  ( stack :|: frame                         )  ( FwStmt BreakStmt )) = case frame of
  WhileMarker e b    ->  (state  heap  envs  ( stack                                   )  ( BwResult a_None  ))
  ForMarker x a_i b  ->  (state  heap  envs  ( stack                                   )  ( BwResult a_None  ))
  FinallyMarker b    ->  (state  heap  envs  ( stack :|: InstrFrame (FwStmt BreakStmt) )  ( FwBlock b        ))
  otherwise          ->  (state  heap  envs  ( stack                                   )  ( FwStmt BreakStmt ))
\end{code}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The continue statement} \label{section:continue}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Like a break statement, a continue statement terminates execution of a loop's body. But rather than jumping to the end of the loop, it starts another iteration. 


\rulesection{Rewinding the stack for a continue statement}
\label{fw-rewrite-ContinueStmt}
Continue statements unwind the stack until the end of a for or while loop. At the end of a loop, the loop is executed again. Like with break statements, any finally clauses will be executed `on the way out'.

In case of a for or while loop, the for or while marker is left on the stack, and normal execution is resumed. This will cause Rule~\ref{bw-rewrite-WhileMarker} or Rule~\ref{bw-rewrite-ForMarker} respectively, to run another iteration.

\begin{code}
rewrite                  (State  heap  envs  ( stack :|: frame                            )  ( FwStmt ContinueStmt )) = case frame of
  WhileMarker e b    ->  (state  heap  envs  ( stack :|: WhileMarker e b                  )  ( BwResult a_None     ))
  ForMarker x a_i b  ->  (state  heap  envs  ( stack :|: ForMarker x a_i b                )  ( BwResult a_None     ))
  FinallyMarker b    ->  (state  heap  envs  ( stack :|: InstrFrame (FwStmt ContinueStmt) )  ( FwBlock b           ))
  otherwise          ->  (state  heap  envs  ( stack                                      )  ( FwStmt ContinueStmt ))
\end{code}



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%if style==newcode
\begin{code}
rewrite _ = skipState
\end{code}
%endif
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%}
