%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Conclusion} \label{chapter:conclusion}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

We have created an executable operational semantics of a significant subset of Python. The subset, called \minpy, covers all major constructs of Python. The semantics are easy to read, insofar the formal semantics of a serious programming language can be. 
%Additionally, this document is self-contained, although we do rely on some basic mathematical notation and definitions.

%Python has shown itself to be a quite suitable subject for formalisation. A simple abstract machine model based on a number of simple data structures, stacks, lists, and dictionaries as mappings, proved sufficient to model its operational semantics.

%The rules that define the semantics have been compiled into an interpreter.
%We have created a test suite that tests the interpreter and, by proxy, the semantics.



%The minimalistic design of Python has also helped. For example, normal functions, class functions and generators are all defined using the same construct. The semantics of the function definition is not overloaded beyond recognition: Rule~\ref{fw-rewrite-DefStmt} defines the semantics of all three variants in a single rule.


% Ultimately this is for the reader to judge. 



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Literate programming}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Developing an executable operational semantics was a challenging exercise of literate programming as introduced by Knuth \cite{knuth1984lp}.

In contrast to Knuth's WEB system, which transparently includes a program's source code into a document, we have tried to hide the Haskell source code in this document. Had we simply included the Haskell code, it would have been perceived as just an implementation of Python in Haskell, rather than an operational semantics of Python.

The Haskell source code was formatted with lhs2\TeX. This system uses formatting rules to translate Haskell to \LaTeX, which can be edited independently from source code that is being formatted. For example, to reveal the Haskell code, one would simply remove the formatting rules and recompile the document. 

%We used the powerful, yet easy to use\footnote{The alignment is controlled by spacing in the source code} mechanisms for horizontal alignment of lhs2\TeX to align the elements of the abstract machine.

We pushed lhs2\TeX to its limits. It was designed for small formatting tweaks, not to completely transform the source code as we did. The sources had to be pre-processed by a simple Perl script and stick to strict coding conventions to get the current results.



% readable code (literate programming), balance reading with coding
% literate programming forces us to be complete, but what to hide, and how to explain dependencies
%- more on the style of the rw rules (complicated abstract machine, format is close to the trace)


%\subsection*{Haskell as a specification language}
The choice of Haskell as the specification language turned out rather well. Its declarative nature and almost mathematical syntax allowed us to write Haskell code that is close to the eventual formatted formulas. 

In addition, Haskell provides many static checks which prevent many small errors. For example, undeclared variables or functions (often caused by typos) are detected, and the order and number of function arguments is verified by the type system.

The type system has also given us much more confidence in the correctness of the semantics. Despite its type inference, the type system of Haskell introduces some overhead. For example, in the semantics we rely on a convention to separate addresses from other types of values: they are all described by a variable~|a| or |env|. Haskell requires us to explicitly tag these variables, e.g. \verb#Addr a#.

Since the type system also captures side effects, it prevents any accidental introduction of a hidden state that can affect the semantics. For example, the guard expressions used to express conditional rewrite rules cannot have side effects. After all, the act of searching for the rewrite rule that applies to a specific machine state should not influence the semantics.


Hiding Haskell was not always easy. Some concepts could have been described very concisely using a number of standard library functions, but to make this document self contained, many library functions could not be used. The formatting of Haskell also limits us to a subset of Haskell.

While we wanted to hide Haskell in the document, we've tried to keep the source code as close to the formatted rules as possible. Haskell proved quite flexible in this respect. For example, the operator~|<+>| is the reformatted version of \verb#<+># in the Haskell code.


Because of the code conventions required for proper formatting and the restrictions on Haskell, the source code of the rewrite rules will not win any beauty contests (see Section \ref{section:rule-source}). Nevertheless, it is almost certainly easier to read and write than any (non-executable) equivalent \LaTeX\ code.

Overall development on the semantics is quite a satisfying experience. When changing the semantics one would typically write some tests to explore the desired semantics, then implement a solution in Haskell, and finally, when the implementation passes the tests, ensure that it is properly formatted.

The Glasgow Haskell Compiler (GHC), directly supports literate Haskell. The source files do not need to be preprocessed by a separate tool but can be fed into GHC directly. Thus, one can disregard the restrictions while experimenting and worry about the formatting later.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The interpreter}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The sources of this document and some supporting code that implements a parser, pretty printer, and an interactive interpreter, can be compiled by the Haskell compiler GHC to produce an interpreter. Like the interpreter of CPython, our interpreter has both a command-line modus and the ability to interpret files.

The following example shows a session of the interpreter in its command-line modus. First, we declare the factorial function in a single function definition statement. The statement is immediately executed if parsing succeeds. Next, we call the factorial function without an argument, which results in a |TypeError| exception. Finally we call it with a correct argument and exit the interpreter.

\begin{verbatim}
gideon@gideon-desktop:~$ minpy 
>>> def fac(x) :
...     if x <= 0 :
...         return 1
...     else :
...         return fac(x-1) * x
... 
>>> fac()
TypeError
>>> fac(4)
24
>>> exiting minpy
\end{verbatim}

Without a standard library and a foreign function interface, the interpreter cannot be used for many serious applications. However, as a platform of experimentation and testing, the interpreter was invaluable.

The interpreter includes a tracer that has proven to be very helpful when debugging the rewrite rules. The tracer prints the changes of the abstract machine state. 

The following interpreter session shows how the tracer is used. The tracer is first enabled by calling \texttt{enableTrace()} of the \texttt{sys} module. After a simple assignment statement the trace shows how the value~1 is stored at address 1006, and the environment with address 22 is updated (in \texttt{state\_5}) with a new mapping. 

The interpreter always prints the result of the last statement if the result is not |None|. So when we execute the expression statement |a|, the value is retrieved (states 7 and 8) and printed (states 10 through 15).

\begin{verbatim}
>>> import sys
>>> sys.enableTrace()
state_1 = <heap_0, []:|:29:|:22, [], 0> 
>>> a = 1
state_3 = <heap_2, envs_2, stack_2:|:AssFrame "a", 1> 
state_4 = <heap_3<+>1006 |-> IntValue 1, envs_2, stack_3, 1006> 
state_5 = <heap_4<+>22 |-> ModValue (fromList [("a",1006),("sys",25)]), 
[]:|:29:|:22, stack_2, 0> 
>>> a
state_7 = <heap_6, envs_6, stack_2, a> 
state_8 = <heap_7, []:|:29:|:22, [], 1006> 
state_10 = <heap_9, envs_9, stack_9:|:PrintFrame1, $1006> 
state_11 = <heap_10, envs_9, stack_10, 1006> 
state_12 = <heap_11, envs_9, stack_9:|:PrintFrame2, int.__str__:: (7) 
-> ?([1006])> 
state_13 = <heap_12, envs_9, stack_12, int.__str__([1006])> 
state_14 = <heap_13<+>1007 |-> StringValue "1", envs_9, stack_12, 1007> 
1
state_15 = <heap_14, envs_9, stack_9, 0> 
\end{verbatim}

Traces usually include a lot of addresses that are hard to keep track of. We implemented address expressions that allow us to handle addresses as first class object in the language for debugging. For example, if we had forgotten what the address 1006 pointed to we can retrieve it by entering the expression statement \texttt{\$1006}.

\begin{verbatim}
>>> $1006
1
>>> exiting minpy
\end{verbatim}


The interpreter does not only help when debugging but also provides insight in the semantics.  Some aspects of the operational semantics `emerge' from a number of rules.  Judging where these emergent semantics occur and exploring their consequences, is often difficult. The interpreter allows us to explore these issues using examples.

For example, the interaction of the try-finally statement with the return statement (see Section~\ref{section:try-finally}) has some unexpected consequences for the semantics. It is quite hard to infer the semantics of this exceptional case from the rules. Running an example in the interpreter however, immediately reveals the emergent semantics.


%\subsection*{Performance of the interpreter}
Our interpreter runs considerably slower than CPython. This is no surprise. We paid little attention to the performance of the interpreter. 

It should be noted though, that the difference is not a simple constant factor. The heap is implemented with a balanced binary tree, whereas CPython uses a direct mapping of memory addresses to values. For a simple address lookup, the former has a complexity of $O(log(n))$, whereas the latter has a constant complexity.

Due to the absence of any garbage collector in the interpreter our interpreter uses much more memory than CPython. Indeed, it will be hard to test any memory intensive programs and even a simple program such as \texttt{while True: pass} will eventually crash due to a lack of memory.

%Moreover, Haskell has call-by-need semantics as opposed to the more traditional call-by-value semantics of for example C and Python. In call-by-value semantics, the arguments of function are always evaluated before executing the function. Haskell, with its call-by-need semantics, postpones evaluation until the value of an expression is needed. 

%The call-by-need semantics might cause computations to occur in a different order than they would in CPython. The actual result of the Python program would not be affected, but the execution of a program could have different performance characteristics.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The test suite} \label{section:tests}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
As mentioned before, our executable operational semantics is testable: we can execute programs and verify whether the \minpy\ interpreter gives the same answer as the CPython interpreter.

The CPython implementation is developed with a suite of regression tests. The regression tests could unfortunately not be used to test \minpy\, because it relies on many Python features and libraries that are not available in \minpy. 

Fortunately the efforts to reverse engineer the semantics of CPython provided an excellent opportunity to write new tests. The resulting test suite consists of 134 tests and is completely written in \minpy, including the test driver.

Beside testing the basic behaviour of constructs, the tests verify aspects such as control flow, order of evaluation, type/class requirements, and variable scoping. Each test covers a single aspect of a single construct. 

We have been able to run the test suite for a number of implementations of Python (see Table~\ref{table:test-results}). Testing other implementations improves confidence in the tested implementations. 

Judging from the low number of failed tests across implementations, the other implementations seem to implement the same language, and conversely, \minpy\ and the test-suite seem to capture a common subset of Python. Hopefully \minpy\ can converge on the essential core of Python.

\begin{table}[h]
\begin{center}
\begin{tabular}{ll}
\textbf{Implementations}               & \textbf{Failed} \\
CPython 2.5.2 (reference)              & 1 \\
\minpy                                 & 13 \\
PyPy 1.0.0 build 41081                 & 3 \\
Iron Python 1.1.1 on .net 2.0.50727.42 & 15 \\
Jython 2.2.1 on java1.6.0\_0            & 17
\end{tabular}
\end{center}
\caption{Number of failed tests for various Python implementations.}
\label{table:test-results}
\end{table}

Since CPython is the reference implementation one would expect no errors. There is, however, one test that crashes the Python interpreter completely. The developers of CPython acknowledged that this was a bug and immediately solved the problem. None of the other implementations crashed in the same situation.  

Our interpreter fails 13 tests, all of which are by definition errors in our semantics. Most of the errors are considered missing features. For example, lists in \minpy\ raise an exception when a negative index is used. CPython on the other hand, treats negative indices as normal indices on the reversed list with an offset of $1$. Thus in CPython $[1,2][-1]$ evaluates to the value $2$ but raises an exception in \minpy.

Of course there are many more errors in \minpy\ than have been tested or even discovered. Some of the known issues are listed in Section \ref{section:future-work}.

%- import self fails
%- missing __dict__
%- construction: new doesn't call init
%- construction: don't type check result of __new__
%- lambda not implemented
%- exec without explicit environment not implemented
%- neg indices for lists not implemented

Interpreting the test results of other Python implementations is difficult. Some differences are intentional and can therefore not be considered bugs. Whether or not a failed test indicates a bug is a matter of opinion. Nevertheless we tried to interpret the results somewhat:

The PyPy interpreter failed three tests due to incompatibilities only; none of the failures indicated bugs. Iron Python failed quite a number of tests, but only five of those can be considered bugs. Of the failed tests on Jython, only three can be considered bugs.

%Fortunately \minpy\ has turned out to roughly correspond with the intersection of quite a few versions and implementations of Python (see Section~\ref{section:tests}).

%Especially in conjunction with the test suite, the interpreter enables experimentation. For example, when we added the specification of the generators, we first wrote tests and ran them in the interpreter during development. At the same time, the old tests in the test suite ensured that we did not break earlier work.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%\section{Discussion}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%The handling of local function variables (see Section~\ref{section:functions-local-variables}), which we also consider one of the weaker parts of the semantics.


%\subsubsection*{The type system}
%Python is often said to support ``duck typing''\footnote{Duck typing refers to the duck test of J.W. Riley, paraphrasing: ``If it walks like a duck and quacks like a duck, it must be a duck.''}, a typing mechanism that discriminates classes by their structure. This is usually contrasted with the nominal typing scheme employed by for example Java, where classes are discriminated by their name and names of its parents.

%Our operational semantics however, reveal both a nominal type system and a structural one. For example, numbers and containers are structural types: any object implementing the right attributes, is considered a number or container (see Chapter \ref{chapter:operators}). Whereas exception handling relies on nominal types to discriminate one exception from the other (see \ref{chapter:exceptions}).

%Apart from the operational semantics in general, there are two specific contributions worth pointing out. Firstly, we have created a new, non-algorithmic specification of the method resolution order (see Section~\ref{section:MRO}). Secondly, we have specified the behaviour of generators, which is a concept unique to Python.

%The object system has been specified in great detail. We have found a new, non-algorithmic specification of the method resolution order (see Section~\ref{section:MRO}), and gained insight into the role of primitive functions in the object system (see Chapter~\ref{chapter:objects}).



%\subsubsection*{Object system and primitives}
%The primitive class functions (see Chapter~\ref{chapter:objects}) have a central role in the semantics of Python's objects. For example the primitive function~|__getattribute__| handle the lookup of attributes and instantiation of class functions.


%Python's extremely dynamic nature actually simplify the semantics. Many languages have complicated static semantics, such as a static type system. In Python there are no separate phases of execution other than parsing and execution phases.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Future work} \label{section:future-work}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
There is an enormous body of work that can be based on this operational semantics. First of all, \minpy\ can be expanded to cover all, or at least a more complete subset of Python. 

%Even though we limited the scope of the semantics (see Section~\ref{section:scope}), not all of our goals have been achieved, due to time constraints:

The grammar of \minpy\ and therefore the specification, is much simpler than Python's syntax (see Chapter~\ref{chapter:AST}). Most significantly, list and generator comprehensions, keyword arguments and other advanced function parameters, the with statement, and the global statement are missing. 

Many optional elements and other syntactic sugar are also missing in \minpy. This makes the semantics less verbose and allows us to concentrate on the most interesting aspects of Python. Nevertheless, for the sake of completeness, they should eventually be specified.

It should be noted that many constructs that could be considered syntactic sugar, actually have semantics that cannot completely be described by a simple transformation. For example, the augmented assignment $x~{+=}~1$ is not equivalent to $x~=~x+1$: in the former case Python would call the class function \texttt{\_\_iadd\_\_}, but for the latter it would call |__setattr__| and |__add__|.

We are confident that all constructs of Python that are missing from \minpy\ can be added without significant changes to the existing semantic rules. The structure of the abstract machine will not have to be changed, and the framework used to describe the rules will not need significant extensions.


Secondly, there are some errors in the current semantics that should be corrected. The test-suite pointed out some errors in the semantics of class initialization. Although we do use the class of a class --the metaclass-- to construct classes, the mechanics of metaclasses have not been covered completely (see Section~\ref{section:class-definition}, specifically Rule~\ref{bw-rewrite-ClassFrame}).

Most of the built-in types have been specified to make the interpreter usable (see Chapter~\ref{chapter:builtins}). But the current semantics of \minpy's built-in types are not conform to Python. For example, \minpy\ does not support coercion, and the dictionary class only accepts strings as keys.


Thirdly, the scope of the semantics (see Section~\ref{section:scope}) can be expanded to encompass more of Python's aspects: 
\begin{itemize}
\item Garbage collection and its interaction with the operational semantics can be explored an formalised.

\item Concurrent programming, notorious for its complexity, can be formalised.

\item The interaction of Python with the outside world --the foreign function interface (FFI)-- can be specified.

\item The various reflective features can be formalised. It seems that most of these features can be specified fairly easily. For example, the |__dict__| attribute that exposes the attribute mappings of an object, would not require big changes.
\end{itemize}

For completeness one might consider to add so called old-style classes as well. It might be more productive to update the entire specification to a newer version of Python, that will does not support old style classes.


Fourthly, the interpreter, which will benefit from all improvements of the semantics, can be improved. Currently the error feedback is minimal. It would be very useful if raised exceptions also print a stack trace.

The performance of the interpreter can probably be improved, but it is doubtful it will ever achieve performance comparable to CPython without compromising the legibility of the semantics.

The interpreter could also be used as a basis for various other tools. The current basic tracing functionality could be extended to a fully fledged debugger for example.


%describe the semantics of non-core aspects of Python
%Improve the interpreter
%improve the presentation of rewrite rules
%
%develop tools analyses and extentions such as a static type system
%prove theorems over the python language


% expand the interpreter to a full Python interpreter:
% improve error feedback of the interpreter
% improve the trace output
%While the interpreter cannot execute most Python programs due to missing libraries and limited grammar, it does serve its purposes: it shows what changes of the semantics do with the language, it enables quick experimentation and makes the semantics testable (see Section~\ref{section:tests}).

%Even more interesting are the avenues of new research opened by our operational semantics. The lack of a formal semantics for Python prevented any formal work on Python. Now however we can pursue various new directions of research:

Finally, our work opens many new avenues of research. The lack of a formal semantics for Python prevented any formal work on Python. Now however we can pursue various new directions of research:

\begin{itemize}
\item One can prove properties of Python with respect to our semantics. For example, we would very much like to prove that the execution of any possible Python program will never end in a state, other than the final state, that cannot be rewritten. Such a proof would have prevented the bug that we found in CPython.

\item New tools for and analyses of Python programs can be developed more easily. For example, a tool that type-checks a Python program could be developed. Perhaps some properties of the type system could even be proven correct.

\item Language extensions can be developed within our framework. Resulting in both a formal description of the extension and a proof of concept interpreter.
\end{itemize}
