%if style == newcode
\begin{code}
module Exceptions
  ( rewrite
  , baseException
  , typeError
  , nameError
  , attributeError
  ) where
import VirtualMachine
import ObjectTheory
\end{code}
%endif

%{

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Exceptions} \label{chapter:exceptions}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Exceptions simplify handling of errors by providing an alternative mode of execution. When an exceptional situation occurs i.e., an exception is raised, normal execution stops, until the situation is dealt with.


Listing~\ref{example:exceptions} shows how exceptions can be used to handle divisions by zero, how normal execution stops when an exception is raised, and how the exception is handled.

\begin{lstlisting} [ caption={Raising and catching a divide by zero exception}
                   , label=example:exceptions ]
class ZeroDivisionError(Exception) :
    def __str__(self) :
        return "Divided a number by zero!"

def div(x, y) :
    if y == 0 :
        raise ZeroDivisionError()
    
    return x / y

try :
    div(2, 0) # will raise the exception
    print("This will never be printed")
except Exception, e : # catch all exceptions from within the try block
    print(e)

print("Normal exectution continues")
\end{lstlisting}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%\section{The exception handling statements}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Several constructs handle exceptions together:
%\begin{itemize}
%\item Raise statements throw exceptions.
%\item Try-except statements catch expressions that are raised in their body.
%\item Try-finally statements execute their body, with the guarantee that the statements in the finally-clause will be executed, even if an exception is raised.
%\end{itemize}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The raise statement}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
A raise statement |RaiseStmt e| stops normal execution and raises |e| if and only if it evaluates to an instance of |BaseException|. During normal execution we would continue with the next frame on the stack, whereas most stack frames are simply discarded while raising an exception. Only some stack frames and markers will be handled while raising an exception.


\rulesection{Executing the raise statement}
\label{fw-rewrite-RaiseStmt}
The raise statement first evaluates the exception expression |e|.

\begin{code}
rewrite  (State  heap  envs  ( stack                )  ( FwStmt (RaiseStmt e) )) =
         (state  heap  envs  ( stack :|: RaiseFrame )  ( FwExpr e             ))
\end{code}


\rulesection{Start unwinding of the stack}
\label{bw-rewrite-RaiseFrame}
If the raised value |a| is an instance of |BaseException|, unwinding of the stack starts. Otherwise, a |TypeError| is raised. Raising an exception has no effect on the environment.

\begin{code}
rewrite                                   (State  heap  envs  ( stack :|: RaiseFrame )  ( BwResult a           ))
  | (isInstOf a a_BaseException heap)  =  (state  heap  envs  ( stack                )  ( Raise a              ))
  | otherwise                          =  (state  heap  envs  ( stack                )  ( raise a_TypeError [] ))
\end{code}


\rulesection{Raising an exception}
\label{raise-rewrite-frame}
By default, stack frames are simply discarded while unwinding the stack until the stack is empty and execution stops. Note that this is a very unspecific rule and hence will be overridden by rules such as Rule~\ref{raise-rewrite-ReturnMarker}.

%if style==newcode
\begin{code}
rewrite s = rwpart1 s
\end{code}
%endif

\begin{code}
rwpart2  (State  heap  envs  ( stack :|: frame )  ( Raise a )) =
         (state  heap  envs  ( stack           )  ( Raise a ))
\end{code}

%if style==newcode
\begin{code}
rwpart2 _ = skipState
\end{code}
%endif


\rulesection{Raising an exception in a function}
\label{raise-rewrite-ReturnMarker}
If an exception remains uncaught in a function, we return from the function by reinstating the environment |envs'| found on the return marker introduced in Rule~\ref{fw-rewrite-FwApp}, and resume unwinding.

\begin{code}
rwpart1  (State  heap  envs   ( stack :|: ReturnMarker envs' )  ( Raise a )) =
         (state  heap  envs'  ( stack                        )  ( Raise a ))
\end{code}


%format a_g
\rulesection{Raising an exception in a generator}
\label{raise-rewrite-GenValue}
When an exception is raised in a generator, we will encounter the yield marker~|GenMarker envs' a_g| introduced in Rule~\ref{fw-rewrite-Prim-generator.__next__}. Once a generator has raised an exception, it is stopped: the current generator~|a_g| is updated with an empty generator~|GenValue EmptyStack EmptyStack| and the old environment |envs'| is reinstated (see also Rule~\ref{bw-rewrite-GenMarker}).

\begin{code}
rwpart1  (State  ( heap                                              )  envs   ( stack :|: GenMarker envs' a_g )  ( Raise a )) =
         (state  ( heap <+> [a_g |-> GenValue EmptyStack EmptyStack] )  envs'  ( stack                         )  ( Raise a ))
\end{code}


%format b_e
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The try-except statement}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
A raised exception is handled by a try-except statement. A try-except statement~|TryStmt b e x b_e| executes its body~|b|. If an exception is raised in the body and it is an instance of |e|, the raised exception is bound to |x| and normal execution resumes with the except clause~|b_e|. If a raised exception is not an instance of |e|, the exception is re-raised and unwinding continues.


\rulesection{Executing the try-except statement}
\label{fw-rewrite-TryStmt}
The try-except statement places the except clause~|ExceptMarker e x b_e| on the stack, and executes the body~|b|.

\begin{code}
rwpart1  (State  heap  envs  ( stack                          )  ( FwStmt (TryStmt b e x b_e) )) =
         (state  heap  envs  ( stack :|: ExceptMarker e x b_e )  ( FwBlock b                  ))
\end{code}


\rulesection{Popping an exception clause while unwinding}
\label{bw-rewrite-Raise-ExceptMarker}
When an except clause is encountered while unwinding, its exception class~|e| is evaluated in order to check whether it catches the exception. The raised exception~|a| is put on the stack, along with the exception clause.

%format b_e
\begin{code}
rwpart1  (State  heap  envs  ( stack :|: ExceptMarker e x b_e  )  ( Raise a  )) =
         (state  heap  envs  ( stack :|: CatchingFrame a x b_e )  ( FwExpr e ))
\end{code}


%format a_e
\rulesection{Catching an exception}
\label{raise-rewrite-ExceptMarker}
Once the class of the except clause has been evaluated, we determine whether it catches the raised exception:

\begin{enumerate}
\item If the class |a_e| of the except clause is an exception (i.e. it is a |BaseException|) and the raise exception |a| is an instance of |a_e|, we catch the exception. The exception |a| is bound to |x| and the |b_e| is executed.
\item Otherwise, we re-raise the exception |a| and pop the except clause.
\end{enumerate}

%{
%format isSubclass heap a_o a_e = a_o "<:_{\Theta}" a_e
\begin{code}
rwpart1                                                           (State  heap  ( envs :|: env_1 )  ( stack :|: CatchingFrame a x b_e )  ( BwResult a_e ))
  | isSubclass heap a_e a_BaseException && isInstOf a a_e heap =  (state  heap' ( envs :|: env_1 )  ( stack                           )  ( FwBlock b_e  ))
  | otherwise                                                  =  (state  heap  ( envs :|: env_1 )  ( stack                           )  ( Raise a      ))
   where heap' = heap <+> [env_1 |-> get heap env_1 <+> [x |-> a]]
\end{code}

The class of an except clause |a_e| can be any value. It is not at all limited to classes, or even exceptions. However, if |a_e| is not an exception class, it will never catch anything.
%}


\rulesection{Popping the exception marker}
\label{bw-rewrite-ExceptMarker}
The exception clause is simply popped from the stack when no exception has been raised in the body.

\begin{code}
rwpart1  (State  heap  envs  ( stack :|: ExceptMarker e x b_e )  ( BwResult a )) = 
         (state  heap  envs  ( stack                          )  ( BwResult a ))
\end{code}


%format b_f
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The try-finally statement} \label{section:try-finally}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Sometimes it is necessary to perform some operations even when an error has occurred. The try-finally statement~|FinallyStmt b b_f| executes its body~|b| with a guarantee that |b_f| will be executed. We would typically use a try-finally statement to ensure some clean-up code is executed, even if an exception is raised in its body.

The guarantee of executing |b_f| applies not only to exception handling. For example, if there is a return statement in the body~|b|, we will execute |b_f| before we return from the function. If |b_f| contains another return statement, the last return statement will return from the function. This rather unintuitive behaviour has prompted the CPython developers to give a warning when the |b_f| contains a yield or return statement.


\rulesection{Executing the try-finally statement}
\label{fw-rewrite-FinallyStmt}
The try-finally starts like the try-except statement, by executing its body. The finally block |b_f| is placed on the stack.

\begin{code}
rwpart1  (State  heap  envs  ( stack                       )  ( FwStmt (FinallyStmt b b_f) )) =
         (state  heap  envs  ( stack :|: FinallyMarker b_f )  ( FwBlock b                  ))
\end{code}


\rulesection{Popping a finally clause while unwinding}
\label{raise-rewrite-FinallyMarker}
When a finally clause in encountered while unwinding, the raised exception |a| is put on the stack and the finally clause |b_f| is executed.

%format b_f
%format a_e
\begin{code}
rwpart1  (State  heap  envs  ( stack :|: FinallyMarker b_f      )  ( Raise a_e   )) =
         (state  heap  envs  ( stack :|: InstrFrame (Raise a_e) )  ( FwBlock b_f ))
\end{code}


\rulesection{Popping the finally marker}
\label{bw-rewrite-FinallyMarker}
When no exception has occurred, the try-finally statement executes the finally clause that has been placed on the stack.

\begin{code}
rwpart1  (State  heap  envs  ( stack :|: FinallyMarker b_f )  ( BwResult a  )) =
         (state  heap  envs  ( stack                       )  ( FwBlock b_f ))
\end{code}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%if style==newcode
\begin{code}
rwpart1  (State  ( heap                                                                  )  envs  stack  ( FwApp (Prim "BaseException" "__str__") [a_self] )) =
         (state  ( heap <+> [a |-> StringValue ("<exception at " ++ show a_self ++ ">")] )  envs  stack  ( BwResult a                                      ))
  where a = freeAddr heap
\end{code}

\begin{code}
rwpart1  (State  ( heap                                       )  envs  stack  ( FwApp (Prim "TypeError" "__str__") [a_self] )) =
         (state  ( heap <+> [a |-> StringValue ("TypeError")] )  envs  stack  ( BwResult a                                  ))
  where a = freeAddr heap
\end{code}

\begin{code}
rwpart1  (State  ( heap                                       )  envs  stack  ( FwApp (Prim "NameError" "__str__") [a_self] )) =
         (state  ( heap <+> [a |-> StringValue ("NameError")] )  envs  stack  ( BwResult a                                  ))
  where a = freeAddr heap
\end{code}

\begin{code}
rwpart1  (State  ( heap                                            )  envs  stack  ( FwApp (Prim "AttributeError" "__str__") [a_self] )) =
         (state  ( heap <+> [a |-> StringValue ("AttributeError")] )  envs  stack  ( BwResult a                                       ))
  where a = freeAddr heap
\end{code}
%endif

%if style==newcode
\begin{code}
rwpart1 s = rwpart2 s
\end{code}
%endif
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%if style==newcode
\begin{code}
baseException :: Value
baseException = new
  [ __bases__  |-> a_objectT
  , __class__  |-> a_Type
  , __str__    |-> Typed (Prim "BaseException" "__str__") [a_BaseException]
  ]

typeError :: Value
typeError = new
  [ __bases__  |-> a_BaseExceptionT
  , __class__  |-> a_Type 
  , __str__    |-> Typed (Prim "TypeError" "__str__") [a_TypeError]
  ]

nameError :: Value
nameError = new
  [ __bases__  |-> a_BaseExceptionT
  , __class__  |-> a_Type 
  , __str__    |-> Typed (Prim "NameError" "__str__") [a_NameError]
  ]

attributeError :: Value
attributeError = new
  [ __bases__  |-> a_BaseExceptionT
  , __class__  |-> a_Type 
  , __str__    |-> Typed (Prim "AttributeError" "__str__") [a_AttributeError]
  ]
\end{code}
%endif
%}
