%if style == newcode
\begin{code}
module Functions(rewrite, o_function, o_generator) where
import VirtualMachine hiding (localVars)
import ObjectTheory
\end{code}
%endif

%{
%if not appendix
%format hasYield b = "\mathtt{yield} \in" b
%format hasNotYield b = "\mathtt{yield} \notin" b
%format localVars b = "\text{lv}"(b)
%format e_f
%format envs_lambda
%format lambda = "\lambda"
%format g = "\text{gen}"
%format b_lambda
%format a_lambda
%format x_lambda
%format noArgs =
%format xs_1
%format xs_2
%format foo = \mathtt{foo}
%format isObjValue (v) = v "\in \Omega"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Functions and generators} \label{chapter:functions}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
This chapter discusses functions and generators. Following the lifetime of a function, we define the semantics of functions in three steps: first the function definition, followed by the function call, execution, and the return. Next we discuss the operation of generators: (re)starting a generator, and 'returning' from a generator with the yield statement.

Functions in Python behave very similar to functions in other imperative language. A function takes a number of arguments, executes its body, and returns some value. For example Listing~\ref{example:functions} shows how a factorial function is defined and called.

\begin{lstlisting}[ caption={A recursive factorial function} 
                  , label=example:functions ]
def fac(x) :
	if x < 1 :
		return 1
	else :
		return x * fac(x-1)

fac(5) # returns 120
\end{lstlisting}

Generators are a unique concept of Python. They are defined using a normal function definition that contains one or more yield statements and no return statements (see Chapter~\ref{chapter:AST}). Such a generator function returns a generator when called.

When a generator's primitive function~|__next__| is called, the generator executes the body of the generator function until a yield expression~|YieldExpr e| is evaluated. Evaluating the yield expression causes the generator to suspend execution and return |e|. When |__next__| is called again, execution continues at the point where it left off, until either another yield expression is encountered or the end of the body is reached. At the end of its body, a generator stops and raises an exception.

For example Listing~\ref{example:generators} shows how a generator is used to create a generator that returns the elements of a list one by one.

\begin{figure}
\begin{lstlisting}[ caption={A generator function that produces a list iterator}
                  , label=example:generators ]
def createListGenerator(list) :
	try :
		i = 0
		while True :
			yield list[i]
			i = i + 1
	except IndexError, e :        # catch index out of bounds
		pass

it = createListGenerator([1,2])
it.next() # returns 1
it.next() # returns 2
it.next() # raises a StopIteration exception
\end{lstlisting}
\end{figure}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Defining functions and generators}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
A function definition~|DefStmt x_lambda xs b_lambda| consists of: the function's name~|x_lambda|, the list of parameter names~|xs|, and the body~|b_lambda|. When executed, a new function object is added to the heap, the function name is bound in the environment, and |None| is returned. 

Functions which body contains a yield expression, produce generators when called (see Section~\ref{section:generators}). The definition of these generator functions however, is no different from normal functions. 

\rulesection{Executing a function definition statement}
\label{fw-rewrite-DefStmt}
The new function object stored at |a_lambda| consists of two parts: the map of attributes~|[__class__ ||-> a_function]|, which only contains the class attribute, and a function closure value. The closure~|FuncValue xs b_lambda envs_lambda| consists of: the function arguments~|xs|, its body~|b_lambda|, and its environment~|envs_lambda|.

\begin{code}
rewrite  (State  heap   ( envs :|: env_1 )  stack  ( FwStmt (DefStmt x_lambda xs b_lambda) )) =
         (state  heap'  ( envs :|: env_1 )  stack  ( BwResult a_None                       ))
   where  a_lambda     = freeAddr heap
          envs_lambda  = if isObjValue (get heap env_1) then envs else envs :|: env_1
          heap'        = heap <+>  [  a_lambda  |-> ObjValue (new [__class__ |-> a_function]) (FuncValue xs b_lambda envs_lambda)
                                   ,  env_1     |-> get heap env_1 <+> [x_lambda |-> a_lambda]
                                   ]
\end{code}

Class functions, recognized by the topmost environment (see Section \ref{section:class-definition}), cannot access other class members directly. Therefore, the environment~|envs_lambda| of class functions does not include the local environment~|env_1| of the function definition statement. See Example~\ref{example:class-function-scope}.

\begin{lstlisting}[ caption={Class variables are not in the scope of class variables}
                  , label=example:class-function-scope ]
x = 1
class A(object) :
	x = x*2 # define class variable A.x

	def foo(self) :
		return x

print(A().foo()) # prints 1
print(A.x)       # prints 2
\end{lstlisting}



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Executing functions} \label{section:function-execution}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsection*{Evaluating calls}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
A call expression~|CallExpr e_f es| executes a function~|e_f| with arguments~|es|. Before the function is executed, |e_f| and |es| are evaluated from left to right.


\rulesection{Evaluating the function expression}
\label{fw-rewrite-CallExpr}
First, the function expression |e_f| is evaluated. A frame~|CallFrame es| is pushed onto the stack containing the remaining argument expressions.

\begin{code}
rewrite  (State  heap  envs  ( stack                  )  ( FwExpr (CallExpr e_f es) )) =
         (state  heap  envs  ( stack :|: CallFrame es )  ( FwExpr e_f               ))
\end{code}


%format es'
%format e'_1
\rulesection{Evaluating the first argument expression}
\label{bw-rewrite-CallFrame}
Then, when |e_f| has been evaluated to |a_f|, we continue to evaluate the first argument~|e'_1| and push the remaining arguments~|es'| and |a_f| on the stack. If there are no arguments, |a_f| is executed (see Section \ref{section:closure-execution}).

%format fstArgFrame a (es) = a("\circ, "es)
\begin{code}
rewrite          (State  heap  envs  ( stack :|: CallFrame es        )  ( BwResult a_f      )) = case es of
  (e:es')    ->  (state  heap  envs  ( stack :|: fstArgFrame a_f es' )  ( FwExpr e          ))
  otherwise  ->  (state  heap  envs  ( stack                         )  ( FwApp  a_f noArgs ))
\end{code}


\rulesection{Evaluating the remaining argument expressions}
\label{bw-rewrite-ArgFrame}
Finally, the remaining arguments~|es| on the stack are evaluated one by one. When all argument expressions have been evaluated, the function is executed.

\begin{code}
rewrite          (State  heap  envs  ( stack :|: ArgFrame a_f as es           )  ( BwResult a            )) = case es of
  (e:es')    ->  (state  heap  envs  ( stack :|: ArgFrame a_f (as ++ [a]) es' )  ( FwExpr e              ))
  otherwise  ->  (state  heap  envs  ( stack                                  )  ( FwApp a_f (as ++ [a]) ))
\end{code}


%format cs = "\overline{c}"
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsection*{Type checking an annotated function}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The instance-of relation is not very informative for functions, since all functions are instances of the class~|function|. However, we can describe a function more precisely by the types of arguments it accepts. 

Function annotations describe the type of a function's arguments, which are checked right before execution of the function. For example the primitive addition function (see Rule~\ref{fw-rewrite-Prim-int.__add__}) is annotated (see Section~\ref{section:integer-class}) to accept only instances of |int| as arguments.

Annotations are only introduced by rewrite rules and object definitions in these semantics. There is no annotation in the abstract syntax.


\rulesection{Type checking function arguments}
\label{fw-rewrite-App-Typed}
An annotated function~|Typed a_f cs| consists of a function address~|a_f| and a list of classes~|cs|. We execute the annotated function call with arguments~|as|, if all arguments~|as| are instances of classes~|cs|. There may be fewer classes in |cs| than there are arguments, but not more. If the arguments fail to satisfy the type, we raise an exception (see Chapter~\ref{chapter:exceptions}).

%format areInstsOf as cs heap = "\forall_{0<i<|"cs"|}~" as"_i :_{"heap"}" cs"_i"
\begin{code}
rewrite                                                 (State  heap  envs  stack  ( FwApp (Typed a_f cs) as ))
  | length as >= length cs && areInstsOf as cs heap  =  (state  heap  envs  stack  ( FwApp a_f as            ))
  | otherwise                                        =  (state  heap  envs  stack  ( raise a_TypeError []    ))
\end{code}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsection*{Function closure execution} \label{section:closure-execution}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The execution of a fully evaluated call~|FwApp a_lambda as| has four (successful) cases: |a_lambda| is a function closure, |a_lambda| is a generator function, |a_lambda| is a callable object, or |a_lambda| is a primitive function. The latter case is not handled in this chapter. Instead, every primitive function has a specific rewrite rule. See for example the primitive addition function which is handled in Rule~\ref{fw-rewrite-Prim-int.__add__}.

%Apart from the user defined functions there are primitive functions and callable objects.  Callable objects are objects with a |__call__| attribute referencing to a function, primitive, or callable object. Thus, there are three cases when executing a function call:


\rulesection{Execution of a function}
\label{fw-rewrite-FwApp}
The execution of a function call has four cases:
\begin{enumerate}
\item 
If |a_f| is a instance of~|function| then the value of |a_f| is the closure~|FuncValue xs b_lambda envs_lambda|. When the number of arguments~|as| equals the number of parameters~|xs|, and |a_f| is not a generator function, we execute the function body~|b_lambda| with the environment~|envs_lambda :||: env_1|. The current environment~|envs| is pushed on the stack in the return marker~|ReturnMarker envs|.

\item 
If |a_f| is a closure with the correct number of arguments, and |a_f| is a generator function, we return a generator~|GenValue (sSingleton (BlockFrame b_lambda)) (envs_lambda :||: env_1)| (see Section~\ref{section:generators}). The generator's environment is the same as a function's environment would be, and the only element on its stack is the function body.

\item 
If |a_f| is not a closure, but an object with a class attribute~|__call__|, we forward the call to that attribute. The function is instantiated with the object |a_f|.

\item 
Otherwise, the call had an incorrect number of arguments or |a_f| was neither a function nor a callable object, so we raise a |TypeError|.
\end{enumerate}


%{
%format a_call
%format a_g
%format a_e
%format (FuncEnvValue (o) (f)) = (o,f)
%format (zip xs as) = [xs_1 |-> as_1, "\cdots", xs_n |-> as_n]
%format as_n
%format xs_n
%format ++ = "\cup"
%format isInstOf a_f a_function heap = a_f ":_{\Theta}" a_function
%format a_ff = a_f
\begin{code}
rewrite                                                                                             (State  heap    ( envs                  )  ( stack                       )  ( FwApp (Addr a_ff) as         ))
    | isInstOf a_f a_function heap        && length as == length xs  && hasNotYield b_lambda     =  (state  heap'   ( envs_lambda :|: env_1 )  ( stack :|: ReturnMarker envs )  ( FwBlock b_lambda             ))
    | isInstOf a_f a_function heap        && length as == length xs  && hasYield b_lambda        =  (state  heap''  ( envs                  )  ( stack                       )  ( BwResult a_g                 ))
    | not (isInstOf a_f a_function heap)  && __call__ `isin` getInhAttrs heap a_f                =  (state  heap    ( envs                  )  ( stack                       )  ( FwApp (Inst a_f a_call) (as) ))
    | otherwise                                                                                  =  (state  heap    ( envs                  )  ( stack                       )  ( raise a_e []                 ))
  where  env_1   = freeAddr heap
         heap'   = heap   <+> [ env_1  |-> FuncEnvValue (new (zip xs as)) (localVars b_lambda ++ xs) ]
         a_g     = freeAddr heap'
         heap''  = heap'  <+> [ a_g    |-> GenValue (sSingleton (BlockFrame b_lambda)) (envs_lambda :|: env_1) ]
         a_call  = get (getInhAttrs heap a_f) __call__
         a_e     = a_TypeError
         FuncValue xs b_lambda envs_lambda = getFuncValue heap a_f
\end{code}
%if style == newcode
\begin{code}
         a_f     = Addr a_ff
\end{code}
%endif

The function's local environment~|env_1| has two parts: the parameter-argument bindings~|new (zip xs as)| and the set of local variables~|localVars b_lambda ++ xs|.
%}

The local variables of a block are the variables used as a target in body of the function, i.e., the variable that can be bound in the local environment when the body is executed. Variable expressions and delete statements have different semantics for local variables (see Chapter~\ref{chapter:variables}).

% glueing rw functions
%if style == newcode
\begin{code}
rewrite s = rwpart1 s

emptyset = []
\end{code}
%endif

\subsubsection{Local variables}
\label{section:functions-local-variables}
The set of local variables~|localVars b| consists of all the variables that can be (un)bound during the execution a block~|b|. Apart from the assign and delete statements, the for and try statements can also bind variables. There is no expression that binds variables. In that respect they are truly stateless, in contrast to statements.

%{
%format emptyset = "\emptyset"
%format : = "\cup"
%format ++ = "\cup"
%format b_x
%format b_y
\begin{code}
localVars EmptyBlock                            = emptyset
localVars (ConsBlock (AssStmt x e)          b)  = x  :  localVars b
localVars (ConsBlock (DelStmt x)            b)  = x  :  localVars b
localVars (ConsBlock (WhileStmt e b_x)      b)  =       localVars b  ++  localVars b_x
localVars (ConsBlock (ForStmt x e b_x)      b)  = x  :  localVars b  ++  localVars b_x
localVars (ConsBlock (IfStmt e b_x b_y)     b)  =       localVars b  ++  localVars b_x  ++  localVars b_y
localVars (ConsBlock (TryStmt b_x e x b_y)  b)  = x  :  localVars b  ++  localVars b_x  ++  localVars b_y
localVars (ConsBlock (FinallyStmt b_x b_y)  b)  =       localVars b  ++  localVars b_x  ++  localVars b_y
localVars (ConsBlock s                      b)  =       localVars b
\end{code}
%}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsection*{Returning from a function}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
A function stops execution at the end of its block or when a return statement is encountered. When a function returns, it returns a value and the environment stored on the stack is reinstated. Functions always return a value; those without a return statement return |None|.

\rulesection{Returning from the end of a function}
\label{bw-rewrite-ReturnMarker}
When all statements in |b_lambda| have been executed, but no return statement has been executed, we will encounter the return marker~|ReturnMarker envs|. The return marker contains the old environment~|envs| which is reinstated. Since there was no return statement, we return |None|.

\begin{code}
rwpart1  (State  heap  envs_lambda   ( stack :|: ReturnMarker envs )  ( BwResult a      )) =
         (state  heap  envs          ( stack                       )  ( BwResult a_None ))
\end{code}


\rulesection{Executing a return statement}
\label{fw-rewrite-RetStmt}
At a return statement, the returned expression~|e| is evaluated and a return frame~|ReturnFrame| is placed on the stack.

\begin{code}
rwpart1  (State  heap  envs  ( stack                 )  ( FwStmt (RetStmt e) )) =
         (state  heap  envs  ( stack :|: ReturnFrame )  ( FwExpr e           ))
\end{code}


\rulesection{Start stack unwinding}
\label{bw-rewrite-ReturnFrame}
Next, once the returned value~|e| has been evaluated to |a|, we use the return instruction to unwind the stack and return |a|

\begin{code}
rwpart1  (State  heap  envs  ( stack :|: ReturnFrame )  ( BwResult a )) =
         (state  heap  envs  ( stack                 )  ( Return a   ))
\end{code}


\rulesection{Unwinding the stack}
\label{rw-rewrite-Return-frame}
Finally the return instruction unwinds (pops frames from) the stack until the return frame is found on top of the stack. At the return frame the old environment is reinstated, |a| is returned, and unwinding stops. 

\begin{code}
rwpart1                   (State  heap  envs   ( stack :|: frame                 )  ( Return a   )) = case frame of
  ReturnMarker envs'  ->  (state  heap  envs'  ( stack                           )  ( BwResult a ))
  FinallyMarker b     ->  (state  heap  envs   ( stack :|: InstrFrame (Return a) )  ( FwBlock b  ))
  otherwise           ->  (state  heap  envs   ( stack                           )  ( Return a   ))
\end{code}

If the return statement is in the body of a try-finally statement, we execute the finally clause |FinallyMarker b| before returning from the function.

%Technically it is possible to jump directly to the return marker when a return statement is encountered. Stack unwinding however, makes the interaction with the try-finally statement is simpler to specify.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Executing generators} \label{section:generators}
%format stack_g
%format envs_g
%format a_g
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Generators produce values on demand. The primitive class function~|__next__| starts the execution of a generator and returns a value when evaluating a yield expression. At the yield expression execution is suspended. Calling |__next__| again, will cause the generator to continue execution until the next yield statement returns a new value. If no more yield statements are encountered |__next__| will raise an exception when it is called.

In the previous sections we have already seen how a generator function is defined and executed. In this section we describe how generators are executed.

The execution of a generator function returned a generator~|GenValue stack envs| with a stack~|stack| and environment~|envs|. When the generator itself is executed, the stack~|stack| and environment~|envs| become the stack and environment of the abstract machine. The generator is suspended by storing the stack and environment in the generator value, and reinstating the old stack and environment.

The following rules describe how the generator is (re)started using the primitive functions |__next__| or |__send__|, and how it is suspended by the yield expression.


\rulesection{Getting the next yielded value}
\label{fw-rewrite-Prim-generator.__next__}
The primitive function~|__next__| starts or resumes a generator if it hasn't already started.

If |__next__| is called for generator that has already started, i.e., it has a value of |SGenValue|, we simply raise a |ValueError|. A running generator cannot be (re)started.

If |__next__| is called for a suspended or new generator~|GenValue stack_g envs_g| the stack~|stack_g| is pushed on the machine stack and the environment is set to |envs_g|. A yield marker~|GenMarker envs a_self| with the current environment~|envs| and the generator's address~|a_self| is pushed on the stack, before |stack_g| the generator's stack.

\begin{code}
rwpart1                         (State  heap   envs    ( stack                                                    )  ( FwApp (Prim "generator" "__next__") [a_self] )) = case v of
   GenValue stack_g envs_g  ->  (state  heap'  envs_g  ( stack `sConcat` GenMarker envs a_self `sPrepend` stack_g )  ( BwResult a_None                              ))
   SGenValue                ->  (state  heap   envs    ( stack                                                    )  ( raise a_ValueError []                        ))
  where  heap'  = heap <+> [a_self |-> SGenValue]
         v      = get heap a_self
\end{code}


\rulesection{Getting the next value with yield value}
\label{fw-rewrite-Prim-generator.__send__}
The primitive~|__send__| resumes a generator like |__next__|, but there is one difference: calling |__send__| with argument~|a| causes the yield expression where the generator has suspended, to evaluate to~|a|.

\begin{code}
rwpart1                         (State  heap   envs    ( stack                                                    )  ( FwApp (Prim "generator" "__send__") [a_self, a] )) = case v of
   GenValue stack_g envs_g  ->  (state  heap'  envs_g  ( stack `sConcat` GenMarker envs a_self `sPrepend` stack_g )  ( BwResult a                                      ))
   SGenValue                ->  (state  heap   envs    ( stack                                                    )  ( raise a_ValueError []                           ))
  where  heap'  = heap <+> [a_self |-> SGenValue]
         v      = get heap a_self
\end{code}


\subsection*{The yield expression}

\rulesection{Evaluating a yield expression}
\label{fw-rewrite-YieldExpr}
First, the value~|e| to be returned is evaluated.

\begin{code}
rwpart1  (State  heap  envs  ( stack                )  ( FwExpr (YieldExpr e) )) =
         (state  heap  envs  ( stack :|: YieldFrame )  ( FwExpr e             ))
\end{code}


\rulesection{Start unwinding of the stack}
\label{bw-rewrite-YieldFrame}
Then, with the yielded value~|a|, we start to unwind the stack. 

\begin{code}
rwpart1  (State  heap  envs  ( stack :|: YieldFrame )  ( BwResult a         )) =
         (state  heap  envs  ( stack                )  ( Yield a EmptyStack ))
\end{code}

The second part of the yield instruction~|Yield a EmptyStack| stores the part of the stack that has been unwound. Unwinding starts with an empty stack (|EmptyStack|).


\rulesection{Unwinding the stack}
\label{bw-rewrite-Yield-frame}
Unwinding of the stack continues until a yield marker~|GenMarker env a_g| is on top. At that point, the generator~|a_g| is updated with the unwound stack~|stack_g| and the generator's environment~|envs_g|. Thus the state of the generator at the yield statement is stored.

\begin{code}
rwpart1                  (State  ( heap                                        )  envs_g  ( stack :|: frame )  ( Yield a stack_g                    )) = case frame of
  GenMarker envs a_g  -> (state  ( heap <+> [a_g |-> GenValue stack_g envs_g ] )  envs    ( stack           )  ( BwResult a                         ))
  otherwise           -> (state  ( heap                                        )  envs_g  ( stack           )  ( Yield a (frame `sPrepend` stack_g) ))
\end{code}


\subsection*{Stopping a generator}
A generator stops when it reaches the end of its body, or an exception is raised in its body (see Rule \ref{raise-rewrite-GenValue}). Stopped generators cannot be restarted, they will raise a |StopIteration| exception when calling |__next__|.


\rulesection{Stopping at the end of a generator}
\label{bw-rewrite-GenMarker}
When the yield marker~|GenMarker envs a_g| is on top of the stack, the generator's body has been executed. We stop the generator, raise a |StopIteration| and reinstate the old environment. 

\begin{code}
rwpart1  (State  ( heap                                              )  envs_g  ( stack :|: GenMarker envs a_g )  ( BwResult a               )) =
         (state  ( heap <+> [a_g |-> GenValue EmptyStack EmptyStack] )  envs    ( stack                        )  ( raise a_StopIteration [] ))
\end{code}

The generator stored on the heap at address |a_g| is updated to the empty generator~|GenValue EmptyStack EmptyStack|. This will cause subsequent executions of the generator to immediately stop.


%if style==newcode
\begin{code}
rwpart1 s = rwpart2 s
\end{code}
%endif

%else
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The function and generator}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%if style==newcode
\begin{code}
o_generator :: Value
o_function :: Value
\end{code}
%endif
%format o_function
%format o_generator

\begin{code}
o_generator = new  
  [ __class__  |-> a_Type
  , __bases__  |-> a_objectT
  , __next__   |-> Prim "generator" "__next__"
  , __send__   |-> Prim "generator" "__send__"
  , __iter__   |-> Prim "generator" "__iter__"
  ]
\end{code}

\begin{code}
o_function = new  
  [ __class__  |-> a_Type
  , __bases__  |-> a_objectT
  , __call__   |-> Prim "function" "__call__"
  ]
\end{code}

\rulesection{A function as a callable object}
\begin{code}
rwpart2  (State  heap  envs  stack  ( FwApp (Prim "function" "__call__") [a] )) =
         (state  heap  envs  stack  ( BwResult a                             ))
\end{code}


\rulesection{The generator as an iterable}
\begin{code}
rwpart2  (State  heap  env  stack  ( FwApp (Prim "generator" "__iter__") [a_self] )) =
         (state  heap  env  stack  ( BwResult a_self                              ))
\end{code}

%if style==newcode
\begin{code}
rwpart2 _ = skipState
\end{code}
%endif


%endif

%}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%if style==newcode
\begin{code}

areInstsOf as cs heap =
  null cs || (not (null as) || error "this shouldn't happen")
          && isInstOf (head as) (head cs) heap 
          && areInstsOf (tail as) (tail cs) heap

noArgs = []

fstArgFrame a = ArgFrame a []

hasNotYield = not . hasYield

hasYield :: Block -> Bool
hasYield EmptyBlock       = False
hasYield (ConsBlock s b)  = stmtHasYield s || hasYield b

stmtHasYield :: Stmt -> Bool
stmtHasYield (ExprStmt e)               = exprHasYield e
stmtHasYield (AssStmt x e)              = exprHasYield e
stmtHasYield (AttrAssStmt e_a x e_b )   = exprHasYield e_a || exprHasYield e_b
stmtHasYield (SliceAssStmt e_a e_b e_c )= exprHasYield e_a || exprHasYield e_b || exprHasYield e_c
stmtHasYield (SliceDelStmt e_a e_b)     = exprHasYield e_a || exprHasYield e_b
stmtHasYield (AttrDelStmt e x)          = exprHasYield e
stmtHasYield (PrintStmt e)              = exprHasYield e
stmtHasYield (IfStmt e b_x b_y)         = exprHasYield e || hasYield b_x || hasYield b_y
stmtHasYield (WhileStmt e b_x)          = exprHasYield e || hasYield b_x
stmtHasYield (ForStmt x e b_x)          = exprHasYield e || hasYield b_x
stmtHasYield (TryStmt b_x e x b_y)      = exprHasYield e || hasYield b_x || hasYield b_y
stmtHasYield (FinallyStmt b_x b_y)      = hasYield b_x || hasYield b_y
stmtHasYield (RaiseStmt e)              = exprHasYield e
stmtHasYield (ExecStmt e_a e_b)         = exprHasYield e_a || exprHasYield e_b
stmtHasYield _                          = False

exprHasYield :: Expr -> Bool
exprHasYield (CallExpr e es) = exprHasYield e || any exprHasYield es
exprHasYield (AttrExpr e x) = exprHasYield e
exprHasYield (ListExpr es) = any exprHasYield es
exprHasYield (TupleExpr es) = any exprHasYield es
exprHasYield (DictExpr []) = False
exprHasYield (DictExpr ((e_a,e_b):ts)) = exprHasYield e_a || exprHasYield e_b || exprHasYield (DictExpr ts)
exprHasYield (SliceExpr e_a e_b) = exprHasYield e_a || exprHasYield e_b
exprHasYield (BinOpExpr o e_a e_b) = exprHasYield e_a || exprHasYield e_b
exprHasYield (UnaryOpExpr o e) = exprHasYield e
exprHasYield (YieldExpr e) = True
exprHasYield _ = False
\end{code}
%endif


