%if style == newcode
\begin{code}
module Import where
import Control.Monad.State(liftIO)
import System.FilePath(dropFileName,pathSeparator)
import System.IO.Unsafe(unsafePerformIO)
--import Data.List(intercalate)
import Directory(doesFileExist)

import VirtualMachine
import ObjectTheory

import PythonParser
\end{code}
%endif


%{
%format a_m
%format x_m
%format env_p
%format ms = "\overline{"m"}"
%format ys = "\overline{"y"}"
%format a_ms
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The import statement} \label{section:import}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Import statements initialize modules or packages, and include them in the current environment as module objects.

%format py = "\mathtt{py}"
Packages are directories containing an initialization module~|__init__.py| and usually a number of other modules. Modules are normal Python files that are implemented by executing them. The variables that have been bound after the initialization of a module, define the attributes of the new module object. Similarly, the initialized modules and packages in a package define the attributes of its module object.

Modules and packages are initialized only once. The first time a module or package is imported, it is stored in a dictionary value at |a_ms|. If a module is imported a second time, it will be retrieved from the dictionary |a_ms|. This mechanism also prevents the built-in modules from being redefined.

%format (last xs) = x_n
%format (firsts xs) = x"_{1 \cdots n-1}"
An import statement~|ImportStmt xs| imports a or package~|xs|. The dot separated list of names~|xs| describes the path of a module or package, where the first names~|firsts xs| are packages and the last name~|last xs| is usually a module but can be a package. 

%format x_1
%format x_2
When executing an import statement |ImportStmt xs| with |length xs > 1|, each package in the path is initialized before the full path is imported. It is as if we were to execute |ImportStmt x_1|, |ImportStmt x_1.x_2|, and so forth.


%{
%format ImportFrame env xs ys = env "\vdash \mathtt{import}~"ys
\rulesection{Executing an import statement}
\label{fw-rewrite-ImportStmt}
The import statement is executed by placing the import frame~|ImportFrame envs [] xs| on the stack. The import frame contains the environment~|envs| that is reinstated when the execution of the import statement has completed (or fails with an exception), and the path~|xs| that will be imported.

\begin{code}
rewrite  (State  heap  envs  ( stack                            )  ( FwStmt (ImportStmt xs) )) =
         (state  heap  envs  ( stack :|: ImportFrame envs [] xs )  ( BwResult a_None        ))
\end{code}
%}

\noindent
This first rule is always immediately followed by Rule~\ref{bw-rewrite-ImportFrame}.


%format ps = "/"
%format a_IE
%format a_SE
%format IE = "\mathtt{IE}"
%format SE = "\mathtt{SE}"
\rulesection{Importing modules or packages}
\label{bw-rewrite-ImportFrame}
The import frame~|ImportFrame envs xs ys| indicates that we have just imported the first part~|xs| of the import statement's path, and have to import the remaining part~|ys|. The topmost environment~|env_p| is the environment of the module that has just been imported. The address |a|, which is immediately discarded, is returned by the last statement of the previously imported module.

The environment~|env_p|, in which the module |xs| was executed, defines the module's attributes. This is not unlike the class definition statements (see Rule \ref{fw-rewrite-ClassStmt}).

\begin{enumerate}
\item If a module or package~|x_m| was already imported, its address is in list of previously imported modules~|ms|. In this case we bind the module or package in the environment~|env_p|.
\item If |x_m| is a package (there is a directory by that name with an initialization module in it), we create a new environment~|a_m| in which we execute the parsed initialization module. The environment~|a_m| is bound to |ys| in the module~|env_p| and stored in the list of modules~|ms|.
\item If we are importing the last element of the path and it is a module, we execute the parsed module in the environment~|a_m|. The environment~|a_m| is bound to |ys| in the module~|env_p| and stored in the list of modules~|ms|.
\item Otherwise the import is invalid because a module or package could not be found, so we raise an |ImportError|.
\end{enumerate}

%{
%format x_1
%format x_n
%format / = "\mathtt{/}"
%format mkModuleName (xs) (y) = x_1/"\cdots"/x_n/y
%format mkModuleFile pwd (xs) (y) = x_1/"\cdots"/x_n/y."\mathtt{py}"
%format mkPackageFile pwd (xs) (y) = x_1/"\cdots"/x_n/y/__init__."\mathtt{py}"
%format isPackage pwd xs y x = x"\text{ is package}"
%format isModule pwd xs y x = x"\text{ is module}"
%format py = "\mathtt{py}"
%format parseModule pwd xs y x = "\text{parse }" x.py
%format parsePackage pwd xs y p = "\text{parse }" p
%format mkPackageName x = x"\mathtt{/}"__init__.py
\begin{code}
rewrite                                                                                   (State  heap    ( envs' :|: env_p          )  ( stack :|: ImportFrame envs  xs (y:ys) )  ( BwResult a                                      ))
   | x `isin` ms                                                                       =  (state  heap'   ( sSingleton (get ms x)    )  ( stack :|: frame                       )  ( BwResult a_None                                 ))
   | x `isnotin` ms  && isPackage (getStringValue heap a_pwd) xs y x                   =  (state  heap''  ( sSingleton env_0 :|: a_m )  ( stack :|: frame                       )  ( parsePackage (getStringValue heap a_pwd) xs y p ))
   | x `isnotin` ms  && isModule (getStringValue heap a_pwd) xs y x && length ys == 0  =  (state  heap''  ( sSingleton env_0 :|: a_m )  ( stack :|: frame                       )  ( parseModule (getStringValue heap a_pwd) xs y x  ))
   | otherwise                                                                         =  (state  heap    ( envs                     )  ( stack                                 )  ( raise a_IE []                                   ))
  where  a_m     = freeAddr heap
         x       = mkModuleName xs y
         p       = mkPackageName x
         ms      = get heap a_ms
         heap'   = heap <+>  [ env_p  |-> get heap env_p <+> [y |-> get ms x] ]
         heap''  = heap <+>  [ env_p  |-> get heap env_p <+> [y |-> a_m] 
                             , a_ms   |-> ms <+> [x |-> a_m]
                             , a_m    |-> ModValue empty
                             ]
         a_IE    = a_ImportError
         frame   = ImportFrame envs  (xs ++ [y])  ys
\end{code}

Like the parser used for the exec statement (see Section~\ref{section:exec}), this parser returns a block when it succeeds and raises a |SyntaxError| exception when it fails. Note that we're parsing files here, not strings.

This rule applies repeatedly until either an exception is raised, or all the packages and the module have been imported. It is then followed by Rule~\ref{bw-rewrite-ImportFrame2}.


\rulesection{Returning from the import statement}
\label{bw-rewrite-ImportFrame2}
When all packages and the final module have been imported, we reinstate the environment~|env| and return |None|.

\begin{code}
rewrite  (State  heap    ( envs' :|: env_p )  ( stack :|: ImportFrame envs xs [] )  ( BwResult a      )) =
         (state  heap    ( envs            )  ( stack                            )  ( BwResult a_None ))
\end{code}


\rulesection{Raising an exception in an imported module}
\label{raise-rewrite-ImportFrame}
If an imported module raises an exception, the previous environment |envs'| is reinstated and unwinding continues (see Chapter~\ref{chapter:exceptions}).

%format ys = "\overline{y}"
\begin{code}
rewrite  (State  heap  envs  ( stack :|: ImportFrame envs' xs ys )  ( Raise a )) =
         (state  heap  envs' ( stack                             )  ( Raise a ))
\end{code}
%}




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%if style==newcode
\begin{code}
rewrite _ = skipState
\end{code}
%endif
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%}

%if style==newcode
\begin{code}
mkModuleName xs y = intercalate "." (xs ++ [y])
mkModuleFile pwd xs y = dropFileName pwd ++ intercalate ps (xs ++ [y]) ++ ".py"
mkPackageFile pwd xs y = dropFileName pwd ++ intercalate ps (xs ++ [y]) ++ ps ++ "__init__.py"

isModule pwd xs y name = not (fileExists p_p) && fileExists p_m
  where p_p = mkPackageFile pwd xs y
        p_m = mkModuleFile pwd xs y
isPackage pwd xs y name = fileExists p_p
  where p_p = mkPackageFile pwd xs y

ps = [pathSeparator]

fileExists = unsafePerformIO . doesFileExist

parseModule pwd xs y _ = parse (mkModuleFile pwd xs y)
parsePackage pwd xs y _ = parse (mkPackageFile pwd xs y)
mkPackageName = id

parse file = 
  case unsafePerformIO (strictParseFile pBlock file) of
    Just b -> FwBlock b
    Nothing -> raise a_SyntaxError []

moduleClass :: Value
moduleClass = new
  [ __class__  |-> a_Type
  , __bases__  |-> a_objectT
  ]
\end{code}
%endif
