%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Introduction}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
This introduction is organized as follows. First we review Python with respect to formal semantics, and define a scope for our work. Then we discuss executable operational semantics and our approach to it. Finally, we list the main contributions, and give an overview of the rest of this thesis.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Python}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Python is an imperative, dynamic, object-oriented programming language originally developed by Guido van Rossum at the CWI in the Netherlands in the 1980s. Since the first publication of the code in 1991, Python has gained much popularity and now is a major programming platform.

%The language is still evolving and many different implementations exist~\cite{ironpython,jython,pypy}. Van Rossum still plays a leading role in the development of the language and the original implementation CPython~\cite{cpython}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsection*{A formal semantics for Python}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
There are no formal semantics for Python. CPython is the de facto reference implementation. Although it maintains high coding standards, CPython is not written with legibility as its primary focus.

The PyPy~\cite{pypy} implementation of Python is written in Python itself. To be exact, it is written in a restricted form of Python called RPython~\cite{ancona2007rst}. The introduced restrictions facilitate static analyses such as typing, and enable good run-time efficiency. 

Because it is easy to read for Python programmers, it has been suggested that PyPy might one day become the reference implementation of Python. However, even for the restricted form of Python, no systematic documentation of the operational semantics exists.

Despite the lack of a formal semantics, Python appears to be a particularly good candidate for formal specification, for a number of reasons.

\begin{itemize}
\item Python is a relatively simple language. It has a limited number of constructs, most of which are found in many other languages. Python introduces no radically new features, but has a unique combination of features found in its predecessors.

\item Despite being a simple language, it is a widely used language which is still gaining in popularity. As a language and platform, Python has proven to be mature: many sizable projects use it to great success.

\item The language is quite well documented. There is an extensive reference manual~\cite{pythonman} and all past changes to the language have been documented in Python enhancement proposals (PEPs)~\cite{pep0}.

\item There are a number of independent, mature implementations of Python available \cite{cpython,ironpython,jython,pypy}. Semantics can be used to compare interpreters and maintain language conformism.
\end{itemize}

These semantics target Python version 2.5. The behaviour of the CPython implementation, not the documentation, is taken as the definition of Python. 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsection*{Scope of the semantics} \label{section:scope}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Of course we would have liked to document all of Python, but time constraints did not allow that. Therefore the scope of these semantics is limited to a subset of Python that we call \minpy. The language \minpy\ includes virtually all syntactical constructs of Python, but leaves out much of the syntactic sugar (see Chapter~\ref{chapter:AST} for an overview of the abstract syntax).

While most of the concepts introduced by the syntax are included in these semantics, much of the language's features have been ignored:

\begin{itemize}
\item The standard library has not been specified wherever possible. Some elements of the library are needed to support other features. For example, exception handling naturally requires the Exception class.

\item
Garbage collection is not modeled by these semantics. Apart from managing the storage of objects in memory, garbage collection affects the operational semantics by calling an object's finalizer, which can be specified by a programmer. Typically different implementations have their own garbage collectors and some even allow a user to control it.

\item
Python's support for multi-threading is ignored. While multi-threading is an interesting subject by itself, it is simply beyond the scope of this project. It could also be argued that threads are not truly part of the language, since they are only supported through the interpreter's libraries.

\item
The ability of Python to interact with the `outside world' is not described. In practice, the `outside world' is represented by other libraries. Interaction between two different languages is achieved through what is sometimes called a foreign function interface.

\item
The specification of simple types in this document, such as integers and strings, do not describe the exact behaviour of CPython (see Chapter~\ref{chapter:builtins}).
The semantics of simple types tend to differ in the small details. Types and their associated operators usually have the semantics of the platform underpinning the interpreter. For example, Jython uses Java's integers while CPython's integers are based on C's integer types.

\item
These semantics only work with the so called new style classes that have been introduced in CPython version 2.2 \cite{newstyleclasses}. The old-style classes had been kept for backwards compatibility only, and will not be in future versions of Python.

\item
Python has various reflexive features. For example, it is possible to inspect the dictionary that implements objects, and the body of a function can be replaced at runtime. These features are not covered by our semantics.
\end{itemize}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Formal semantics}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Formal semantics seek to precisely describe the meaning of a programming language using mathematical constructs. All aspects of programming languages can be formally specified, but usually a formal semantics describes the run-time behaviour of programs.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsection*{Operational semantics}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
A number of different styles of formal semantics exist. Operational semantics as introduced by Plotkin~\cite{plotkin1981sao}, describe language semantics in terms of a state transition system. Using an abstract machine for the state, the operational semantics will closely resemble an interpreter. Because a state machine uses well known constructs like stacks and heaps, the semantics will be comparatively easy to understand.

In general, formal semantics improve our understanding of a language to its finest details, simplifying reasoning about programs at an abstract level. Operational semantics are of significant practical value as well:

\begin{itemize}
\item
The detailed documentation provided by an operational semantics facilitate the creation of interpreters of compilers. Especially if accompanied by a test suite, standardization is much simpler to achieve and maintain.

\item
Program analyses can be designed in a more systematic fashion and formal reasoning can be used to verify properties of analyses. More generally, many language tools are easier to develop with an operational semantics at hand.

\item
A language can be extended and altered more safely, since interaction of extensions with the original language can be analyzed systematically. For example backwards compatibility can be guarded more closely.
\end{itemize}

While the case for operational semantics is easy to make, some disadvantages of formal semantics in general also apply to operational semantics, albeit to a lesser degree.

\begin{itemize}
\item
Formal semantics are often perceived to be an ivory tower matter, i.e., be of little value to neither the users nor the developers of a language. The following anecdote illustrates this nicely. When inquiring after previous work on formal semantics on the Python mailing list, someone answered ``[\ldots]~I don't think Python culture operates like that very much.''

\item
Formal semantics are often defined separately from the actual implementation. Whether or not a compiler or interpreter actually adheres to the defined semantics is unclear and difficult to judge. Very few compilers or interpreters have been proven correct with respect to the semantics.

\item
It has been argued that formal semantics might inhibit evolution of a language~\cite{hudak2007hhb}; the maintenance costs of formal semantics could discourage experimentation and extension.
\end{itemize}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsection*{An executable operational semantics}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The listed shortcomings of formal semantics can be eliminated by making the operational semantics executable. In other words, by writing the semantics in such a way, that it can be compiled into an interpreter.

An executable operational semantics would be of immediate practical value as an interpreter. Furthermore it simplifies maintenance of the operational semantics, as the semantics and implementation are developed simultaneously. An executable operational semantics enables experimentation and simplifies language evolution.

The operational semantics in this document have been written in the programming language Haskell, or more specifically literate Haskell, a variant that allows Haskell code to be mixed with \LaTeX. The Haskell code has been formatted using lhs2\TeX\cite{lhs2TeX} and some simple scripts. Thus the sources used to produce this document, can also be compiled into an interpreter for \minpy.

The source code has been reformatted in such a way, that no knowledge of Haskell is required. One might even forget that the semantics were written in a programming language at all, as the rules look like common mathematical equations.

Normally writing the semantics of a complete language is a tedious and error prone process. Writing an executable operational semantics on the other hand is exciting, as one's progress is clearly reflected in the interpreter. The strong type system of Haskell prevents many kinds of mistakes, and we can compare the behaviour of our executable semantics to that of Python with test cases. While testing cannot prove the correctness of the semantics, it does justify confidence in the semantics' quality.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Contributions}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The main contributions of this master's thesis are the following:

\begin{enumerate}
\item
Firstly, we have described the operational semantics of a significant subset of Python;
\item
secondly, we have created an interpreter from the same sources as the operational semantics;
\item
and finally, we have created a test suite that compares the behaviour as described by our operational semantics to the CPython implementation, or other implementations.
\end{enumerate}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Overview}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
This thesis is organised as follows. First, we set the stage for the semantics by introducing some notational conventions in Chapter~\ref{chapter:preliminaries}, describing the abstract syntax of \minpy in Chapter~\ref{chapter:AST}, and introducing the abstract machine model in Chapter~\ref{chapter:abstract-machine}. Then, in Chapter~\ref{chapter:object-theory} we introduce the object model of \minpy.

The remaining chapters, except for the conclusion, contain the semantic rules that describe the behaviour of specific statements and expressions. We start with some basic rules in Chapter~\ref{chapter:basic-rules}, followed by the rules describing variables in Chapter~\ref{chapter:variables}, functions and generators in Chapter~\ref{chapter:functions}, classes and objects in Chapter~\ref{chapter:objects}, exception handling in Chapter~\ref{chapter:exceptions}, control structures in Chapter~\ref{chapter:control-structures}, printing in Chapter~\ref{chapter:print}, operators in Chapter~\ref{chapter:operators}, and finally the exec and import statements in Chapter~\ref{chapter:dynamic-code}.

Finally, in Chapter~\ref{chapter:conclusion}, we discuss the results of this work and the lessons that we learned in the process.

%in Chapter~\ref{chapter:preliminaries}, we introduce some notational conventions used throughout this document. Then, in Chapter~\ref{chapter:AST}, we describe the abstract syntax of \minpy. This is followed by Chapter~\ref{chapter:abstract-machine}, which introduces the abstract machine used as the basis of our semantics. Then, we introduce the object model used by \minpy.

