module Main where

import System(getArgs)
import System.Console.Readline(addHistory, readline, setInhibitCompletion)
import Control.Monad.State(evalStateT, liftIO)

import VirtualMachine
import Tracer
import RewriterMonad
import ObjectTheory

import qualified BasicRewrites
import qualified Functions
import qualified Objects
import qualified ControlStructures
import qualified Variables
import qualified Numbers
import qualified Print
import qualified Exceptions
import qualified Lists
import qualified Import
import qualified Strings
import qualified Dictionaries
import qualified BasicIO
import qualified Exec
import qualified Booleans
import qualified Tuples
import qualified Operators

import PythonParser

eitherRewrite :: Rewriter -> Rewriter -> Rewriter
eitherRewrite x y step = do
  mstep <- x step
  case mstep of
    Just step -> return (Just step)
    Nothing -> y step

combinedRewrite :: Rewriter
combinedRewrite = foldr1 eitherRewrite
  [ BasicRewrites.rewrite
  , Functions.rewrite 
  , Booleans.rewrite
  , Operators.rewrite
  , Objects.rewrite
  , Variables.rewrite
  , ControlStructures.rewrite-- must be before Exceptions.rewrite
  , Numbers.rewrite
  , Print.rewrite
  , Lists.rewrite
  , Tuples.rewrite
  , BasicIO.rewrite
  , Strings.rewrite
  , Dictionaries.rewrite
  , Exec.rewrite-- must be before Exceptions.rewrite
  , Import.rewrite-- must be before Exceptions.rewrite
  , Exceptions.rewrite
  , primCallErr
  ]

--TODO: move primCallErr this somewhere else!? in parts?
-- enable and disable tracing primitives
primCallErr (State  heap  envs  stack  ( FwApp (Prim "sys" "enableTrace") [] )) =
  enableTrace >> (state  heap  envs  stack  ( BwResult a_None ))
primCallErr (State  heap  envs  stack  ( FwApp (Prim "sys" "disableTrace") [] )) =
  disableTrace >> (state  heap  envs  stack  ( BwResult a_None ))
-- raise a type error if we've called a primitive but don't have a matching rule
-- usually this occurs when a prim is called with an incorrect number of arguments
primCallErr (State  heap  envs  stack  ( FwApp (Prim obj func) as )) =
            (state  heap  envs  stack  ( raise a_TypeError []     ))
-- skip over the stack and environment markers used for tracing
primCallErr (State  heap  envs  ( MarkedStack _ stack ) i) =
            combinedRewrite (State  heap  envs  (               stack ) i)
primCallErr (State  heap  ( MarkedStack _ envs ) stack i) =
            combinedRewrite (State  heap  (               envs ) stack i)
primCallErr _ = skipState


-- emptyheap is the start heap, not actually empty
-- before using it, also insert:
-- * the argumenst (a_args), represented by a list of strings
-- * the current working directory (a_pwd), represented by a string
emptyHeap = new
  [ a_None                  |-> NoneValue
  , a_emptyTuple            |-> TupleValue []
  , a_NoneType              |-> new [__class__ |-> a_Type, __bases__ |-> a_objectT]
  , a_True                  |-> BoolValue True
  , a_False                 |-> BoolValue False
  , a_Type                  |-> Objects.o_Type
  , a_object                |-> Objects.o_object
  , a_objectT               |-> TupleValue [a_object]
  , a_bool                  |-> Booleans.bool
  , a_int                   |-> Numbers.o_int
  , a_list                  |-> Lists.o_list
  , a_tuple                 |-> Tuples.o_tuple
  , a_ListIterator          |-> Lists.listIterType
  , a_generator             |-> Functions.o_generator
  , a_BaseException         |-> Exceptions.baseException
  , a_BaseExceptionT        |-> TupleValue [a_BaseException]
  , a_TypeError             |-> Exceptions.typeError
  , a_NameError             |-> Exceptions.nameError
  , a_AttributeError        |-> Exceptions.attributeError
  , a_IndexError            |-> new [__bases__ |-> a_BaseExceptionT, __class__ |-> a_Type]
  , a_SyntaxError           |-> new [__bases__ |-> a_BaseExceptionT, __class__ |-> a_Type]
  , a_StopIteration         |-> new [__bases__ |-> a_BaseExceptionT, __class__ |-> a_Type]
  , a_UnboundLocalError     |-> new [__bases__ |-> a_BaseExceptionT, __class__ |-> a_Type]
  , a_ImportError           |-> new [__bases__ |-> a_BaseExceptionT, __class__ |-> a_Type]
  , a_ValueError            |-> new [__bases__ |-> a_BaseExceptionT, __class__ |-> a_Type]
  , a_ModuleType            |-> new [__class__ |-> a_Type , __bases__  |-> a_objectT]
  , a_str                   |-> Strings.o_string
  , a_dict                  |-> Dictionaries.o_dict
  , a_function              |-> Functions.o_function
  , a_file                  |-> BasicIO.fileClass
  , a_name                  |-> StringValue "__main__"
  , a_ms                    |-> new ["__main__" |-> a_main, "os" |-> a_os, "sys" |-> a_sys]
  , a_os                    |-> ModValue (new ["listdir" |-> Prim "os" "listdir", "path" |-> a_path])
  , a_path                  |-> ModValue (new ["isdir" |-> Prim "path" "isdir"])
  , a_sys                   |-> ModValue (new [ "argv"    |-> a_args
                                              , "stdout"  |-> a_stdout
                                              , "disableTrace" |-> Prim "sys" "disableTrace"
                                              , "enableTrace"  |-> Prim "sys" "enableTrace"
                                              ]
                                         )
  , a_stdout                |-> ModValue (new ["write" |-> Prim "stdout" "write"])
  , a_main                  |-> ModValue empty
  , env_0                   |-> ModValue startEnv
  ] where mkInstance tp = new ["__class__" |-> tp]


startEnv :: Object
startEnv = new
  [ "None"              |-> a_None
  , "bool"              |-> a_bool
  , "True"              |-> a_True
  , "False"             |-> a_False
  , "type"              |-> a_Type
  , "object"            |-> a_object
  , "str"               |-> a_str
  , "int"               |-> a_int
  , "list"              |-> a_list
  , "dict"              |-> a_dict
  , "generator"         |-> a_generator
  , "TypeError"         |-> a_TypeError
  , "BaseException"     |-> a_BaseException
  , "Exception"         |-> a_BaseException --TODO: make a separate class for user defined exceptions, not just an alias
  , "NameError"         |-> a_NameError
  , "file"              |-> a_file
  , "SyntaxError"       |-> a_SyntaxError
  , "AttributeError"    |-> a_AttributeError
  , "StopIteration"     |-> a_StopIteration
  , "IndexError"        |-> a_IndexError
  , "UnboundLocalError" |-> a_UnboundLocalError
  , "ImportError"       |-> a_ImportError
  , "ValueError"        |-> a_ValueError
  ]

runStateMachine :: State -> RewriterM State
runStateMachine startState = do
  mStartState <- markState startState
  mNewState <- combinedRewrite mStartState
  case mNewState of
    Just newstate -> 
      do stepNo <- getStepCount
         trace (printString (showStateDelta stepNo startState newstate))
         runStateMachine newstate
    Nothing       -> return startState

startState :: String -> [String] -> Instr -> State
startState pwd args result = State heap (EmptyStack :|: env_0 :|: a_main) EmptyStack result
  where heap = Lists.storeNewValuesAsList a_args (map StringValue args) 
                 (emptyHeap <+> [a_pwd |-> StringValue pwd]) 

printAddr :: Result -> Stmt
printAddr (Addr a) = PrintStmt (AddrExpr a)

---------------------------------------------------------------------
-- Main
---------------------------------------------------------------------

main :: IO ()
main = do
  args <- getArgs
  let (pargs, options) = getOptions args
  case pargs of
    (file:_)  -> runFile options file pargs
    otherwise -> runCommandline options

type Options = ()

getOptions :: [String] -> ([String], Options)
getOptions args = (args, ())

---------------------------------------------------------------------
-- File interpreter
---------------------------------------------------------------------
runFile :: Options -> String -> [String]-> IO ()
runFile options filename args = do
  program <- parseFile pBlock filename
  let state = startState filename args (FwBlock program)
  state <- evalStateT (runStateMachine state) startMetaState
  case state of
    State heap env stack (BwResult a) -> return ()
    State heap env stack (Raise a)    -> 
      do evalStateT (execStmt (printAddr a) state) startMetaState
         return ()
    otherwise -> 
      error $ "illigal state:\n" ++ showStateDelta 0 undefined state


---------------------------------------------------------------------
-- Commandline interpreter
---------------------------------------------------------------------
runCommandline :: Options -> IO ()
runCommandline options = do
  setInhibitCompletion True
  evalStateT (loop $ Just ("", startState "./" [] $ BwResult a_None)) startMetaState
  return ()
  where loop :: Maybe (String, State) -> RewriterM ()
        loop Nothing = liftIO $ putStrLn("exiting minpy")
        loop (Just (prevCommands, prevState)) = do
          input <- liftIO $ readline (if null prevCommands then ">>> " else "... ")
          case input of
            Just str  -> liftIO (addHistory str)
            otherwise -> return ()
          result <- execCommand input prevCommands prevState
          loop result


execCommand :: Maybe String -> String -> State -> RewriterM (Maybe (String, State))
execCommand Nothing _ _ = return Nothing
execCommand (Just "") "" prevState = return (Just ("", prevState))
execCommand (Just "") prevCommands prevState =
  case strictParseString pBlock prevCommands of
    Just (ConsBlock prog EmptyBlock) -> do newState <- execStmt prog prevState
                                           return (Just ("", newState))
    Just prog -> do newState <- execBlock prog prevState
                    return (Just ("", newState))
    Nothing   -> do liftIO $ parseString pBlock prevCommands --print error messages
                    return (Just ("", prevState))
execCommand (Just command) prevCommands prevState =
  case strictParseString pStmtLine command of
    Just prog -> case strictParseString pCompStmt fullCommand of
                   Just prog -> return (Just (fullCommand, prevState))
                   Nothing   -> execCommand (Just "") fullCommand prevState
    Nothing   -> return (Just (fullCommand, prevState))
  where fullCommand = prevCommands ++ "\n" ++ command


execStmt :: Stmt -> State -> RewriterM State
execStmt prog (State heap envs stack _) =
  execState (State heap envs stack (FwStmt prog))

execBlock :: Block -> State -> RewriterM State
execBlock prog (State heap envs stack _) =
  execState (State heap envs stack (FwBlock prog))

execState :: State -> RewriterM State
execState startState = do
  state <- runStateMachine startState
  case state of
    State heap env stack (BwResult a) -> 
      if a == a_None 
        then return state
        else execStmt (printAddr a) state
    State heap env stack (Raise a)    -> 
      execStmt (printAddr a) state
    otherwise                         -> 
      error $ "illigal state:\n" ++ showStateDelta 0 undefined state
