%if style == newcode
\begin{code}
module ObjectTheory
( getAllAttrs
, getInhAttrs
, unions
, mro
, isInstOf
, isSubclass
, isNotSubclass
, getObj
, isValidClass
, module Attributes
) where

import Data.List(intersect)

import VirtualMachine
import Attributes

getObj :: Heap -> Result -> Object
getAllAttrs :: Heap -> Result -> Object
getInhAttrs :: Heap -> Result -> Object
\end{code}
%endif


%format a_f
%format a_o
%format a_v
%format a_x
%format e_o
%format a_c
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Theory of objects and classes} \label{chapter:object-theory}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Python is an object-oriented language. Values, including built-in values such as integers, strings, and functions, are represented by objects. Objects are mappings of names to addresses. The addresses point to the object's data and the associated operations (functions).

Every object has exactly one class. Even classes, being objects themselves have a class. The class-of relations between object and their classes can be represented by a tree. At the root of such a tree is the class~|Type|, which is its own class. See for example Figure~\ref{example:simple-instance-tree}.

%Every object is an instance of a class. Even classes, being objects themselves, are instances of classes. The instance relations between objects and their classes can be represented by a tree, with at the root the class~|Type|, which is an instance of itself. See for example Figure \ref{example:simple-instance-tree}.

\tikzstyle{class} = [rectangle, draw, text centered, minimum height=2em, minimum width=2em]
\tikzstyle{object} = [circle, draw, text centered, minimum height=2em]
\tikzstyle{inherits} = [draw, -open triangle 60]
\tikzstyle{instance} = [draw, -angle 60]
   

\begin{figure}
\begin{center}
\begin{tikzpicture}[scale=1.5, auto]
    %\draw[help lines] (0,0) grid (3,2);

    % Place nodes
    \node [class] (type)   at (1,2) {|Type|};
    \node [class] (object) at (3,2) {|object|};
    \node [class] (A)      at (0,1) {|A|};
    \node [class] (B)      at (2,1) {|B|};
    \node [object] (b)     at (4,1) {|b|};
    \node [class] (C)      at (1,0) {|C|};
    \node [object] (c)     at (4,0) {|c|};

    % Draw edges
    \path [instance] (object) -- (type);
    \path [instance] (type) edge [loop above] node {} (type);

    \path [instance] (A) -- (type);
    \path [instance] (B) -- (type);
    \path [instance] (C) -- (type);

    \path [instance] (b) -- (B);
    \path [instance] (c) -- (C);
\end{tikzpicture}
\end{center}
\caption{Example tree showing class-of relations. The classes |object|, |A|, |B|, and |C| have class~|Type|, which is its own class. The objects |a| and |b| have classes |A| and |B| respectively.}
\label{example:simple-instance-tree}
\end{figure}

Every class has one or more base classes, with the exception of the object~|object|, which has no bases itself. The bases of a class are sorted by their local precedence order.
%Classes are said to inherit the attributes of its bases. 
%The bases of a class are sorted by their local precedence order, which determines what bases are inherited first.
The bases of a class~|A|, the bases of the bases, and so on, are the superclasses of |A|. The object~|object| is a superclass of all classes. 

The base-of relation between types can be represented by a directed (from class to base), acyclic graph with edges ordered by their local precedence order. In this inheritance graph, all paths eventually lead to the base object |object|. See for example Figure \ref{example:simple-inheritance-graph}.

\begin{figure}
\begin{center}
\begin{tikzpicture}[scale=1.5, auto]
    %\draw[help lines] (0,0) grid (3,2);

    % Place nodes
    \node [class] (type)   at (-1,2) {|Type|};
    \node [class] (object) at (1,2) {|object|};
    \node [class] (A)      at (0,1) {|A|};
    \node [class] (B)      at (2,1) {|B|};
    \node [class] (C)      at (1,0) {|C|};

    % Draw edges
    \path [inherits] (type) -- (object);
    \path [inherits] (A) -- (object);
    \path [inherits] (B) -- (object);
    \path [inherits] (C) edge [] node {1} (A);
    \path [inherits] (C) edge [] node {2} (B);

\end{tikzpicture}
\end{center}
\caption{Example inheritance graph. The classes |Type|, |A|, and |B| inherit from |object|. The class |C| inherits |A| and |B|, in that order.}
\label{example:simple-inheritance-graph}
\end{figure}

%format (getObj (heap) (a)) = "\Omega^{"heap"}_{"a"}"

Both the class and bases are object attributes: the class of an object is stored in the attribute~|__class__|, and the |__bases__| attribute contains a tuple of a class' base classes. The order of bases in the |__bases__| tuple define the local precedence order. Classes are differentiated from other objects, only by their |__bases__| attribute, which `normal' objects lack.

All values~|get heap a| on a heap~|heap| have an object representation~|getObj heap a|. The object representation is the mapping of names to values, i.e. the object's attributes. The value of an object is determined by the class of an object. For example, objects of class |Type| or |object| have a value~|None|, and objects of class |int| have a primitive integer value.

%format a_C
The object representations of built-in types and functions, are not stored on the heap. Built-ins have an immutable object representation |[__class__ ||-> a_C]|, where |C| is the class of the built-in type. Only a module value returns an object representation extended with its own mapping.

%if style == newcode
\begin{code}
getObj heap (Prim _ _) = new [__class__ |-> a_function]
getObj heap (Typed a _) = getObj heap a
getObj heap (Inst _ a) = getObj heap a
getObj heap a = case get heap a of
    ObjValue o _       -> o
    IntValue i         -> new    [__class__ |-> a_int]
    BoolValue b        -> new    [__class__ |-> a_bool]
    ListValue lst      -> new    [__class__ |-> a_list]
    ListItValue _ _    -> new    [__class__ |-> a_ListIterator]
    ModValue o         -> o <+>  [__class__ |-> a_ModuleType]
    StringValue s      -> new    [__class__ |-> a_str]
    DictValue dict     -> new    [__class__ |-> a_dict]
    GenValue _ _       -> new    [__class__ |-> a_generator]
    SGenValue          -> new    [__class__ |-> a_generator]
    TupleValue _       -> new    [__class__ |-> a_tuple]
    NoneValue          -> new    [__class__ |-> a_NoneType]
\end{code}
%endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{The method resolution order}
\label{section:inheritance-graph} 
\label{section:inheritance}
\label{section:MRO}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%The relations between classes and their bases can be represented as a directed graph with ordered edges, the inheritance graph. The ordering of the edges is determined by the local precedence of bases. 

%format MRO = "\text{\sc mro}"

A class inherits the attributes of its superclasses. Objects have access to the attributes of their class, including the inherited attributes, as if they are their own attributes. However, when classes have different definitions for the same attributes, it is unclear which takes precedence. 

To resolve conflicts between inherited attributes, we define a so called method resolution order (|MRO|). The order of classes in the |MRO| determines which attribute overrides the other attributes of the same name. Despite the name, the |MRO| determines not only the resolution of methods (function attributes), but the resolution of all attributes.

%The MRO of a class is a topological sort of the classes reachable from that class in the inheritance graph. Often there are multiple topological sorts for the same inheritance graph.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsubsection{The C3 algoritm}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Python uses the C3 algorithm originally developed for Dylan~\cite{simionatoPythonMRO,barrett1996msl} to find an |MRO| that satisfies a number of requirements. Most importantly, 
\begin{itemize}
\item The |MRO| observes local precedence order. For example, a class~|A| that precedes a class~|B| in the local precedence order of a class~|C|, must precede |B| in the |MRO| of |C|.
\item The |MRO| is monotonic. For example, a class~|A| in the |MRO| of a class~|B|, must precede |B| in any |MRO| that contains |A|.
%\item The |MRO| is consistent with the extended precedence graph as defined in \cite{ducournau1992mcr,ducournau1994pmm}.
\end{itemize}

%format CS = "\overline{C}"
%format C_1
%format C_2
%format C_3
To compute the C3 linearization, the inheritance graph is extended with edges for the local precedence order of bases: for each class with ordered list of bases |CS| we insert an edge from |C_1| to |C_2|, an edge from |C_2| to |C_3|, and so forth. The ordering of edges is no longer relevant in the extended inheritance graph.

The C3 algorithm produces a depth first topological sort of the extended inheritance graph, provided that the extended graph is acyclic. See for example Figure \ref{example:extended-inheritance-graph}.

\begin{figure}
\begin{center}
\begin{tikzpicture}[scale=1.5, auto]
    %\draw[help lines] (0,0) grid (3,2);

    % Place nodes
    \node [class] (object) at (1,2) {|object|};
    \node [class] (A)      at (0,1) {|A|};
    \node [class] (B)      at (2,1) {|B|};
    \node [class] (C)      at (1,0) {|C|};

    % Draw edges
    \path [inherits] (A) -- (object);
    \path [inherits] (B) -- (object);
    \path [inherits] (C) -- (A);
    \path [inherits] (C) -- (B);

	\path [inherits] (A) edge [] node {new} (B);

\end{tikzpicture}
\end{center}
\caption{A (partial) extended inheritance graph. The marked arrow is added to the inheritance graph in Figure \ref{example:simple-inheritance-graph}. Thus the class |C| has the |MRO|~|C, A, B, object|.}
\label{example:extended-inheritance-graph}
\end{figure}


% this is an example where it matters that we have a _depth first_ ordering
%  object
%   / \
%  A   C
%  |   |
%  B-->D
%   \ /
%    E

%if style == newcode
\begin{code}
mro :: Heap -> Result -> [Result]
merge :: [Result] -> [Result] -> [Result]
\end{code}
%endif

%format b_1
%format b_2
%format mro heap a = MRO"~"heap(a)
%format `merge` = "~\sqcup~"
%{
%format foldl merge first map = MRO"~"heap(b_1) `merge` MRO"~"heap(b_2) `merge` "\cdots"
\mathhs
\begin{equation}
\begin{code}
mro heap a = a:bs `merge` foldl merge [] (map (mro heap) bs)
  where bs = getTupleValue heap (get (getObj heap a) __bases__)
\end{code}
\label{equation:mro}
\end{equation}
\plainhs
%}

The |MRO| is defined using a left associative merge operator~|as `merge` bs| that merges the two topological sorts~|as| and~|bs|. Note that, if |as| contains a class that is also in |bs|, that class will occur only once in |as `merge` bs|.

%{
%format | = "\text{ if }"
\mathhs
\begin{equation}
\begin{code}
[]      `merge`  bs                                = bs
as      `merge`  []                                = as
(a:as)  `merge`  (b:bs)  | a == b                  = a : as      `merge`  bs
(a:as)  `merge`  (b:bs)  | a /= b && a `nelem` bs  = a : as      `merge`  (b:bs)
(a:as)  `merge`  (b:bs)  | a /= b && b `nelem` as  = b : (a:as)  `merge`  bs
\end{code}
\end{equation}
\plainhs
%}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsubsection{Illegal inheritance graphs}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Not all inheritance graphs are considered legal, e.g., cyclic extended inheritance graphs are not allowed. This sometimes has unintuitive consequences, see for example Figure \ref{example:illigal-inheritance-graph}.

Furthermore, all classes must have at least one base, with the exception of |object|, and classes cannot have duplicate bases. Only classes can serve as bases. Finally, a class can only have one built-in superclass (e.g. |int| or |list|).

\begin{figure}
\begin{center}
\subfigure{
\begin{tikzpicture}[scale=1.5, auto]
    %\draw[help lines] (0,0) grid (3,2);

    % Place nodes
    \node [class] (object) at (1,2) {|object|};
    \node [class] (A)      at (0,1) {|A|};
    \node [class] (B)      at (2,1) {|B|};

    % Draw edges
    \path [inherits] (A) edge [] node {1} (object);
    \path [inherits] (A) edge [] node {2} (B);
    \path [inherits] (B) -- (object);
\end{tikzpicture}
} % end subfigure
\hspace{2em}
\subfigure{
\begin{tikzpicture}[scale=1.5, auto]
    %\draw[help lines] (0,0) grid (3,2);

    % Place nodes
    \node [class] (object) at (1,2) {|object|};
    \node [class] (A)      at (0,1) {|A|};
    \node [class] (B)      at (2,1) {|B|};

    % Draw edges
    \path [inherits] (A) edge [] node {2} (object);
    \path [inherits] (A) edge [] node {1} (B);
    \path [inherits] (B) -- (object);
\end{tikzpicture}
} % end subfigure
\end{center}
\caption{The inheritance graph on the right has an illegal ordering of its base classes: the extended inheritance graph has a cycle |object -> B -> object|. The inheritance graph on the right is legal, because |A| inherits |B| before |object| and thus has no cycle in its extended inheritance graph.}
\label{example:illigal-inheritance-graph}
\end{figure}

%if style==newcode
\begin{code}
hasCycles :: Heap -> [Result] -> Result -> Bool
hasCycles heap as a
  | a `elem` as = True
  | otherwise   = anyHasCyles [] bases
  where bases = getTupleValue heap (get (getObj heap a) __bases__)
        anyHasCyles cs [] = False
        anyHasCyles cs (b:bs)
          =  hasCycles heap (a:as ++ cs) b 
          || anyHasCyles (b:cs) bs
\end{code}
%endif

%if style==newcode
\begin{code}
noDuplicateBases :: [Result] -> Bool
noDuplicateBases [] = True
noDuplicateBases (a:as) 
  =  not (a `elem` as) 
  && noDuplicateBases as
\end{code}
%endif

%if style==newcode
\begin{code}
isClass :: Heap -> Result -> Bool
isClass heap a = __bases__ `isin` getObj heap a
\end{code}
%endif


%if style==newcode
\begin{code}
hasSingleBaseInst :: Heap -> Result -> Bool
hasSingleBaseInst heap a 
  = length (intersect [a_int, a_str, a_list, a_tuple, a_dict] (mro heap a)) < 2
\end{code}
%endif


%if style==newcode
\begin{code}
hasValidBaseInst :: [Result] -> Bool
hasValidBaseInst = not . any (`elem` [a_NoneType, a_bool, a_generator, a_function])

isValidClass :: Heap -> Result -> Bool
isValidClass heap a =  noDuplicateBases bases
                    && all (isClass heap) bases
                    && (a == a_object || length bases > 0)
                    && hasValidBaseInst bases
                    && not (hasCycles heap [] a)
                    && hasSingleBaseInst heap a
  where bases = getTupleValue heap (get (getObj heap a ) __bases__)
\end{code}
%endif


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Mappings of inherited attributes}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
As mentioned before, the |MRO| determines what attributes a class inherits. In order to lookup the attributes of an object~|getObj heap a|, we define two derived mappings that encapsulate the inheritance of attributes.


%format a_b
%format unions (heap) (as) = "\bigoplus_{"heap"}" as
%format (getInhAttrs (heap) (a)) = "\Phi^{"heap"}_{"a"}"
%{
%format a_1
%format a_n
\subsubsection{A mapping of class attributes}
We define the derived mapping of class attributes~|getInhAttrs heap a| to contain the inherited attributes of the class of |a| and its own attributes. It is defined using a capital version of the join operator~|<+>|. The capital join operator~|unions heap as| merges the objects |getObj heap a_1| through |getObj heap a_n|, where |n = length as|.
%}

\begin{code}
unions heap []      = empty
unions heap (a:as)  = unions heap as `union` getObj heap a
\end{code}

%The mapping of class attributes~|getInhAttrs heap a| joins the base attributes of the class~|getObj heap a_c| of object~|getObj heap a| with the mapping of base attributes~|getBaseAttrs heap a_c|.

\noindent
The derived mapping~|getInhAttrs heap a| is defined as the joined mappings of the |MRO| of the class of |a|.

\mathhs
\begin{equation}
\begin{code}
getInhAttrs heap a = unions heap (mro heap a_c)
  where a_c = get (getObj heap a) __class__
\end{code}
\label{equation:getInhAttrs}
\end{equation}
\plainhs


%format (getAllAttrs (heap) (a)) = "\Upsilon^{"heap"}_{"a"}"
\subsubsection{A mapping of all attributes}
Finally, the full collection of attributes~|getAllAttrs heap a| of an object~|getObj heap a| is defined as the combination of its inherited attributes and the object's own attributes. 

\mathhs
\begin{equation}
\begin{code}
getAllAttrs heap a = getInhAttrs heap a `union` getObj heap a
\end{code}
\label{equation:getAllAttrs}
\end{equation}
\plainhs


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Classes as types}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
In the object-oriented world the terms ``type'' and ``classes'' are often used interchangeably. In Python, the term ``type'' is somewhat ambiguously defined. We define two typing relations based on the class-of and base-of relations.

%For example the built-in function |Type(o)| returns the class of |o|, 

%Throughout these semantics ``type'' refers to a class unless specifically stated otherwise.


%format a_a
%format a_b

% unfortunately there seems to be a bug in lhs2tex that prevents us from 
% surrounding the heap with curly brackets in the rules for isSubclass and isInstOf
%format isSubclass heap a c = a "<:_"heap c
%format isNotSubclass heap a c = a "\nless:_"heap c
%format isInstOf a c heap = a ":_"heap c
%{
%format = = "\equiv"


\subsubsection{The subclass relation}
Classes are said to be subclasses, or subtypes, of their base classes and the base classes thereof. We define the subclass relation~|isSubclass heap a_a a_b| to hold if and only if |a_a| is a class, and |a_b| is in the |MRO| of |a_a|. If a class |a_a| is a subclass of |a_b|, we also say that |a_b| is a superclass of |a_a|.

\mathhs
\begin{equation}
\begin{code}
isSubclass heap a_a a_b = __bases__ `isin` getObj heap a_a && a_b `elem` mro heap a_a
\end{code}
\label{equation:subclass}
\end{equation}
\plainhs

\noindent
For example, the class |C| in Figure \ref{example:extended-inheritance-graph} is a subclass of itself, |A|, |B|, and |object|.


\subsubsection{The instance-of relation} \label{section:instance-of}
Before, we described the class-of relation as a relation between object and class: every object has exactly one class. The instance-of relation extends this notion to the superclasses of an object, i.e. an object is an instance of its class and all the superclasses of its class.
We define the instance-of relation~|isInstOf a_o a_c heap| to hold if and only if the class of |a_o| is a subclass of |a_c|.

%{
%format (get (map) (a)) = map(a)
\mathhs
\begin{equation}
\begin{code}
isInstOf a_o a_c heap = isSubclass heap (get (getObj heap a_o) __class__) a_c
\end{code}
\label{equation:instance-of}
\end{equation}
\plainhs
%}

For example, the object |c| in Figure \ref{example:simple-instance-tree} is an instance of |C|, |A|, |B| and |object| (assuming the inheritance graph of Figure \ref{example:simple-inheritance-graph}). 
%}

%if style == newcode
\begin{code}
isNotSubclass heap a_a a_b = not (isNotSubclass heap a_a a_b)
\end{code}
%endif
