%if style == newcode
\begin{code}
module Objects
( rewrite
, o_object
, o_Type
) where

import Data.List(intersect)

import VirtualMachine
import ObjectTheory
\end{code}
%endif

%{
%format a_f
%format a_o
%format a_v
%format a_x
%format e_o
%format a_c

%if not appendix
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Classes and objects} \label{chapter:objects}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
A number of statements and expressions are devoted to classes and objects. Most of which are common to object-oriented language. Class definitions, attribute access, attribute assignments and class functions are all common programming constructs.

Python differs from most other languages in its extremely dynamic implementation of objects: Python's objects, and even classes, can be altered at any time. Attributes, including the |__class__| and |__bases__| attributes, can be removed and (re)defined using delete and assignment statements. Furthermore, the semantics of attribute references, assignments, and deletions can be changed by overriding class members.

The example in Listing~\ref{example:objects} shows how the semantics of attribute assignment are overridden by the class function |__setattr__|, and how the attribute deletion statement removes |__setattr__| even after the class has been created.

\begin{lstlisting}[ caption={A simple use-case for classes and objects. The Counter class overrides |__setattr__| to prevent direct access to the internal counter~|x|. Because the class overrides |__setattr__| we are forced to use |object.__setattr__| to set the attributes of a counter.}
, label=example:objects 
]
class Counter(object) :
	def __init__(self) :
		object.__setattr__(self, "x", 0)
	def inc(self) :
		object.__setattr__(self, "x", self.x + 1)
	def __setattr__(self, attr, val) :
		print("Use inc() to change the counter value")
	

c = Counter()
print(c.x)               # prints 0
c.inc()
print(c.x)               # prints 1
c.x = 0                  # prints the error message
print(c.x)               # prints 1

del Counter.__setattr__  # remove the class function __setattr__
c.x = 5                  # nothing is printed
print(c.x)               # prints 5
\end{lstlisting}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Class definition} \label{section:class-definition}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
A class definition~|ClassStmt x e b| creates a new class object. Its bases are set to the value of |e|, and the variables bound in |b| define its attributes.

Class definitions are not declarative, like in many other object oriented languages, but imperative in style. The body~|b| and bases~|e| of a new class are executed and evaluated when the class definition is executed.


\rulesection{Evaluating the bases of a new class}
\label{fw-rewrite-ClassStmt}
Before the body is executed, the base objects have to be evaluated. The bases are represented by a tuple expression~|e|.

\begin{code}
rewrite  (State  heap  envs  ( stack                     )  ( FwStmt (ClassStmt x e b) )) =
         (state  heap  envs  ( stack :|: ClassFrame1 x b )  ( FwExpr e                 ))
\end{code}

Although one would expect to derive only from instances of |Type|, any object is accepted as a base at this point.


%format envs'
\rulesection{Executing the body of a class definition}
\label{bw-rewrite-ClassFrame1}
When |e| has been evaluated, its result~|a_b| is stored on the stack.

The body is executed with a new object~|env_c| on top of the environment stack. All variable bindings within the body will be bound in the object~|env_c|. Thus attributes of the new class are defined by ordinary variable bindings in the body.

\begin{code}
rewrite  (State  ( heap                                          )  ( envs :|: env_1 )  ( stack :|: ClassFrame1 x b                    )  ( BwResult a_b )) =
         (state  ( heap <+> [env_c |-> ObjValue empty NoneValue] )  ( envs'          )  ( stack :|: ClassFrame2 x a_b (envs :|: env_1) )  ( FwBlock b    ))
   where  env_c   =  freeAddr heap
          envs'   =  if isObjValue (get heap env_1)  then  envs :|: env_c 
                                                     else  envs :|: env_1 :|: env_c
\end{code}

Class definitions that are nested in class definitions, can not access the attributes of the class enclosing them through the environment. If the topmost environment~|env_1| is an object, the body of the class definition is executed without |env_1| in the environment stack. Nested function definitions are limited in the same way (see Equation~\ref{fw-rewrite-DefStmt}).


\rulesection{Creating a new class}
\label{bw-rewrite-ClassFrame}
After the body has been executed, we store the class name~|x| on the heap and construct a new class object. The new object is constructed by the constructor of the metaclass i.e., the class function~|__new__|. See Rule~\ref{fw-rewrite-Prim-Type.__new__} for the default constructor.

%format a_t
%format ClassConstrFrame x a_x a_b env_c = "\circ_{"x"}".__init__(a_x, a_b, env_c)
\begin{code}
rewrite (State  heap   ( envs'  :|:  env_c )  ( stack :|: ClassFrame2 x a_b (envs :|: env_1) )  ( BwResult a                                                     )) =
        (state  heap'  ( envs   :|:  env_1 )  ( stack :|: ClassConstrFrame x a_x a_b env_c   )  ( FwApp (get (getInhAttrs heap a_t) __new__) [a_t,a_x,a_b,env_c] ))
  where  heap'  =  heap <+> [ a_x |-> StringValue x ]
         a_x    =  freeAddr heap
         a_t    =  if __metaclass__ `isin` get heap env_c  then  get (get heap env_c) __metaclass__ 
                                                           else  a_Type
\end{code}

Note that this implementation of metaclasses is incomplete: in Python the metaclass is inherited from the base(s) of a class. This is more complicated than it seems at first glance. For example, when a new class |C| is created that inherits from two classes |A| and |B| which inherit from two different metaclasses $M_A$ and $M_B$ respectively, it is unclear which metaclass constructs |C|. Python will raise an exception in this particular situation. For the moment \minpy\ avoids these complications by disabling inheritance of metaclasses.


\rulesection{The default class constructor}
\label{fw-rewrite-Prim-Type.__new__}

The default class constructor extends the new class~|a_c| with name~|a_x|, bases~|a_b| and class |a_self|. The |__metaclass__| attribute, which typically refers to |a_self|, is removed from the new class.
If no valid inheritance graph can be created because the bases are bad (see \ref{section:inheritance-graph}), we raise a |TypeError|.

%format isValidClass heap a = heap(a)"~\text{is valid}"
\begin{code}
rewrite                       (State  heap   envs  stack  ( FwApp (Prim "Type" "__new__") [a_self,a_x,a_b,a_c] ))
  | isValidClass heap' a_c =  (state  heap'  envs  stack  ( BwResult a_c                                       ))
  | otherwise              =  (state  heap   envs  stack  ( raise a_TypeError []                               ))
  where  heap'  =  heap <+>  [a_c |-> (get heap a_c  <+>  [ __class__ |-> a_self
                                                          , __bases__ |-> a_b 
                                                          ]) 
                                                     <-> __metaclass__
                             ]
\end{code}


\rulesection{Initializing the new class}
\label{bw-rewrite-ClassConstrFrame}
New classes, created by the constructor |__new__| of the metaclass, are initialized by the |__init__| function of the metaclass. So when the constructor returns, the initializer is called.

The assignment on the stack binds the class name |x| to the new class~|a_o|. If the initialization function raises an exception, the assignment will be popped from the stack and the class will not be bound.

\begin{code}
rewrite (State  heap  envs  ( stack :|: ClassConstrFrame x a_x a_b env_c               )  ( BwResult a_o                                                       )) =
        (state  heap  envs  ( stack :|: InstrFrame (FwStmt (AssStmt x (addrExpr a_o))) )  ( FwApp (get (getInhAttrs heap a_t) __init__) [a_t, a_x, a_b, env_c] ))
  where a_t = get (get heap a_o) __class__
\end{code}

The original metaclass, whose constructor was called, can return any object. Therefore we call the initialization function of the returned object's class, rather than the original metaclass.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Object creation}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The creation of objects is handled by the |__call__|, |__new__|, and |__init__| attributes. Most classes will inherit those attributes from the base object |object|, whose primitives define the default construction mechanism.


%format ConstrFrame as = "\circ".__init__(as)
\rulesection{The default constructor}
\label{fw-rewrite-Prim-Type.__call__}
Every class can be called as if it were a function because all classes inherit, or override, this |__call__| primitive. If a class is called, its |__call__| attribute function is called (see Rule \ref{fw-rewrite-FwApp}). The |__call__| primitive consecutively calls the |__new__| and |__init__| to construct a new class and initialize it.

\begin{code}
rewrite  (State  heap  envs  ( stack                    )  ( FwApp (Prim "Type" "__call__") (a_c:as) )) =
         (state  heap  envs  ( stack :|: ConstrFrame as )  ( FwApp a_f (a_c:as)                      ))
  where a_f = get (unions heap (mro heap a_c)) __new__
\end{code}

The |__new__| function |a_f|, is looked up in the class itself or its bases. The frame |ConstrFrame as|, is put onto the stack to execute the |__init__| with the same arguments.


\rulesection{The default object creation}
\label{fw-rewrite-Prim-object.__new__}
The default implementation of |__new__| in |object| creates a new object with base |None|, and sets its class to |a_c|.

\begin{code}
rewrite   (State  ( heap                                                          )  envs  stack  ( FwApp (Prim "object" "__new__") (a_c:as) )) =
          (state  ( heap <+> [a |-> ObjValue (new [__class__ |-> a_c]) NoneValue] )  envs  stack  ( BwResult a                               ))
  where a = freeAddr heap
\end{code}

Primitive values that allow subclassing, have their own |__new__| attributes which initialize the base value to their primitive value. See for example Rule \ref{fw-rewrite-Prim-int.__new__}.


\rulesection{From creation to initialization}
\label{bw-rewrite-ConstrFrame}
The new object |a| returned by |__new__| is initialized by its class. If the initialization function does not raise an exception, the new object~|a| on the stack will be returned.

\begin{code}
rewrite  (State  heap  envs  ( stack :|: ConstrFrame as          )  ( BwResult a       )) = 
         (state  heap  envs  ( stack :|: InstrFrame (BwResult a) )  ( FwApp a_f (a:as) ))
  where a_f = get (getInhAttrs heap a) __init__
\end{code}


\rulesection{The default object initialization}
\label{fw-rewrite-Prim-object.__init__}
The default initialization does nothing at all and returns |None|.

\begin{code}
rewrite  (State  heap  envs  stack  ( FwApp (Prim "object" "__init__") as )) =
         (state  heap  envs  stack  ( BwResult a_None                     ))
\end{code}

Like |Prim "object" "__new__"|, this primitive accepts an arbitrary number of arguments. This allows a user defined |__new__| function to accept any number of arguments.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Attribute access}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The attribute access expression |AttrExpr e x| evaluates to the attribute |x| of object |e|. However, its semantics can be changed by overriding the |__getattribute__| function.


\rulesection{Evaluating the attribute access expression}
\label{fw-rewrite-AttrExpr}
First the left hand side expression |e| of the expression |AttrExpr e x| is evaluated.

\begin{code}
rewrite  (State  heap  envs  ( stack                 )  ( FwExpr (AttrExpr e x) )) =
         (state  heap  envs  ( stack :|: AttrFrame x )  ( FwExpr e              ))
\end{code}


\rulesection{Calling |__getattribute__|}
\label{bw-rewrite-AttrFrame}
When the left hand side expression |e| has been evaluated to an object |a|, the |__getattribute__| function of |a| is called. The function |__getattribute__| is called with the evaluted object |a| (we instantiate |__getattribute__| with |a|) and the attribute name |a_x|.

\begin{code}
rewrite  (State  ( heap                              )  envs  ( stack :|: AttrFrame x )  ( BwResult a         )) =
         (state  ( heap <+> [a_x |-> StringValue x ] )  envs  ( stack                 )  ( FwApp a_f [a, a_x] ))
  where  a_f  =  get (getInhAttrs heap a) __getattribute__ 
         a_x  =  freeAddr heap
\end{code}

The |__getattribute__| function is retrieved from the class of the object |a|. This means that it cannot be overridden by instance attributes.


\rulesection{Getting attributes and instantiating function attributes}
\label{fw-rewrite-Prim-object.__getattribute__}
The default implementation of |__getattribute__| for objects retrieves an attribute according to the |MRO| (see Section \ref{section:inheritance}).  When a class function of an object~|o| is retrieved it is automatically instantiated with the object, i.e., its first argument is set to |o|. Once instantiated, a class function will not be instantiated again.

We have four cases:
\begin{enumerate}
\item If |a_x| is an instance attribute, we return its value uninstantiated. 
\item If |a_x| is an inherited attribute and is instantiatable, we return the instantiated attribute.
\item If |a_x| is an inherited attribute but is not instantiatable, we return it.
\item Otherwise the attribute is not defined so we raise an |AttributeError|.
\end{enumerate}

%format isInstantiable heap a = heap(a) "~\text{is instantiable}"
%format isNotInstantiable heap a = heap(a) "~\text{is not instantiable}"
\begin{code}
rewrite                                                                                              (State  heap  envs  stack  ( FwApp (Prim "object" "__getattribute__") [a_o,a_x] ))
    | x `isin` getObj heap a_o                                                                    =  (state  heap  envs  stack  ( BwResult a                 ))
    | x `isnotin` getObj heap a_o  && x `isin` getInhAttrs heap a_o  && isInstantiable heap a     =  (state  heap  envs  stack  ( BwResult (Inst a_o a)      ))
    | x `isnotin` getObj heap a_o  && x `isin` getInhAttrs heap a_o  && isNotInstantiable heap a  =  (state  heap  envs  stack  ( BwResult a                 ))
    | otherwise                                                                                   =  (state  heap  envs  stack  ( raise a_AttributeError []  ))
  where  a  =  get (getAllAttrs heap a_o) x
         x  =  getStringValue heap a_x
\end{code}

Note that only uninstantiated functions are instantiable. Callable objects are instantiated when it is called (see Rule \ref{fw-rewrite-FwApp}).


\rulesection{Executing an instantiated function}
\label{fw-rewrite-Inst}
A function |a_f| instantiated with object |a_o| is denoted as |Inst a_o a_f|. When an instantiated function is evaluated in a function application, the object |a_o| is inserted as the first argument of the function call.

\begin{code}
rewrite (State  heap  envs  stack  ( FwApp (Inst a_o a_f) (      as) )) =
        (state  heap  envs  stack  ( FwApp (a_f         ) (a_o : as) ))
\end{code}


\rulesection{Getting class members and inserting type checks}
\label{fw-rewrite-Prim-Type.__getattribute__}
The |__getattribute__| primitive of |Type| is equivalent to the |__getattribute__| primitive of |object|, except for one difference: we insert type annotations for instantiatable class functions. The type annotations ensure that class functions are only applied to instances of their class.

\begin{code}
rewrite                                                    (State  heap  envs  stack  ( FwApp (Prim "Type" "__getattribute__") [a_o,a_x]   ))
    | x `isin` getObj heap a_o && isInstantiable heap a =  (state  heap  envs  stack  ( BwResult (Typed a [a_o])                           ))
    | otherwise                                         =  (state  heap  envs  stack  ( FwApp (Prim "object" "__getattribute__") [a_o,a_x] ))
  where  a  =  get (getAllAttrs heap a_o) x
         x  =  getStringValue heap a_x
\end{code}
    


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Attribute assignment} \label{section:attribute-assignment}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
By default, the attribute assignment statement |AttrAssStmt e_o x e| (re)defines the attribute |x| of the target object |e_o| to the value |e|. The semantics of attribute assignment statements are determined by the |__setattr__| attribute of an object's class. By overloading it, a programmer can redefine the semantics.


\rulesection{Evaluating the right hand side of an attribute assignment}
\label{fw-rewrite-AttrAssStmt}
The evaluation of the assignment starts with the right hand side: the value~|e|. Note that this behaviour is different from the usual left to right evaluation.

\begin{code}
rewrite  (State  heap  envs  ( stack                         )  ( FwStmt (AttrAssStmt e_o x e) )) =
         (state  heap  envs  ( stack :|: AttrAssFrame1 e_o x )  ( FwExpr e                     ))
\end{code}


\rulesection{Evaluating the left hand side of an attribute assignment}
\label{bw-rewrite-AtrrAssFrame1}
Next, the left hand side~|e_o| is evaluated, while the value |a| of the right hand side is placed on the stack.

\begin{code}
rewrite  (State  heap  envs  ( stack :|: AttrAssFrame1 e_o x )  ( BwResult a )) =
         (state  heap  envs  ( stack :|: AttrAssFrame2 x a   )  ( FwExpr e_o ))
\end{code}


\rulesection{Calling |__setattr__|}
\label{bw-rewrite-AttrAssFrame2}
Finally we call the inherited attribute |__setattr__| of |a_o|, with the target object |a_o|, the assigned attribute name |a_x|, and the new value |a|.

\begin{code}
rewrite  (State  ( heap                             )  envs  ( stack :|: AttrAssFrame2 x a )  ( BwResult a_o          )) =
         (state  ( heap <+> [a_x |-> StringValue x] )  envs  ( stack                       )  ( FwApp a_f [a_o,a_x,a] ))
  where  a_f  =  get (getInhAttrs heap a_o) __setattr__
         a_x  =  freeAddr heap
\end{code}


\rulesection{The default implementation of |__setattr__| for objects}
\label{fw-rewrite-Prim-object.__setattr__}
The default implementation of |__setattr__| for objects updates the object |a_o| with |[x ||-> a]|.

\begin{code}
rewrite  (State  ( heap                  )  envs  stack  ( FwApp (Prim "object" "__setattr__") [a_o, a_x, a] )) =
         (state  ( heap <+> [a_o |-> o'] )  envs  stack  ( BwResult a_None                                   ))
  where o' = get heap a_o <+> [getStringValue heap a_x |-> a]
\end{code}


%\rulesection{The default implementation of |__setattr__| for types}
%\label{fw-rewrite-Prim-Type.__setattr__}
%The default implementation of |__setattr__| for types updates the class |a_o| with |[x ||-> a]|.
%
%\begin{code}
%rewrite  (State  heap   envs  stack  ( FwApp (Prim "Type" "__setattr__") [a_o, a_x, a] )) =
%         (state  heap'  envs  stack  ( BwResult a_None                                 ))
%  where heap' = heap <+> [ a_o |-> get heap a_o <+> [getStringValue heap a_x |-> a] ]
%\end{code}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Attribute deletion}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
By default, an attribute deletion statement |AttrDelStmt e_o x| removes the attribute |x| from object |e_o| if it exists. The semantics can be redefined by overloading |__delattr__|.


\rulesection{Evaluating the left hand side of an attribute deletion}
\label{fw-rewrite-AttrDelStmt}
We start by evaluating the object |e_o|.

\begin{code}
rewrite  (State  heap  envs  ( stack                    )  ( FwStmt (AttrDelStmt e_o x) )) =
         (state  heap  envs  ( stack :|: AttrDelFrame x )  ( FwExpr e_o                 ))
\end{code}


\rulesection{Calling |__delattr__|}
\label{bw-rewrite-AttrDelFrame}
When |e_o| has been evaluated to |a_o|, we call the |__delattr__| attribute inherited by |a_o|.

\begin{code}
rewrite  (State  ( heap                             )  envs  ( stack :|: AttrDelFrame x )  ( BwResult a_o        )) =
         (state  ( heap <+> [a_x |-> StringValue x] )  envs  ( stack                    )  ( FwApp a_f [a_o,a_x] ))
  where  a_f  =  get (getInhAttrs heap a_o) __delattr__
         a_x  =  freeAddr heap
\end{code}


\rulesection{The default implementation of |__delattr__|}
\label{bw-rewrite-Prim-object.__delattr__}
The default implementation of |__delattr__| removes the attribute |a_x| from the object |a_o| if |x| is defined in |a_o|. Otherwise an |AttributeError| is raised.

\begin{code}
rewrite                                             (State  ( heap                  )  envs  stack  ( FwApp (Prim "object" "__delattr__") [a_o, a_x] ))
  | getStringValue heap a_x `isin` get heap a_o  =  (state  ( heap <+> [a_o |-> o'] )  envs  stack  ( BwResult a_None                                ))
  | otherwise                                    =  (state  ( heap                  )  envs  stack  ( raise a_AttributeError []                      ))
   where o' = get heap a_o <-> getStringValue heap a_x
\end{code}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%if style==newcode
\begin{code}
rewrite state = rwpart1 state
\end{code}
%endif

%if style==newcode
\begin{code}
o_object :: Value
o_Type :: Value
\end{code}
%endif


%else
%format o_object
%format o_Type

\section{The base objects |object| and |Type|}
\begin{code}
o_object = new  
  [  __getattribute__  |-> Prim "object" "__getattribute__"
  ,  __setattr__       |-> Prim "object" "__setattr__"
  ,  __delattr__       |-> Prim "object" "__delattr__"
  ,  __new__           |-> Prim "object" "__new__"
  ,  __init__          |-> Prim "object" "__init__"
  ,  __str__           |-> Prim "object" "__str__"
  ,  __bases__         |-> a_emptyTuple
  ,  __class__         |-> a_Type
  ]
\end{code}

\begin{code}
o_Type = new  
  [  __getattribute__  |-> Prim "Type" "__getattribute__"
  ,  __bases__         |-> a_objectT
  ,  __class__         |-> a_Type
  ,  __new__           |-> Prim "Type" "__new__"
  ,  __call__          |-> Prim "Type" "__call__"
  ,  __str__           |-> Prim "Type" "__str__"
  ]
\end{code}

\rulesection{The default string representation}
\label{fw-rewrite-Prim-object.__str__}
\begin{code}
rwpart1  (State  ( heap                                                               )  envs  stack  ( FwApp (Prim "object" "__str__") [a_self] )) =
         (state  ( heap <+> [a |-> StringValue ("<object at " ++ show a_self ++ ">")] )  envs  stack  ( BwResult a                               ))
  where a = freeAddr heap
\end{code}

\rulesection{Default string representation for classes}
\begin{code}
rwpart1  (State  ( heap                                                              )  envs  stack  ( FwApp (Prim "Type" "__str__") [a_self] )) =
         (state  ( heap <+> [a |-> StringValue ("<class at " ++ show a_self ++ ">")] )  envs  stack  ( BwResult a                             ))
  where a = freeAddr heap
\end{code}
%endif

%if style == newcode
\begin{code}
rwpart1 _ = skipState

isInstantiable heap a = isNotInsted a && isInstOf a a_function heap
isNotInstantiable heap a = not (isNotInsted a && isInstOf a a_function heap)
\end{code}
%endif
%}
