%if style == newcode
\begin{code}
module Operators(rewrite) where
import VirtualMachine
import ObjectTheory
\end{code}
%endif


%{
%format op = "\bigcirc"
%format a_l
%format a_r
%format e_l
%format e_r
%format getUOpFunc o = "\text{attribute of}~" o
%format getOpFunc o = "\text{attribute of}~" o
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Operators} \label{chapter:operators}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The semantics of most operators are determined by the class function of their operands. For example the addition operator~$+$ is implemented by the class function |__add__| of the left operand. An expression |BinOpExpr Add e_l e_r| could\footnote{This simple rewrite of the expression has a different order of evaluation and assumes |__getattribute__| is not overridden.} be translated to |CallExpr (AttrExpr e_l __add__) [e_r]|.

Boolean operators are an exception to this rule. They are not implemented by class function but depend on the class function~|__nonzero__| to coerce values to booleans.


\section{Unary operators}
Unary operators have only one operand. Evaluation takes three simple steps: evaluation of the operand, getting the operator's attribute from the operand's class, execute the class function.


\rulesection{Evaluating the operand of a unary operator expression}
\label{fw-rewrite-UnaryOpExpr}
\begin{code}
rewrite  (State  heap  envs  ( stack                     )  ( FwExpr (UnaryOpExpr op e) )) =
         (state  heap  envs  ( stack :|: UnaryOpFrame op )  ( FwExpr e                  ))
\end{code}


\rulesection{Evaluating the unary not operator}
\label{bw-rewrite-UnaryOpFrame-Not}
The |Not| operator negates the value returned by the class function~|__nonzero__| of its operand. If the operand has no attribute~|__nonzero__|, we raise a |TypeError|.

%format NotFrame = Not "\circ_{"__nonzero__"}"
\begin{code}
rewrite                                       (State  heap  envs  ( stack :|: UnaryOpFrame Not )  ( BwResult a           ))
  | __nonzero__ `isin` getInhAttrs heap a  =  (state  heap  envs  ( stack :|: NotFrame         )  ( FwApp a_f [a]        ))
  | otherwise                              =  (state  heap  envs  ( stack                      )  ( raise a_TypeError [] ))
   where a_f = get (getInhAttrs heap a) __nonzero__
\end{code}


\rulesection{Return the negated value of a not operand}
\label{bw-rewrite-NotFrame}
If |__nonzero__| returned a boolean, we return its negated value. Otherwise, if |__nonzero__| returned something else, we raise a |TypeError|.

\begin{code}
rewrite                        (State  heap  envs  ( stack :|: NotFrame )  ( BwResult a                                    ))
  | isInstOf a a_bool heap  =  (state  heap  envs  ( stack              )  ( FwExpr (BoolExpr (not (getBoolValue heap a))) ))
  | otherwise               =  (state  heap  envs  ( stack              )  ( raise a_TypeError []                          ))
\end{code}


\rulesection{Executing the function of a unary operator}
\label{bw-rewrite-UnaryOpFrame}
For other operators, we call the appropriate operator function.

\begin{code}
rewrite                                         (State  heap  envs  ( stack :|: UnaryOpFrame op )  ( BwResult a           ))
  | getUOpFunc op `isin` getInhAttrs heap a  =  (state  heap  envs  ( stack                     )  ( FwApp a_f [a]        ))
  | otherwise                                =  (state  heap  envs  ( stack                     )  ( raise a_TypeError [] ))
  where a_f = get (getInhAttrs heap a) (getUOpFunc op)
\end{code}


\section{Binary operators}
Most operators have two operands: a left and a right hand side operator. The semantics of the operator are defined by the left hand side operator.


\rulesection{Evaluating the left hand side of a binary operator}
\label{fw-rewrite-BinOpExpr}
First, the left hand side operator is evaluated while the remaining expression is put on the stack.

\begin{code}
rewrite (State  heap  envs  ( stack                           )  ( FwExpr (BinOpExpr op e_l e_r) )) =
        (state  heap  envs  ( stack :|: LeftBinOpFrame op e_r )  ( FwExpr e_l                    ))
\end{code}


\rulesection{Evaluating the right hand side of a binary operator}
\label{bw-rewrite-LeftBinOpFrame}
The boolean operators |And| and |Or| are lazy. They only evaluate their right hand side if needed. Therefore, we call |__nonzero__| for the |And| and |Or| operators before |e_r| is evaluated. If |__nonzero__| is not defined, we raise a |TypeError|.

Other operators are executed after both their operators have been evaluated. Therefore, we evaluate |e_r| and store |a_l| on the stack.

\begin{code}
rewrite                                                                  (State  heap  envs  ( stack :|: LeftBinOpFrame op e_r  )  ( BwResult a_l         ))
  | op `elem` [And,Or] && __nonzero__ `isin` getInhAttrs heap a_l     =  (state  heap  envs  ( stack :|: LeftBoolOpFrame op e_r )  ( FwApp a_f [a_l]      ))
  | op `elem` [And,Or] && __nonzero__ `isnotin` getInhAttrs heap a_l  =  (state  heap  envs  ( stack                            )  ( raise a_TypeError [] ))
  | otherwise                                                         =  (state  heap  envs  ( stack :|: RightBinOpFrame a_l op )  ( FwExpr e_r           ))
   where a_f = get (getInhAttrs heap a_l) __nonzero__
\end{code}

Evaluation of the boolean operators continues at Rules \ref{bw-rewrite-LeftBoolOpFrame-And} and \ref{bw-rewrite-LeftBoolOpFrame-Or}. The |In| operator continues at Rule \ref{bw-rewrite-RightBinOpFrame-Is}. Other operators are evaluated by Rule \ref{bw-rewrite-RightBinOpFrame}.


\rulesection{Evaluating an |is| expression}
\label{bw-rewrite-RightBinOpFrame-Is}
The |Is| operator determines whether its operands are the same object. This is achieved by comparing the addresses of the operands.

\begin{code}
rewrite  (State  heap  envs  ( stack :|: RightBinOpFrame a_l Is )  ( BwResult a_r                   )) =
         (state  heap  envs  ( stack                            )  ( FwExpr (BoolExpr (a_l == a_r)) ))
\end{code}


\rulesection{Calling the operator attribute}
\label{bw-rewrite-RightBinOpFrame}
When both operands of the operator have been evaluated, the operator's class function of |a_l| is called with the operands as its arguments.

\begin{code}
rewrite                                          (State  heap  envs  ( stack :|: RightBinOpFrame a_l op )  ( BwResult a_r         ))
  | getOpFunc op `isin` getInhAttrs heap a_l  =  (state  heap  envs  ( stack                            )  ( FwApp a_f [a_l,a_r]  ))
  | otherwise                                 =  (state  heap  envs  ( stack                            )  ( raise a_TypeError [] ))
  where a_f = get (getInhAttrs heap a_l) (getOpFunc op)
\end{code}


\rulesection{Lazy execution of the |And| operator}
\label{bw-rewrite-LeftBoolOpFrame-And}
The |And| expression returns |e_r| if the |e_l| is not zero (its |__nonzero__| returned |True|), otherwise it returns |False|.
Note that the right hand side |e_r| can evaluate to any value.

\begin{code}
rewrite                                                     (State  heap  envs  ( stack :|: LeftBoolOpFrame And e_r )  ( BwResult a              ))
  | isInstOf a a_bool heap  && getBoolValue heap a       =  (state  heap  envs  ( stack                             )  ( FwExpr e_r              ))
  | isInstOf a a_bool heap  && not (getBoolValue heap a) =  (state  heap  envs  ( stack                             )  ( FwExpr (BoolExpr False) ))
  | otherwise                                            =  (state  heap  envs  ( stack                             )  ( raise a_TypeError []    ))
\end{code}



\rulesection{Lazy execution of the |Or| operator}
\label{bw-rewrite-LeftBoolOpFrame-Or}
Conversely, the |Or| expression returns |e_r| when |e_l| is zero and True otherwise.

\begin{code}
rewrite                                                    (State  heap  envs  ( stack :|: LeftBoolOpFrame Or e_r )  ( BwResult a             ))
  | isInstOf a a_bool heap && getBoolValue heap a       =  (state  heap  envs  ( stack                            )  ( FwExpr (BoolExpr True) ))
  | isInstOf a a_bool heap && not (getBoolValue heap a) =  (state  heap  envs  ( stack                            )  ( FwExpr e_r             ))
  | otherwise                                           =  (state  heap  envs  ( stack                            )  ( raise a_TypeError []   ))
\end{code}


%format e_i
%format e_v
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Container operators}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Containers are values containing values, or actually addresses pointing to values. Individual items of a container can be retrieved, set, or deleted using the container operators. The semantics of those operators depend on the implementation of the container's class functions |__setitem__|, |__getitem__|, and |__delitem__|.

For example, lists, tuples and dictionaries are containers with their own implementation of the container attributes (see Appendix \ref{chapter:builtins}). Not just built-in values are containers though, any object with the right class functions is considered a container.


\subsection*{Retrieving a container item}
The subscription~|SliceExpr e e_i| retrieves an item with key or index~|e_i| from a container~|e|.


%format a_v
\rulesection{Evaluating the container expression}
\label{fw-rewrite-SliceExpr}
First, the container~|e| is evaluated.

\begin{code}
rewrite (State  heap  envs  ( stack                        )  ( FwExpr (SliceExpr e e_i) )) =
        (state  heap  envs  ( stack :|: ContSliceFrame e_i )  ( FwExpr e                 ))
\end{code}


\rulesection{Evaluating the key expression}
\label{bw-rewrite-ContSliceFrame}
Then, the index or key~|e_i| is evaluated.

\begin{code}
rewrite (State  heap  envs  ( stack :|: ContSliceFrame e_i )  ( BwResult a )) =
        (state  heap  envs  ( stack :|: KeySliceFrame a    )  ( FwExpr e_i ))
\end{code}


\rulesection{Calling |__getitem__| of a container}
\label{bw-rewrite-KeySliceFrame}
Finally, the container's class function |__getitem__| is called to retrieve the item.

\begin{code}
rewrite                                       (State  heap  envs  ( stack :|: KeySliceFrame a )  ( BwResult a_i         ))
  | __getitem__ `isin` getInhAttrs heap a  =  (state  heap  envs  ( stack                     )  ( FwApp a_f [a,a_i]    ))
  | otherwise                              =  (state  heap  envs  ( stack                     )  ( raise a_TypeError [] ))
   where a_f = get (getInhAttrs heap a) __getitem__ 
\end{code}


\subsection*{Setting a container item}
The subscription assignment statement~|SliceAssStmt e e_i e_v| sets an item~|e_v| in the container~|e| at key or index~|e_i|. Like the object attribute assignment (see Section~\ref{section:attribute-assignment}) the expressions in the container assignment are not evaluated from left to right, but the value~|e_v| is evaluated first.


\rulesection{Evaluating the right hand side expression}
\label{fw-rewrite-SliceAssStmt}
First, the item's value |e_v| is evaluated.

\begin{code}
rewrite (State  heap  envs  ( stack                              )  ( FwStmt (SliceAssStmt e e_i e_v) )) =
        (state  heap  envs  ( stack :|: ValueSliceAssFrame e e_i )  ( FwExpr e_v                      ))
\end{code}


\rulesection{Evaluating the container expression}
\label{bw-rewrite-ValueSliceAssFrame}
Then, the container |e| is evaluated.

\begin{code}
rewrite (State  heap  envs  ( stack :|: ValueSliceAssFrame e e_i  )  ( BwResult a_v )) =
        (state  heap  envs  ( stack :|: ContSliceAssFrame e_i a_v )  ( FwExpr e     ))
\end{code}


\rulesection{Evaluating the key expression}
\label{bw-rewrite-ContSliceAssFrame}
Next, the index or key |e_i| is evaluated.

\begin{code}
rewrite (State  heap  envs  ( stack :|: ContSliceAssFrame e_i a_v )  ( BwResult a )) =
        (state  heap  envs  ( stack :|: KeySliceAssFrame a a_v    )  ( FwExpr e_i ))
\end{code}


\rulesection{Calling |__setitem__|}
\label{bw-rewrite-KeySliceAssFrame}
Finally, the container's class function|__setitem__| is called to insert the new item.

\begin{code}
rewrite                                       (State  heap  envs  ( stack :|: KeySliceAssFrame a a_v )  ( BwResult a_i          ))
  | __setitem__ `isin` getInhAttrs heap a  =  (state  heap  envs  ( stack                            )  ( FwApp a_f [a,a_i,a_v] ))
  | otherwise                              =  (state  heap  envs  ( stack                            )  ( raise a_TypeError []  ))
   where a_f = get (getInhAttrs heap a) __setitem__ 
\end{code}


\subsection*{Deleting a container item}
The subscription delete statement~|SliceDelStmt e e_i| removes an item with key or index~|e_i| from the container~|e|.


\rulesection{Evaluating the container expression}
\label{fw-rewrite-SliceDelStmt}
First, the container~|e| is evaluated.

\begin{code}
rewrite (State  heap  envs  ( stack                           )  ( FwStmt (SliceDelStmt e e_i) )) =
        (state  heap  envs  ( stack :|: ContSliceDelFrame e_i )  ( FwExpr e                    ))
\end{code}


\rulesection{Evaluating the key expression}
\label{bw-rewrite-ContSliceDelFrame}
Then, the key or index~|e_i| is evaluated.

\begin{code}
rewrite (State  heap  envs  ( stack :|: ContSliceDelFrame e_i )  ( BwResult a )) =
        (state  heap  envs  ( stack :|: KeySliceDelFrame a    )  ( FwExpr e_i ))
\end{code}


\rulesection{Calling |__delitem__|}
\label{bw-rewrite-KeySliceDelFrame}
Finally, the container's class attribute |__delitem__| is called to delete the item.

\begin{code}
rewrite                                       (State  heap  envs  ( stack :|: KeySliceDelFrame a )  ( BwResult a_i         ))
  | __delitem__ `isin` getInhAttrs heap a  =  (state  heap  envs  ( stack                        )  ( FwApp a_f [a,a_i]    ))
  | otherwise                              =  (state  heap  envs  ( stack                        )  ( raise a_TypeError [] ))
   where a_f = get (getInhAttrs heap a) __delitem__ 
\end{code}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%if style == newcode
\begin{code}
rewrite _ = skipState
\end{code}
%endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%if style==newcode
\begin{code}
getOpFunc op = case op of
  Add -> "__add__"
  Sub -> "__sub__"
  Mul -> "__mul__"
  Div -> "__div__"
  Pow -> "__pow__"
  Mod -> "__divmod__"
  FloorDiv -> "__floordiv__"

  LShift -> "__lshift__"
  RShift -> "__rshift__"
  
  BitOr -> ""
  BitXor -> ""
  BitAnd -> ""

  Eq -> "__eq__"
  NotEq -> "__neq__"
  Lt -> "__lt__"
  LtE -> "__le__"
  Gt -> "__gt__"
  GtE -> "__ge__"

  In -> "__contains__"

  And -> "__and__"
  Or -> "__or__"

getUOpFunc op = case op of
  Invert -> "__invert__"
  USub -> "__neg__"
\end{code}
%endif

%}
