%if style == newcode
\begin{code}
module Preliminaries
  ( Mapping(..)
  , (<+>)
  , (<->)
  , (|->)
  , isnotin
  , Stack(..)
  , sFromList
  , sElem
  , sPrepend
  , sConcat
  , sSingleton
  , intercalate
  , nelem
  ) where

import Data.List(intersperse)
--import Data.List(intercalate) --not available?
\end{code}
%endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Preliminaries} \label{chapter:preliminaries}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Before discussing the semantics themselves we introduce the notational conventions used for lists and mappings.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Lists}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%format head lst = lst"_1"
%%format tail lst  = lst"_{2..}"

%format length lst = "|"lst"|"
%format null lst = "|"lst"|" == 0
%format `nelem` = "\notin"

%format es = "\overline{e}"
%format as = "\overline{a}"
%format bs = "\overline{b}"
%format xs = "\overline{x}"
%format qs = "\overline{q}"
%format ps = "\overline{p}"

In these semantics we use many lists. Lists are sequences which are ordered collections of elements. For example, a list~|[3,5,2,5]| contains the elements 3, 5, 2, and 5 in that order. Lists can also be empty, which would be denoted as pair of empty brackets. 

%format q_2
By convention we overline names of lists, e.g. |as| would be a list of addresses. Individual elements of |as| are referred to by their index starting at 1. For example, |q_2| refers to the number 4 in the list~|qs = [1,4,3]|. The length of a list~| ||qs|| | is defined as the number of elements in the list. The operator |:| prepends a list with a new element and the operator~|++| concatenates two lists.

%if style==newcode
\begin{code}
intercalate xs xss = concat (intersperse xs xss)

nelem x y = not (elem x y)
\end{code}
%endif


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Mappings} \label{section:mappings}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\newcommand{\dom}[1]{#1}
\newcommand{\range}[1]{\text{range}~#1}

%format rm m k = m "\ominus" k
%format `rm` = "\ominus"
%format <-> = "\ominus"
%format (get m (k)) = m"("k")"
%format `isin` = "\in"
%format `isnotin` = "\notin"
%format new lst = lst
%format empty = "\emptyset"
%format union x y = x "\oplus" y
%format `union` = "\oplus"
%format insert x y = x "\oplus" y
%format `insert` = "\oplus"
%format <+> = "\oplus"

%format |-> = "\rightarrow"

%{
A mapping is a partial function that maps keys to values. For example, arrays and hash tables are mappings. Python itself has native support for mappings, which it calls dictionaries.

We use a number of special operators and notations to describe, alter and inspect maps. These operators have the same semantics for all mappings, regardless of their specific contents.

The list of key-value pairs $ \lbrack k_1 \mapsto v_1, ..., k_n \mapsto v_n \rbrack $
represents a mapping $m$ of keys $k_{1...n}$ to values $v_{1...n}$ respectively. Each key $k \in \dom{m}$ maps to a value $v = m(k)$.
Thus, the key-value pairs in the mapping are unique. An empty mapping is denoted as $\emptyset$, which is never used for empty sets.

The operator $\oplus$ combines two mappings $A$ and $B$ to create a new mapping that consists of the key-value pairs of both $A$ and $B$. For duplicate keys, the key-value pairs of the left hand side of $\oplus$ take precedence over the right hand side. Formally defining this operator, we have:
\begin{equation*}
(A \oplus B)(k) =
\begin{cases}
B(k) & \text{if } k \in B \\
A(k) & \text{otherwise}
\end{cases}
\end{equation*}

The operator $\ominus$ removes a key-value pair, identified by its key, from a mapping. If the key is not in the domain of the mapping, nothing changes. Formally defining this operator, we have:

\begin{equation*}
(A \ominus k)(j) =
\begin{cases}
A(j)  & \text{if } k \neq j \\
\perp & \text{otherwise}
\end{cases}
\end{equation*}

%if style == newcode
\begin{code}
class Mapping m k v | m -> k, m -> v where
  rm :: m -> k -> m
  get :: m -> k -> v
  isin :: k -> m -> Bool
  new :: [(k, v)] -> m
  union :: m -> m -> m
  insert :: m -> [(k,v)] -> m
  insert x y = x `union` new y
  empty :: m
  empty = new []

infixl 9 <->
infixr 9 <+>
infix 8 |->

(<+>) :: Mapping m k v => m -> [(k, v)] -> m
(<+>) = insert
(<->) :: Mapping m k v => m -> k -> m
(<->) = rm

isnotin x y = not (isin x y)

(|->) :: x -> y -> (x,y)
x |-> y = (x,y)
\end{code}
%endif
%}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%\section{Stacks}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%format sSingleton (f) = f
%format EmptyStack = "\Box"
%format :|: = "|"
%format `sPrepend` = "|"
%format `sConcat` = "|"
%format `sElem` = "\in"
%{
%if style == newcode
\begin{code}
data Stack a  =  Stack a :|: a
              |  EmptyStack
              |  MarkedStack Int (Stack a)
              deriving (Show, Eq)

infixl 6 :|:

sSingleton :: a -> Stack a
sSingleton a = EmptyStack :|: a

sElem :: (Eq a) => a -> Stack a -> Bool
sElem x EmptyStack = False
sElem x (s :|: y) = x == y ||x `sElem` s

sConcat :: Stack a -> Stack a -> Stack a
sConcat s EmptyStack = s
sConcat sx (sy :|: a) = sConcat sx sy :|: a

sPrepend :: a -> Stack a -> Stack a
sPrepend a EmptyStack = EmptyStack :|: a
sPrepend a (xs :|: b) = sPrepend a xs :|: b

sFromList :: [a] -> Stack a
sFromList [] = EmptyStack
sFromList (a:as) = sFromList as :|: a

infixl 8 `sPrepend`
infixl 7 `sConcat`
infixl 4 `sElem`
\end{code}
%endif
%}

