%if style == newcode
\begin{code}
module PythonAST
  ( Expr(..)
  , Stmt(..)
  , Block(..)
  , BinOp(..)
  , UnaryOp(..)
  , Name
  , stmtBlock
  , blockFromList
  ) where

import Preliminaries(intercalate)
\end{code}
%endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Abstract syntax} \label{chapter:AST}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
This chapter describes the abstract syntax of \minpy. First we will introduce the syntax of expressions and operators, followed by statements and blocks. The semantics of each syntactical construct will be described in the following chapters. 

The syntax definitions in this chapter are reformatted Haskell data definitions that mimic concrete Python syntax. Lists of a syntax element~$A$ are denoted as $\langle A \rangle^{\ast}$.

Because the concrete syntax of Python has already been specified by the reference manual~\cite{vanrossum1995prm}, it will not be discussed here. The parser used by the \minpy\ interpreter only implements a minimal subset of Python's concrete syntax.

The abstract syntax clearly shows some of the limitations of \minpy\ compared to full Python. For example, list comprehension expressions are missing and try-except statements are limited to a single except clause. To keep the specification simple, we have made various optional elements mandatory. For example, the else branch in the if-then-else statement is not optional in this abstract syntax.

%extra limitations:
%\begin{itemize}
%\item list of import names must have a length $> 0$
%\end{itemize}

% formatting rules for expressions
%format VarExpr v = v
%format CallExpr (e) args = e(args)
%format BoolExpr (b) = b
%format IntExpr i = i
%format StringExpr s = s
%format AttrExpr e v = e "." v
%format ListExpr (es) = [es]
%format TupleExpr (es) = (es)
%format DictExpr (dict) = dict
%format SliceExpr x (y) = x"["y"]"
%format BinOpExpr op x y = x op y
%format UnaryOpExpr op x = op x
%format YieldExpr e = "\mathtt{yield}~" e
%format AddrExpr a = a

% operators

%format BinOp = "~BinOp~"
%format UnaryOp = "UnaryOp~"
%%format BinOp = "\text{\textcircled{\small 2}}"
%%format UnaryOp = "\text{\textcircled{\small 1}}"

%format Add = "+"
%format Sub = "-"
%format USub = "-"
%format Not = "\mathtt{not}"
%format Mul = "*"
%format Div = "/"
%format FloorDiv = "//"
%format Mod = "\%"
%format Pow = "**"
%format Eq = "=="
%format NotEq = "!="
%format Lt = "<"
%format LtE = "<="
%format Gt = ">"
%format GtE = ">="
%format Is = "\mathtt{is}"
%format In = "\mathtt{in}"
%format And = "\mathtt{and}"
%format Or = "\mathtt{or}"

% formatting rules for statements
%format ExprStmt e = e
%format (AssStmt (v) (e)) = v "=" e
%format AttrAssStmt e_t v e = e_t "." v "=" e
%format SliceAssStmt e s v = e"["s"] = " v
%format DelStmt v = "\mathtt{del}~" v
%format AttrDelStmt e x = "\mathtt{del}~"e"."x
%format SliceDelStmt x y = "\mathtt{del}~"x[y]
%format DefStmt v pars b = "\mathtt{def}~" v(pars) ":" b
%format RetStmt e = "\mathtt{return}~" e
%format PrintStmt e = "\mathtt{print}"(e)
%format PassStmt = "\mathtt{pass}"
%format IfStmt e bt bf = "\mathtt{if}~" e ":" bt "~\mathtt{else}:" bf
%format WhileStmt e b = "\mathtt{while}~" e ":" b
%format ForStmt x e b = "\mathtt{for}~" x "~\mathtt{in}~" e ":" b
%format ClassStmt (x) (e) (b) = "\mathtt{class}~" x"("e")" ":" b
%format TryStmt (bt) (e) (x) (be) = "\mathtt{try}:" bt "~\mathtt{except}~" e "\comma~" x ":" be
%format FinallyStmt (bt) (bf) = "\mathtt{try}:" bt "~\mathtt{finally}:" bf
%format RaiseStmt (e) = "\mathtt{raise}~" e 
%format ImportStmt n = "\mathtt{import}~" n
%format BreakStmt = "\mathtt{break}"
%format ContinueStmt = "\mathtt{continue}"
%format ExecStmt e env = "\mathtt{exec}~"e"~\mathtt{in}~"env

%format (ConsBlock (s) (b)) = s"\semicolon~"b
%format (EmptyBlock) = "\{\}"

%format (appendStmt (s) (b)) = s"\semicolon~"b

% local formatting rules, for grammar/type declarations
%\definecolor{Gray}{rgb}{0.1,0.1,0.5}
%\newcommand{\token}[1]{{\color{Gray}#1}}
%{
%format data = " "
%format type = " "
%format = = "::="
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Expressions}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

The expression syntax specification indicates no binding preferences. Any ambiguity will be resolved by parenthesis. For example $3+2*5$ will have to be written as $3+(2*5)$.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%format DictExpr tuples = "\{ \langle"Expr : Expr"\rangle^{\ast}\}"
%format Exprs = "\langle"Expr"\rangle^{\ast}"
%format Names = "\langle"Name"\rangle^{\ast}"
%format AddrExpr i = Address
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\begin{code}
data Expr  =  VarExpr Name               -- Variable          ,   see Rule \ref{fw-rewrite-VarExpr}
           |  CallExpr Expr Exprs        -- Function call     ,   see Rule \ref{fw-rewrite-CallExpr}
           |  AttrExpr Expr Name         -- Attribute access  ,   see Rule \ref{fw-rewrite-AttrExpr}
           |  SliceExpr Expr Expr        -- Slice access      ,   see Rule \ref{fw-rewrite-SliceExpr}
           |  BinOpExpr BinOp Expr Expr  -- Binary operator   ,   see Rule \ref{fw-rewrite-BinOpExpr}
           |  UnaryOpExpr UnaryOp Expr   -- Unary operator    ,   see Rule \ref{fw-rewrite-UnaryOpExpr}
           |  YieldExpr Expr             -- Yield expression  ,   see Rule \ref{fw-rewrite-YieldExpr}
           |  IntExpr Int                -- Literal integer   ,   see Rule \ref{fw-rewrite-IntExpr}
           |  BoolExpr Bool              -- Literal boolean   ,   see Rule \ref{fw-rewrite-BoolExpr}
           |  StringExpr String          -- Literal string    ,   see Rule \ref{fw-rewrite-StringExpr}
           |  ListExpr Exprs             -- Literal list      ,   see Rule \ref{fw-rewrite-ListExpr}
           |  TupleExpr Exprs            -- Literal tuple     ,   see Rule \ref{fw-rewrite-TupleExpr}
           |  DictExpr [(Expr, Expr)]    -- Literal dictionary,   see Rule \ref{fw-rewrite-DictExpr}
\end{code}
%if style == newcode
\begin{code}
           |  AddrExpr Int               -- Literal address
  deriving (Eq)
\end{code}
%endif

\begin{code}
data BinOp = Add | Sub | Mul | Div | Mod | Pow | FloorDiv | Eq | NotEq | Lt | LtE | Gt | GtE | Is | In | And | Or
\end{code}
%if style == newcode
\begin{code}
           | LShift | RShift | BitOr | BitXor | BitAnd 
           | NotIn | IsNot 
	deriving (Eq)
\end{code}
%endif

\begin{code}
data UnaryOp = Not | USub
\end{code}
%if style == newcode
\begin{code}
             | Invert | UAdd
	deriving (Eq)
\end{code}
%endif


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\section{Statements}
\label{section:statements-AST}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The command line interface of Python accepts single statements. Full programs as well as modules, consist of a single block of statements --a list of statements-- in a text file.

We often append a semicolon to statements to distinguish statements, and especially expression statements, from expressions. The concrete syntax of Python supports this use of semicolons as well, but does not require it.

Some statements, so called compound statements, contain blocks. In the concrete syntax these blocks are delimited by their indentation level (see Python's reference documentation). In this abstract syntax we delimit blocks using curly brackets where ambiguity arises. Specifically the empty block is denoted as an empty pair of curly brackets.


\begin{code}
data Stmt  =  ExprStmt Expr                  -- expression statement,  see Rule \ref{fw-rewrite-ExprStmt}
           |  AssStmt Name Expr              -- variable assignment ,  see Rule \ref{fw-rewrite-AssStmt}
           |  AttrAssStmt Expr Name Expr     -- attribute assignment,  see Rule \ref{fw-rewrite-AttrAssStmt}
           |  SliceAssStmt Expr Expr Expr    -- slice assignment    ,  see Rule \ref{fw-rewrite-SliceAssStmt}
           |  DelStmt Name                   -- variable deletion   ,  see Rule \ref{fw-rewrite-DelStmt}
           |  SliceDelStmt Expr Expr         -- slice deletion      ,  see Rule \ref{fw-rewrite-SliceDelStmt}
           |  AttrDelStmt Expr Name          -- attribute deletion  ,  see Rule \ref{fw-rewrite-AttrDelStmt}
           |  DefStmt Name Names Block       -- function definition ,  see Rule \ref{fw-rewrite-DefStmt}
           |  RetStmt Expr                   -- return              ,  see Rule \ref{fw-rewrite-RetStmt}
           |  PrintStmt Expr                 -- print               ,  see Rule \ref{fw-rewrite-PrintStmt}
           |  IfStmt Expr Block Block        -- if-then-else        ,  see Rule \ref{fw-rewrite-IfStmt}
           |  WhileStmt Expr Block           -- while               ,  see Rule \ref{fw-rewrite-WhileStmt}
           |  ForStmt Name Expr Block        -- for                 ,  see Rule \ref{fw-rewrite-ForStmt}
           |  ClassStmt Name Expr Block      -- class definition    ,  see Rule \ref{fw-rewrite-ClassStmt}
           |  TryStmt Block Expr Name Block  -- try-catch           ,  see Rule \ref{fw-rewrite-TryStmt}
           |  FinallyStmt Block Block        -- try-finally         ,  see Rule \ref{fw-rewrite-FinallyStmt}
           |  RaiseStmt Expr                 -- raise               ,  see Rule \ref{fw-rewrite-RaiseStmt}
           |  ImportStmt Names               -- import              ,  see Rule \ref{fw-rewrite-ImportStmt}
           |  PassStmt                       -- pass                ,  see Rule \ref{fw-rewrite-PassStmt}
           |  BreakStmt                      -- break               ,  see Rule \ref{fw-rewrite-BreakStmt}
           |  ContinueStmt                   -- continue            ,  see Rule \ref{fw-rewrite-ContinueStmt}
           |  ExecStmt Expr Expr             -- exec                ,  see Rule \ref{fw-rewrite-ExecStmt}
\end{code}
%if style == newcode
\begin{code}
  deriving(Eq)
\end{code}
%endif


Blocks in functions, classes and modules also define a variable scope, the semantics of which are discussed in chapters \ref{chapter:variables}, \ref{chapter:functions}, \ref{chapter:objects}, and \ref{chapter:dynamic-code}. There are some restrictions on the syntax of scoping blocks that are not expressed by the abstract syntax definitions. It could be argued that these are not syntactical restrictions, but Python raises a |SyntaxError| for programs that fail the restrictions:

\begin{itemize}
\item The return and yield statements are limited to the scope of a function. In other words, they can only occur in function definitions or in a while, for, try, or if statement inside the function.
\item A function scope can contain either a return or a yield statement, but not both.
\item The continue and break statements can only occur in for or while loops, or in a nested if or try statement.
\end{itemize}


%if style == newcode
\begin{code}
data Block  =  ConsBlock Stmt Block  |  EmptyBlock
  deriving (Eq)
\end{code}
%endif


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% etc
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%if style == newcode
\begin{code}
type Exprs = [Expr]
type Names = [Name]

type Name = String

instance Show Expr where
  show (VarExpr v) = v
  show (CallExpr e args) = show e ++ "(" 
    ++ intercalate ", " (map show args) ++ ")"
  show (IntExpr i) = show i
  show (BoolExpr b) = show b
  show (AttrExpr e v) = show e ++ "." ++ v
  show (ListExpr es) = show es
  show (TupleExpr []) = "(,)"
  show (TupleExpr [e]) = "(" ++ show e ++ "," ++ ")"
  show (TupleExpr es) = "(" ++ intercalate ", " (map show es) ++ ")"
  show (DictExpr pes) = "{" ++ intercalate ", " (map (\(k,v) -> show k ++ " : " ++ show v) pes) ++ "}"
  show (StringExpr s) = show s
  show (BinOpExpr o x y) = show x ++ " " ++ show o ++ " " ++ show y
  show (UnaryOpExpr o x) = show o ++ show x
  show (SliceExpr lst i) = show lst ++ "[" ++ show i ++ "]"
  show (YieldExpr e) = show "yield " ++ show e
  show (AddrExpr i) = '$' : show i

-- note that these are also used by the parser, so be careful
instance Show BinOp where
  show Add = "+"
  show Sub = "-"
  show Mul = "*"
  show Div = "/"
  show FloorDiv = "//"
  show Mod = "%"
  show Pow = "**"
  show Eq = "=="
  show NotEq = "!="
  show Lt = "<"
  show LtE = "<="
  show Gt = ">"
  show GtE = ">="
  show Is = "is"
  show In = "in"
  show And = "and"
  show Or = "or"
  show BitOr = "|"
  show BitXor = "^"
  show BitAnd = "&"
  show LShift = "<<"
  show RShift = ">>"

-- note that these are also used by the parser, so be careful
instance Show UnaryOp where
  show Invert = "~"
  show Not = "not"
  show UAdd = "+"
  show USub = "-"
\end{code}
%endif


%if style == newcode
\begin{code}
instance Show Stmt where
  show (ExprStmt e) = show e
  show (AssStmt v e) = v ++ " = " ++ show e
  show (AttrAssStmt e_t v e) = show e_t ++ "." ++ v ++ " = " ++ show e
  show (SliceAssStmt e_l e_i e) = show e_l ++ "[" ++ show e_i ++ "] = " ++ show e
  show (DelStmt v) = "del " ++ v
  show (AttrDelStmt e v) = "del " ++ show e ++ "." ++ v
  show (SliceDelStmt e k) = "del " ++ show e ++ "[" ++ show k ++ "]"
  show (DefStmt f pars b) = "def " ++ f ++ "(" 
    ++ intercalate ", " pars ++ ") : " ++ show b
  show (RetStmt e) = "return " ++ show e
  show (PrintStmt e) = "print(" ++ show e ++ ")"
  show (IfStmt e t f) = 
    "if " ++ show e ++ " : " ++ show t 
    ++ " else : " ++ show f
  show (WhileStmt e b) = "while " ++ show e ++ ": " ++ show b
  show (ForStmt x e b) = "for " ++ x ++ " in " ++ show e ++ ": " ++ show b
  show (ClassStmt x e b) = 
    "class " ++ x ++ "(" ++ show e ++ ") :" ++ show b
  show (TryStmt b_t e x b_e) =
    "try : " ++ show b_t ++ " except " ++ show e ++ ", " ++ x ++ " : " ++ show b_e
  show (FinallyStmt b_t b_f) =
    "try : " ++ show b_t ++ " finally : " ++ show b_f
  show (RaiseStmt e) = "raise " ++ show e
  show (ImportStmt xs) = "import " ++ intercalate "." xs
  show (BreakStmt) = "break"
  show (ContinueStmt) = "continue"
  show (PassStmt) = "pass"
  show (ExecStmt x e) = "exec " ++ show x ++ " in " ++ show e
\end{code}
%endif


%if style == newcode
\begin{code}
--instance Lift Block where
--  lift (ConsBlock s b) = [| ConsBlock $(lift s) $(lift b) |]
--  lift (EmptyBlock) = [| EmptyBlock |]

concatBlocks :: Block -> Block -> Block
concatBlocks EmptyBlock b = b
concatBlocks (ConsBlock s bl) br = ConsBlock s (concatBlocks bl br)

appendStmt :: Block -> Stmt -> Block
appendStmt EmptyBlock s = ConsBlock s EmptyBlock
appendStmt (ConsBlock h t) s = ConsBlock h (appendStmt t s)

concatBlocksAround :: Block -> Stmt -> Block -> Block
concatBlocksAround bl s br = concatBlocks bl (ConsBlock s br)

stmtBlock :: Stmt -> Block
stmtBlock s = ConsBlock s EmptyBlock

blockFromList :: [Stmt] -> Block
blockFromList [] = EmptyBlock
blockFromList (s:b) = ConsBlock s (blockFromList b)

instance Show Block where
  show EmptyBlock = "{}"
  show (ConsBlock b EmptyBlock) = "{" ++ show b ++ "}"
  show (ConsBlock b _) = "{" ++ show b ++ "; ..." ++ "}"
\end{code}
%endif
%}
