module PythonParser where

import UU.Parsing
import UU.Scanner.GenToken(position)

import PythonTokenParser
import PythonScanner(scanString, scanFile)
import PythonToken(PythonToken, getTokenLines)

import PythonAST

parseString parser str = 
  do let tokens = scanString "<stdin>" str
     parseTokens parser tokens

parseFile parser file =
  do tokens <- scanFile file
     parseTokens parser tokens

parseTokens = parseIOMessage showMessage

--strictParseString :: (IsParser p PythonToken) => p Block -> String -> Maybe Block
strictParseString parser str =
  case getMsgs steps of
    [] -> Just result
    msgs -> Nothing
  where steps = parse parser (scanString "<stdin>" str)
        (Pair result _) = evalSteps steps

strictParseFile parser file =
  do tokens <- scanFile file
     let steps = parse parser tokens
     let (Pair result _) = evalSteps steps
     case getMsgs steps of
       []   -> return (Just result)
       msgs -> return Nothing

showMessage (Msg expecting mtoken action) =
  let pos = case mtoken of
              Nothing -> "at end of file"
              Just s  -> case action of 
                           Insert _ -> "before \"" ++ show s ++ "\"" ++ show (position s) 
                           Delete t -> "at \"" ++ show t ++ "\"" ++ show (position t) 
  in "\nError      : " ++ pos ++
     "\nExpecting  : " ++ show expecting ++
     "\nRepaired by: " ++ show action ++ "\n"  

-------------------------------------------------------------------------------
-- Expressions
-------------------------------------------------------------------------------
binOpExpr o x y = BinOpExpr x o y
chainBinOp op r = pChainl (BinOpExpr op <$ pKey (show op)) r

pExpr :: (IsParser p PythonToken) => p Expr
pExpr = pOrTest

pOrTest :: (IsParser p PythonToken) => p Expr
pOrTest = foldr chainBinOp pNotTest [Or, And]

pNotTest :: (IsParser p PythonToken) => p Expr
pNotTest =
  pComparison
  <|> UnaryOpExpr Not <$ pKey "not" <*> pNotTest

pComparison :: (IsParser p PythonToken) => p Expr
pComparison = 
  foldr chainBinOp pOrExpr [Lt, LtE, Gt, GtE, Eq, NotEq, Is, In]

pOrExpr :: (IsParser p PythonToken) => p Expr
pOrExpr = foldr chainBinOp pAExpr [BitOr, BitXor, BitAnd, LShift, RShift]

pAExpr :: (IsParser p PythonToken) => p Expr
pAExpr = foldr chainBinOp pUExpr [Add, Sub, Mul, Div, FloorDiv, Mod]

pUExpr :: (IsParser p PythonToken) => p Expr
pUExpr =
  pPower
  <|> UnaryOpExpr USub <$ pKey "-" <*> pUExpr
  <|> UnaryOpExpr Invert <$ pKey "~" <*> pUExpr

pPower :: (IsParser p PythonToken) => p Expr
pPower = foldr chainBinOp pPrimary [Pow]

pPrimary :: (IsParser p PythonToken) => p Expr
pPrimary =
  foldl (\a b -> b a) <$> pAtom <*> pList_ng  pPrimaryTail

pPrimaryTail  :: (IsParser p PythonToken) => p (Expr -> Expr)
pPrimaryTail = 
  pCallTail 
  <|> pAttributeRefTail
  <|> pSliceTail
  
pSliceTail :: (IsParser p PythonToken) => p (Expr -> Expr)
pSliceTail = swapArgs SliceExpr <$ pOBrack <*> pExpr <* pCBrack

pCallTail :: (IsParser p PythonToken) => p (Expr -> Expr)
pCallTail = 
  swapArgs CallExpr <$> pParens (pListSep pComma pExpr)

pAttributeRefTail :: (IsParser p PythonToken) => p (Expr -> Expr)
pAttributeRefTail =
  swapArgs AttrExpr <$ pDot <*> pName

pAtom :: (IsParser p PythonToken) => p Expr
pAtom =
  VarExpr <$> pName
  <|> AddrExpr . read . tail <$> pLitAddr
  <|> IntExpr . read <$> (pDecNumber <|> pOctNumber) -- only octs parse '0'
  <|> StringExpr . read <$> pLitString
  <|> ListExpr <$ pOBrack <*> pListSep pComma pExpr <* pCBrack
  <|> TupleExpr <$ pOParen <*> ((:) <$> pExpr <* pComma <*> pListSep pComma pExpr) <* pCParen
  <|> TupleExpr [] <$ pOParen <* pCParen
  <|> DictExpr <$ pOCurly <*> pListSep pComma ((,) <$> pExpr <* pColon <*> pExpr) <* pCCurly
  <|> pParens pExpr
  <|> pParens (YieldExpr <$ pKey "yield" <*> pExpr)

swapArgs :: (a -> b -> c) -> (b -> a -> c)
swapArgs f x y = f y x
-------------------------------------------------------------------------------
-- Statements
-------------------------------------------------------------------------------
pStmtLine :: (IsParser p PythonToken) => p Stmt
pStmtLine = pStmt <* pNewline

pStmt :: (IsParser p PythonToken) => p Stmt
pStmt =
  ExprStmt <$> pExpr
  <|> AssStmt <$> pName <* pKey "=" <*> pExpr
  <|> AttrAssStmt <$> pExpr <* pDot <*> pName <* pKey "=" <*> pExpr
  <|> SliceAssStmt <$> pExpr <* pOBrack <*> pExpr <* pCBrack <* pKey "=" <*> pExpr
  <|> DelStmt <$ pKey "del" <*> pName
  <|> AttrDelStmt <$ pKey "del" <*> pExpr <* pDot <*> pName
  <|> SliceDelStmt <$ pKey "del" <*> pExpr <* pOBrack <*> pExpr <* pCBrack
  <|> RetStmt <$ pKey "return" <*> pExpr
  <|> yieldStmt <$ pKey "yield" <*> pExpr
  <|> PrintStmt <$ pKey "print" <*> pParens pExpr
  <|> RaiseStmt <$ pKey "raise" <*> pExpr
  <|> ImportStmt <$ pKey "import" <*> pList1Sep pDot pName
  <|> PassStmt <$ pKey "pass"
  <|> BreakStmt <$ pKey "break"
  <|> ContinueStmt <$ pKey "continue"
  <|> ExecStmt <$ pKey "exec" <*> pOrExpr <* pKey "in" <*> pExpr

yieldStmt = ExprStmt . YieldExpr
passBlock = ConsBlock PassStmt EmptyBlock

pCompStmt :: (IsParser p PythonToken) => p Stmt
pCompStmt = 
  DefStmt <$ pKey "def" <*> pName <*> pParens (pListSep pComma pName) 
          <* pColon <* pNewline <*> pSuite
  <|> IfStmt <$ pKey "if" <*> pExpr <* pColon <* pNewline <*> pSuite
             <*> opt (pKey "else" *> pColon *> pNewline *> pSuite) passBlock
  <|> WhileStmt <$ pKey "while" <*> pExpr <* pColon <* pNewline <*> pSuite
  <|> ForStmt <$ pKey "for" <*> pName <* pKey "in" <*> pExpr <* pColon <* pNewline <*> pSuite
  <|> ClassStmt <$ pKey "class" <*> pName <*> pParens (TupleExpr <$> pList1Sep pComma pExpr) <* pColon <* pNewline
                <*> pSuite
  <|> TryStmt <$ pKey "try" <* pColon <* pNewline
              <*> pSuite
              <* pKey "except" <*> pExpr <* pComma <*> pName <* pColon <* pNewline <*> pSuite
  <|> FinallyStmt <$ pKey "try" <* pColon <* pNewline
                  <*> pSuite
                  <* pKey "finally" <* pColon <* pNewline 
                  <*> pSuite
--  <|> TryStmt <$ pKey "try" <* pColon <* pNewline
--              <*> pSuite
--              <*> pList (pKey "except" *> pExpr <* pComma <*> pName <* pColon <* pNewline <*> pSuite)
--              <*> opt (pKey "else" *> pColon *> pNewline *> pSuite) EmptyBlock
--              <*> opt (pKey "finally" *> pColon *> pNewline *> pSuite) EmptyBlock


-------------------------------------------------------------------------------
-- Blocks
-------------------------------------------------------------------------------

pSuite :: (IsParser p PythonToken) => p Block
pSuite = pIndent *> pBlock <* pDedent

pBlock :: (IsParser p PythonToken) => p Block
pBlock =
  pFoldr (ConsBlock, EmptyBlock) (pStmtLine <|> pCompStmt)



-------------------------------------------------------------------------------
-- expressions
-------------------------------------------------------------------------------

