module PythonScanner where

import PythonTokenizer
import PythonToken

import UU.Scanner.GenToken
import UU.Scanner.Position

--TODO: remove newlines between concaternated strings (should not be indented)

scanFile :: String -> IO [PythonToken]
scanFile filename =
  do str <- readFile filename
     return (scanString filename str)

scanString :: String -> String -> [PythonToken]
scanString posFilename = id
  . dropLeadingNewlines
  . replaceIndents
  . dropDelimWhitespace '{' '}'
  . dropDelimWhitespace '[' ']'
  . dropDelimWhitespace '(' ')'
  . dropWhiteLines
  . dropNonIndentWS . (ValToken Newline "" noPos:)
  . dropEscapedNewlines
  . dropComments
  . addPositions (initPos posFilename)
  . alexScanTokens
  . (++ "\n")


dropLeadingNewlines :: [PythonToken] -> [PythonToken]
dropLeadingNewlines (ValToken Newline _ _:tl) =
  dropLeadingNewlines tl
dropLeadingNewlines tl = tl

addPositions :: Pos -> [Pos -> PythonToken] -> [PythonToken]
addPositions _ [] = []
addPositions pos (hd:tl) =
  case hd pos of
    Reserved k p -> Reserved k p : addPositions (updPosStr k p) tl
    ValToken t v p -> ValToken t v p : addPositions (updPosStr v p) tl


updPosStr :: String -> Pos -> Pos
updPosStr str p = foldr updPos p str

replaceIndents :: [PythonToken] -> [PythonToken]
replaceIndents = replaceIndents' [0]

-- assuming all non indent whitespace has been removed
replaceIndents' :: [Int] -> [PythonToken] -> [PythonToken]
replaceIndents' depth [] = []
replaceIndents' depth toks@(ValToken Whitespace v p:tl)
  | indent > head depth  =
      ValToken Indent "" p : replaceIndents' (indent:depth) tl
  | indent < head depth  =
      ValToken Dedent "" p : replaceIndents' (tail depth) toks
  | otherwise            =
      replaceIndents' depth tl
  where indent = indentSize v
replaceIndents' depth (hd@(ValToken Newline _ p):tl) -- nl followed by no ws
  | head depth > 0 && hdIsNoWs tl  =
      hd : map (const $ ValToken Dedent "" p) (tail depth) 
           ++ replaceIndents' [0] tl
  | otherwise                      =
      hd : replaceIndents' depth tl
  where hdIsNoWs ((ValToken Whitespace _ _):_) = False
        hdIsNoWs _ = True
replaceIndents' depth (hd:tl) = hd : replaceIndents' depth tl

indentSize :: String -> Int
indentSize = cnt 0
  where cnt c [] = c
        cnt c (' ':tl) = cnt (c+1) tl
        cnt c ('\t':tl) = cnt (8 * (1 + div c 8)) tl

dropNonIndentWS :: [PythonToken] -> [PythonToken]
dropNonIndentWS [] = []
dropNonIndentWS (ValToken Newline vnl pnl:ValToken Whitespace v p:tl) =
  ValToken Newline vnl pnl : ValToken Whitespace v p : dropNonIndentWS tl
dropNonIndentWS (ValToken Whitespace v p:tl) =
  dropNonIndentWS tl
dropNonIndentWS (hd:tl) = hd : dropNonIndentWS tl

dropWhiteLines :: [PythonToken] -> [PythonToken]
dropWhiteLines [] = []
dropWhiteLines (ValToken Newline _ _
                : ValToken Whitespace _ _
                : ValToken Newline v p
                : tl) =
  dropWhiteLines (ValToken Newline v p : tl)
dropWhiteLines (ValToken Newline _ _
                : ValToken Newline v p
                : tl) =
  dropWhiteLines (ValToken Newline v p : tl)
dropWhiteLines (hd:tl) =
  hd : dropWhiteLines tl

dropEscapedNewlines :: [PythonToken] -> [PythonToken]
dropEscapedNewlines [] = []
dropEscapedNewlines (Reserved "\\" _:ValToken Newline _ _:tl) =
  dropEscapedNewlines tl
dropEscapedNewlines (hd:tl) = hd : dropEscapedNewlines tl

dropComments :: [PythonToken] -> [PythonToken]
dropComments = filter matchComment
  where matchComment (ValToken Comment _ _) = False
        matchComment _ = True

dropDelimWhitespace :: Char -> Char -> [PythonToken] -> [PythonToken]
dropDelimWhitespace oDelim cDelim toks = fst (trackDelim toks)
  where
    trackDelim :: [PythonToken] -> ([PythonToken], Int)
    trackDelim [] = ([],0)
    trackDelim (hd@(Reserved [c] _):tl)
      | c == oDelim  = (hd : if closes > 0 then filteredTl else tl', closes - 1)
      | c == cDelim  = (hd : tl', closes + 1)
      | otherwise    = (hd : tl', closes)
      where (tl',closes) = trackDelim tl
            filteredTl = filterUntilClosed 0 tl'
    trackDelim (hd:tl) = (hd:tl', closes)
      where (tl',closes) = trackDelim tl

    -- TODO: this function can be applied many times on the same tokens
    -- in case of nested parenthesis. This could be prevented...
    filterUntilClosed :: Int -> [PythonToken] -> [PythonToken]
    filterUntilClosed _ [] = []
    filterUntilClosed i (hd@(Reserved [c] _) : tl)
      | c == cDelim  = hd : if i > 0 then filterUntilClosed (i-1) tl else tl
      | c == oDelim  = hd : filterUntilClosed (i+1) tl
      | otherwise    = hd : filterUntilClosed i tl
    filterUntilClosed i (ValToken Newline _ _ : tl) = filterUntilClosed i tl
    filterUntilClosed i (ValToken Whitespace _ _ : tl) = filterUntilClosed i tl
    filterUntilClosed i (hd:tl) = hd : filterUntilClosed i tl

