%if style == newcode
\begin{code}
module Variables(rewrite) where
import VirtualMachine
\end{code}
%endif

%{

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Variables} \label{chapter:variables}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Variables are references to values on the heap. Variables can be assigned, evaluated (dereferenced), and deleted. Because variables are references, assignments and deletions change only the variable binding, not the values themselves. Multiple values can also refer to the same value. The program in Listing~\ref{example:variables} shows how a references can be used to share values.

\begin{lstlisting}[ caption={Assignments change the variable but not the value}
                  , label=example:variables ]
x = [1,2]    # assigning a variable with a list
y = x        # aliasing the variable
x.append(3)  # append 3 to the list
print(y)     # prints [1,2,3]
x = []       # reassigning x to the empty list
print(y)     # prints [1,2,3]
z = y
del y        # removes the binding for y
print(z)     # prints [1,2,3]
\end{lstlisting}


\section{Assignment statements}

\rulesection{Executing an assignment statement}
\label{fw-rewrite-AssStmt}
Executing the assignment~|AssStmt x e| starts with the evaluation of |e|. A continuation frame~|AssFrame x|, holding the variable name~|x|, is placed on the stack.

\begin{code}
rewrite  (State  heap  envs  ( stack                )  ( FwStmt (AssStmt x e) )) =
         (state  heap  envs  ( stack :|: AssFrame x )  ( FwExpr e             ))
\end{code}

The circle~$\circ$ in the stack frame~|AssFrame x| serves as a placeholder for the expression~|e| that is being evaluated.


\rulesection{Binding a variable}
When the evaluation of |e| is completed, the resulting address~|a| is stored in the topmost environment mapping. Any previous binding of~|x| in the mapping~|get heap env_1| will be overwritten.

\begin{code}
rewrite  (State  ( heap                                              )  ( envs :|: env_1 )  ( stack :|: AssFrame x )  ( BwResult a      )) =
         (state  ( heap <+> [env_1 |-> get heap env_1 <+> [x |-> a]] )  ( envs :|: env_1 )  ( stack                )  ( BwResult a_None ))
\end{code}


\section{Variable expressions}

\rulesection{Evaluating a variable expression}
\label{fw-rewrite-VarExpr}
Evaluating a variable in general amounts to a simple lookup in the environment: if the variable is bound in the environment, we return it, otherwise an exception is raised.

Local variables (see Section \ref{section:functions-local-variables}), have slightly different semantics. A local variable can only be retrieved from the local environment, which is topmost in the environment stack.

Thus, the evaluation of a variable has four distinct cases:
\begin{enumerate}
\item If |x| is a local variable and is bound in the local environment~|get heap env_1|, we lookup and return the address of |x|.
\item An |UnboundLocalError| exception is raised if |x| is a local variable but is unbound in the local environment.
\item If |x| not a local variable and is bound in the full environment~|getFullEnv (envs :||: env_1) heap|, we lookup and return the address of |x|.
\item Otherwise the variable is unbound and not local, so we raise a |NameError|.
\end{enumerate}

%format a_e
%format a_u
%format a_n

\begin{code}
rewrite                                                                               (State  heap  ( envs :|: env_1 )  stack  ( FwExpr (VarExpr x)                                  ))
  | x `elem` localVars heap env_1   && x `isin` getEnvValue heap env_1             =  (state  heap  ( envs :|: env_1 )  stack  ( BwResult (get (getEnvValue heap env_1) x)           ))
  | x `elem` localVars heap env_1   && x `isnotin` getEnvValue heap env_1          =  (state  heap  ( envs :|: env_1 )  stack  ( raise a_UnboundLocalError []                        ))
  | x `nelem` localVars heap env_1  && x `isin` getFullEnv (envs :|: env_1 ) heap  =  (state  heap  ( envs :|: env_1 )  stack  ( BwResult (get (getFullEnv (envs :|: env_1) heap) x) ))
  | otherwise                                                                      =  (state  heap  ( envs :|: env_1 )  stack  ( raise a_NameError []                                ))
\end{code}


\section{Delete statements}

\rulesection{Executing a variable deletion}
\label{fw-rewrite-DelStmt}
The delete statement~|DelStmt x| removes the variable~|x| from the local environment |env_1|. As for the variable evaluation, local variables have a special case:
\begin{enumerate}
\item If |x| is bound in the local environment, we delete it.
\item An |UnboundLocalError| is raised if |x| is a local variable and not (yet) bound in the local environment.
\item Otherwise the variable must be unbound and not local, so we raise a |NameError|.
\end{enumerate}

\begin{code}
rewrite                                                                     (State  heap   ( envs :|: env_1 )  stack  ( FwStmt (DelStmt x)           ))
  | x `isin` getEnvValue heap env_1                                      =  (state  heap'  ( envs :|: env_1 )  stack  ( BwResult a_None              ))
  | x `isnotin` getEnvValue heap env_1 && x `elem` localVars heap env_1  =  (state  heap   ( envs :|: env_1 )  stack  ( raise a_UnboundLocalError [] ))
  | otherwise                                                            =  (state  heap   ( envs :|: env_1 )  stack  ( raise a_NameError []         ))
   where  heap'  = heap <+> [env_1 |-> get heap env_1 <-> x]
\end{code}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%if style==newcode
\begin{code}
rewrite _ = skipState
\end{code}
%endif

%}
