%if style == newcode
\begin{code}
module VirtualMachine
  ( State(..)
  , Instr(..)
  , raise
  , Env
  , getFullEnv
  , localVars
  , Frame(..)
  , CallStack()
  , Heap(..)
  , freeAddr
  , Value(..)
  , Object
  , isObjValue
  , getBoolValue
  , getEnvValue
  , getIntValue
  , getStringValue
  , getFuncValue
  , getDictValue
  , getListValue
  , getTupleValue
  , getListItValue
  , state
  , skipState
  , Rewriter
  , module Result
  , module PythonAST
  , module Preliminaries
  )
where
import qualified Data.Map as M
import qualified Control.Monad.State as S

import Preliminaries
import PythonAST
import Result
import RewriterMonad
\end{code}
%endif


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\chapter{Transforming an abstract machine} \label{chapter:abstract-machine}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The operational semantics are defined by state transitions of an abstract machine. A machine state~$\bigl\langle \Theta, \Gamma, S, \iota \bigr\rangle$ consist of a heap~$\Theta$, an environment stack~$\Gamma$, a control stack~$S$, and an instruction~$\iota$. State transitions are defined by rewrite rules that transform the machine state. These rewrite rules have the following shape.

\[ 
\bigl\langle \Theta, \Gamma, S, \iota \bigr\rangle
\Rightarrow
\bigl\langle \Theta', \Gamma', S', \iota' \bigr\rangle 
\]

%There are three different kinds of rewrite rules: forward, backward and rewind rules. Forward rules transform the heap, stack and environment according to the instruction. Backward rules inspect the head frame on the stack to determine what to do with the instruction. Rewind rules are a special backward rules that unwind the stack i.e., repeatedly pop stack frames, until a specific frame is found.

As the semantics of Python are deterministic, there is one, and only one, rule for each machine state. The states are discerned by pattern matches in the rules. In some cases however, pattern matches overlap. 
%For example Rule \ref{bw-rewrite-BlockFrame} and Rule \ref{bw-rewrite-EmptyBlock} have overlapping patterns: both rules apply to a state $\bigl\langle \Theta, \Gamma, S | {}, a \bigr\rangle$. 
In these cases, the most specific pattern takes precedence over the others.

%if style == newcode
\begin{code}
data State = State Heap Env CallStack Instr
\end{code}

\begin{code}
instance Show State where
  show (State heap env stack step) = show step ++ " [ " ++ show stack ++ " ]"
\end{code}
%endif

%if style==newcode
\begin{code}
type Rewriter = State -> RewriterM (Maybe State)

state :: Heap -> Env -> CallStack -> Instr -> RewriterM (Maybe State)
state heap env stack a = return (Just $ State heap env stack a)

skipState :: RewriterM (Maybe State)
skipState = return Nothing
\end{code}
%endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsubsection{Heap}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The heap is a mapping of addresses to values. Addresses are represented by natural numbers. Values include integers, strings, functions and objects.

%As a mapping, the heap is transformed using the operations introduced in Section \ref{section:mappings}. 

%Some values, like integers and strings, are immutable; once placed on the heap, their values do not change. Immutable values do not need to be stored on the heap, but storing them on the heap makes the semantics more uniform.

%format heap = "\Theta"
%format heap'
%format a_1
%format a_2
%format a_3
%format a_4
For example, a heap |heap = [a_1 ||-> "spam", a_2 ||-> 3, a_3 ||-> [a_1, a_2]]|, contains a string, a number, and a list at the addresses |a_1|, |a_2|, and |a_3| respectively. A new heap |heap' = heap <+> [a_1 ||-> "eggs", a_4 ||-> 42]| is defined as an update of the heap |heap|: the string at |a_1| is redefined and the value $42$ is added at address |a_4|.

Values are never removed from the stack. Because of this, the operational semantics presented here will never define a practical interpreter. In a full implementation of Python, unused values will be removed from the heap by a garbage collector.

%format heap'
%format heap''

%format freeAddr heap = "\text{new address}" "\notin" heap

%format (getEnvValue heap (a)) = heap(a)
%format (getIntValue heap (a)) = heap(a)
%format (getStringValue heap (a)) = heap(a)
%format (getListValue heap (a)) = heap(a)
%format (getListItValue heap (a)) = heap(a)
%format (getBoolValue heap (a)) = heap(a)
%format (getFuncValue heap (a)) = heap(a)
%format (getTupleValue heap (a)) = heap(a)
%format (getDictValue heap (a)) = heap(a)

%format (isObjValue (v)) = v "\text{~is object}"


%if style == newcode
\begin{code}
data Heap     =  Heap (M.Map Int Value)
\end{code}

\begin{code}
instance Show Heap where
  show (Heap heap) = show heap

instance Mapping Heap Result Value where
  rm (Heap m) (Addr k) = error "Only the garbage collector may remove values from the heap!"--Heap (M.delete k m)
  get (Heap m) (Addr k) = m M.! k
  get (Heap m) (Inst _ k) = get (Heap m) k
  get (Heap m) (Typed k _) = get (Heap m) k
  isin (Addr k) (Heap m) = M.member k m
  new = Heap . M.fromList . map (\(Addr k, v) -> (k,v))
  union (Heap x) (Heap y) = Heap (M.union y x)

freeAddr :: Heap -> Result 
freeAddr (Heap heap) = Addr (fst (M.findMax heap) + 1)

getValues :: [Result] -> Heap -> Values
getValues as heap = map (\a -> getValue a heap) as
  where getValue (Addr a) (Heap heap) =
          case M.lookup a heap of
            Just v -> v

getBoolValue :: Heap -> Result -> Bool
getBoolValue (Heap heap) (Addr a) =
  case M.lookup a heap of
    Just (BoolValue b) -> b

getFuncValue :: Heap -> Result -> Value
getFuncValue (Heap heap) (Addr a) =
  case M.lookup a heap of
    Just (ObjValue _ f@(FuncValue _ _ _)) -> f
    Just f@(FuncValue _ _ _) -> f

getDictValue :: Heap -> Result -> Dict
getDictValue (Heap heap) (Addr a) =
  case M.lookup a heap of
    Just (DictValue b) -> b
    Just (ObjValue _ (DictValue d)) -> d

getEnvValue :: Heap -> Result -> M.Map Name Result
getEnvValue (Heap heap) (Addr a) =
  case M.lookup a heap of
    Just (FuncEnvValue b _) -> b
    Just (DictValue b) -> b
    Just (ObjValue _ (DictValue d)) -> d
    Just (ObjValue b NoneValue) -> b
    Just (ModValue b) -> b

getStringValue :: Heap -> Result -> String
getStringValue (Heap heap) (Addr a) =
  case M.lookup a heap of
    Just (StringValue i) -> i
    Just (ObjValue _ (StringValue s)) -> s

getIntValue :: Heap -> Result -> Int
getIntValue (Heap heap) (Addr a) =
  case M.lookup a heap of
    Just (IntValue i) -> i
    Just (ObjValue _ (IntValue i)) -> i

getListValue :: Heap -> Result -> [Result]
getListValue (Heap heap) (Addr a) =
  case M.lookup a heap of
    Just (ListValue lst) -> lst
    Just (ObjValue _ (ListValue l)) -> l

getTupleValue :: Heap -> Result -> [Result]
getTupleValue (Heap heap) (Addr a) =
  case M.lookup a heap of
    Just (TupleValue lst) -> lst
    Just (ObjValue _ (TupleValue l)) -> l

getListItValue :: Heap -> Result -> (Result, Int)
getListItValue (Heap heap) (Addr a) =
    case M.lookup a heap of
      Just (ListItValue lst i) -> (lst, i)
\end{code}
%endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% values
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% formatting rules for Values
%format FuncValue ps b e = "\lambda" ps ".~" b "\dashv" e
%format GenValue (e) (s) = " \langle " e, s " \rangle "
%format SGenValue = " \langle \rangle "
%format IntValue i = i
%format BoolValue b = b
%format (ObjValue (m) (p)) = (m, p)
%format StringValue s = s
%format ListValue (as) = [as]
%format TupleValue (as) = (as)
%format DictValue (d) = d
%format ModValue (m) = m
%format NoneValue = "\mathtt{None}"
%format (ListIterator (a) (b)) = (a,b)

%if style == newcode
\begin{code}
data Value
  =  FuncValue [Name] Block Env
  |  GenValue CallStack Env
  |  SGenValue
  |  IntValue Int
  |  BoolValue Bool
  |  StringValue String
  |  ListValue [Result]
  |  ListItValue Result Int
  |  TupleValue [Result]
  |  DictValue Dict
  |  ObjValue Object Value
  |  ModValue Object
  |  FuncEnvValue Object [Name]
  |  NoneValue
  deriving (Show, Eq)
\end{code}
%endif

%if style == newcode
\begin{code}
type Values = [Value]

instance Mapping Value Name Result where
  rm (ObjValue o p) k = ObjValue (rm o k) p
  rm (FuncEnvValue e vs) k = FuncEnvValue (rm e k) vs
  rm (DictValue d) k = DictValue (rm d k)
  rm (ModValue m) k = ModValue (rm m k)
  rm v k = error ("Trying to remove attribute " 
                  ++ show v ++ " from readonly value "
                  ++ show v)
  get v k = get (valToMap v) k
  isin k v = isin k (valToMap v)
  new lst = ObjValue (new lst) NoneValue --TODO: error "New undefined for values"
  insert x lst = case x of
    DictValue d -> DictValue $ union d (new lst)
    ModValue m -> ModValue $ union m (new lst)
    FuncEnvValue e vs -> FuncEnvValue (union e (new lst)) vs
    ObjValue o p -> ObjValue (union o (new lst)) p
    otherwise -> error "Value is no mapping!"
  union x y = error "Union undefined for values"

valToMap :: Value -> Object
valToMap val = case val of
  ObjValue o _ -> o
  ModValue m -> m
  FuncEnvValue e vs -> e
  DictValue d -> d
  otherwise -> error "Value is no mapping!"
\end{code}
%endif

%if style == newcode
\begin{code}
isFunc :: Heap -> Result -> Bool
isFunc heap (Typed a _) = isFunc heap a || error "isFunc encountered typed nonfunction"
isFunc heap (Prim _ _) = True
isFunc heap (Inst _ a) = isFunc heap a || error "isFunc encountered instantiated nonfunction"
isFunc heap a = case get heap a of
  FuncValue _ _ _ -> True
  ObjValue _ (FuncValue _ _ _) -> True
  otherwise -> False

isNotFunc heap a = not (isFunc heap a)
\end{code}
%endif


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Object value
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%if style == newcode
\begin{code}
type Object   =  M.Map Name Result
\end{code}

\begin{code}
isObjValue (ObjValue _ _) = True
isObjValue _ = False

instance Mapping (M.Map Name Result) Name Result where
  rm x v = M.delete v x
  get = (M.!)
  isin = M.member
  new = M.fromList
  union x y = M.union y x
\end{code}
%endif


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Dict value
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%\todo{the dictionary actually has type M.Map Result Result}

%if style == newcode
\begin{code}
type Dict     =  M.Map Name Result
\end{code}
%endif



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsubsection{Environment}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The environment is a stack of addresses. The addresses point to environment mappings on the heap, each of which contains the bound variables in a single scoping block. Environment mappings are values that implement a mapping of variable names -represented by strings- to addresses.

%format (getFullEnv (envs) (heap)) = "\Sigma^{"heap"}_{"envs"}"
%format envs = "\Gamma"
%format envs'
%format envs''
%format env = "\gamma"
%format env_0
%format env_1
%format env_2
%format env_c
%format localVars heap result = heap(result)"_{loc}"

%- indirection
%- getfullenv
%- locals

The environment mappings on an environment stack can be merged to create the environment mapping |getFullEnv envs heap|, that contains all variable bindings of the environment mappings on the environment stack. The merged environment mapping |getFullEnv envs heap| is defined as:

\begin{code}
getFullEnv EmptyStack heap      = empty
getFullEnv (envs :|: env) heap  = getFullEnv envs heap `union` getEnvValue heap env 
\end{code}

%format eHeap = "\Theta"
%format eEnvs = "\Gamma"
For example, the environment |eEnvs| with a heap |eHeap| binds the variables $x$ and $y$ to the values $1$ and $2$ respectively. The variable |x| is bound in both |env_1| and |env_2|, but the binding in |env_2| shadows the binding in |env_1|, i.e., |get (getFullEnv eEnvs eHeap) x = a_2|.

\begin{code}
eEnvs =  EmptyStack :|: env_1 :|: env_2
eHeap =  [  a_1    |-> IntValue 1
         ,  a_2    |-> IntValue 2
         ,  env_1  |-> new ["x" |-> a_2, "y" |-> a_2]
         ,  env_2  |-> new ["x" |-> a_1]
         ]
\end{code}

%if style == newcode
\begin{code}
a_1 = undefined
a_2 = undefined
env_1 = undefined
env_2 = undefined
\end{code}

\begin{code}
type Env = Stack Result
getFullEnv :: Env -> Heap -> M.Map Name Result
\end{code}

\begin{code}
localVars :: Heap -> Result -> [Name]
localVars (Heap heap) (Addr a) =
  case M.lookup a heap of
    Just (FuncEnvValue _ vs) -> vs
    otherwise                -> []
\end{code}
%endif


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsubsection{Control stack}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The control stack is a stack of continuation frames, i.e., frames that indicate `what to do next'. Many frames use a placeholder~$\circ$, to indicate what part of an instruction is currently being executed or evaluated.

%format AssFrame v = v "=" "\circ"
For example, the stack~|EmptyStack :||: AssFrame x| consists of a single frame~|AssFrame x| on top of the empty stack~|EmptyStack|. The circle in the frame~|AssFrame x| replaces the assignment's right hand side to indicate that it is being evaluated. Once the right hand side expression has been evaluated, the assignment is executed.

%The stack $S$ of the virtual machine state is a list of continuation frames. An empty stack is denoted as $\Box$. A stack with head element |frame| and tail |S| is presented as |S :||: F|. A great number of different frames can be placed on the stack. Each frame will be described when it is first used.

%Traditionally a stack is transformed using the push and pop operations. In this document these transformations are only performed indirectly using pattern matches.

\newcommand{\marker}[0]{\!\!\Rsh\!\!}

%format . = "."
%format BlockFrame (b) = b
%format CallFrame        (es) = "\circ"(es)
%format ArgFrame a (as) (es) = a(as", \circ, "es)
%format FuncFrame p (as) = p(as)
%format ReturnMarker env = env "\vdash \circ_{\lambda}"
%format GenMarker env a =  env "\vdash" a
%format ReturnFrame = "\mathtt{return}~\circ"
%format AttrAssFrame1 e x = e"."x "= \circ"
%format AttrAssFrame2 x a = "\circ." x "=" a
%format AttrDelFrame x = "\mathtt{del}~\circ."x
%format IfFrame1 bt bf = "\mathtt{if}~\circ".__nonzero__() ":" bt "~\mathtt{else}:" bf
%format IfFrame2 bt bf = "\mathtt{if}~\circ:" bt "~\mathtt{else}:" bf
%format WhileFrame1 e bw = "\mathtt{while}~\circ_{"e"}".__nonzero__()" :" bw
%format WhileFrame2 e bw = "\mathtt{while}~\circ_{"e"} :" bw
%format WhileMarker e b = "\mathtt{while}~"e": \circ_{"b"}"
%format ForFrame1 x b = "\mathtt{for}~"x"~\mathtt{in}~\circ".__iter__()" :"b
%format ForFrame2 x b = "\mathtt{for}~"x"~\mathtt{in}~\circ :"b
%format ForFrame3 x a b = "\mathtt{for}~"x"~\mathtt{in}~"a":"b
%format ForMarker x a b = "\mathtt{for}~"x"~\mathtt{in}~"a":\circ_{"b"}"
%format AttrFrame v = "\circ." v
%format PrintFrame1 = "\mathtt{print}"("\circ".__str__())
%format PrintFrame2 = "\mathtt{print}(\circ)"
%format ClassFrame1 v b = "\mathtt{class}~" v "(\circ):" b
%format ClassFrame2 (x) (a) (env) = env "\vdash \mathtt{class}~" x (a)": \circ"
%format ClassConstrFrame x ax ab ac = "\marker_{"ac.__class__.__new__("\cdot", ax, ab, ac)"}"
%format ClassConstrMarker x ao = "\marker_{"ao.__class__.__init__("\cdots")"}"
%format RaiseFrame = "\mathtt{raise}~\circ"
%format ExceptMarker e x b = "\mathtt{except}~" e "\comma~" x ":" b
%format FinallyMarker b = "\mathtt{finally}:" b
%format RaiseContFrame a = "\marker^{\mathtt{raise}~"a"}_{\mathtt{finally}:\circ}"
%format CatchingFrame a x b = "\marker^{\mathtt{raise}~"a"}_{\mathtt{except}~\circ\comma~"x":" b"}"
%format LeftBinOpFrame op e = "\circ" op e
%format RightBinOpFrame a op = a op "\circ"
%format LeftBoolOpFrame op e = "\circ" op"_{B}" e
%format UnaryOpFrame op = op "\circ"

%format ContSliceFrame k = "\circ"[k]
%format KeySliceFrame a = a["\circ"]
%format ValueSliceAssFrame c k = c[k] "= \circ"
%format ContSliceAssFrame k a = "\circ" [k] "=" a
%format KeySliceAssFrame c v = c ["\circ"] "=" v
%format ContSliceDelFrame k = "\mathtt{del}~\circ"[k]
%format KeySliceDelFrame a = "\mathtt{del}~"a["\circ"]

%format ExecFrame1 e = "\mathtt{exec}~\circ~\mathtt{in}~"e
%format ExecFrame2 a = "\mathtt{exec}~"a"~\mathtt{in}~\circ"
%format ExecMarker env = env "\vdash \mathtt{exec}~\circ"
%format ImportFrame (env) xs ys = env "\vdash \mathtt{import}~"xs."\!\circ\!".ys
%format YieldFrame = "\mathtt{yield}~\circ"
%format (ListFrame (as) (es)) = [as, "\circ", es]
%format (TupleFrame (as) (es)) = (as, "\circ", es)
%format (DictKeyFrame (d) (e) (eps)) = "\{"d, "\circ:" e, eps"\}"
%format (DictValFrame (d) (a) (eps)) = "\{"d, a ":\circ", eps"\}"
%format InstrFrame (i) = i

%format frame = "f"
%format stack = "S"


%if style == newcode
\begin{code}
type CallStack = Stack Frame

data Frame  =  BlockFrame Block
            |  CallFrame [Expr]
            |  ArgFrame Result [Result] [Expr]
            |  ReturnMarker Env
            |  ReturnFrame
            |  AssFrame Name
            |  AttrAssFrame1 Expr Name
            |  AttrAssFrame2 Name Result
            |  AttrDelFrame Name
            |  IfFrame1 Block Block
            |  IfFrame2 Block Block
            |  WhileFrame1 Expr Block
            |  WhileFrame2 Expr Block
            |  WhileMarker Expr Block
            |  ForFrame1 Name Block
            |  ForFrame2 Name Block
            |  ForFrame3 Name Result Block
            |  ForMarker Name Result Block
            |  AttrFrame Name
            |  PrintFrame1
            |  PrintFrame2
            |  ClassFrame1 Name Block
            |  ClassFrame2 Name Result Env
            |  ClassConstrFrame Name Result Result Result
            |  ClassConstrMarker Name Result
            |  RaiseFrame
            |  FinallyMarker Block
            |  ExceptMarker Expr Name Block
            |  CatchingFrame Result Name Block
            |  ListFrame [Result] [Expr]
            |  TupleFrame [Result] [Expr]
            |  DictKeyFrame Dict Expr [(Expr, Expr)]
            |  DictValFrame Dict Result [(Expr, Expr)]
            |  ContSliceFrame Expr
            |  KeySliceFrame Result
            |  ValueSliceAssFrame Expr Expr
            |  ContSliceAssFrame Expr Result
            |  KeySliceAssFrame Result Result
            |  ContSliceDelFrame Expr
            |  KeySliceDelFrame Result
            |  ExecFrame1 Expr
            |  ExecFrame2 Result
            |  ExecMarker Env
            |  ConstrFrame [Result]
            |  ConstrMarker Result
            |  ImportFrame Env [Name] [Name]
            |  YieldFrame
            |  GenMarker Env Result
            |  LeftBinOpFrame BinOp Expr
            |  LeftBoolOpFrame BinOp Expr
            |  RightBinOpFrame Result BinOp
            |  UnaryOpFrame UnaryOp
            |  NotFrame
            |  InstrFrame Instr
            deriving(Show,Eq)
\end{code}
%endif

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsubsection{Instruction}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The instruction is a kind of program counter for the virtual machine. There are three different kinds of instruction: The instruction can be the expression, statement, or block that is being executed; it can be the result of the previously executed instruction; or an unwind instruction.
%The latter two instructions depend on the control stack to determine what happens next.

% steps
%format FwBlock (b) = b
%format FwStmt (s) = s";"
%format FwExpr (e) = e
%format FwApp (r) (args) = r(args)
%format BwResult (r) = r
%format Raise (a) = "\mathtt{raise}~" a
%%format raise a es = FwStmt (RaiseStmt (CallExpr (AddrExpr a) es))
%format raise a (es) = "\mathtt{raise}~" a()
%format Return (a) = "\mathtt{return}~" a
%format Yield (a) (stack) = "\mathtt{yield}~" a ",~" stack

%if style == newcode
\begin{code}
data Instr 
  = FwBlock Block | FwStmt Stmt | FwExpr Expr | FwApp Result [Result]
  | BwResult Result | Raise Result | Return Result | Yield Result CallStack
  deriving (Eq)

raise (Addr a) es = FwStmt $ RaiseStmt $ CallExpr (AddrExpr a) es

instance Show Instr where
  show (FwBlock block) = show block
  show (FwStmt stmt) = show stmt
  show (FwExpr expr) = show expr
  show (FwApp f args) = show f ++ "(" ++ show args ++ ")"
  show (BwResult r) = show r
  show (Raise r) = "raise " ++ show r
  show (Return r) = "return" ++ show r
  show (Yield r a) = "yield" ++ show r ++ " " ++ show r
\end{code}
%endif


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsubsection{Side effects}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Some rules have an effect on a hidden state not part of the abstract machine. Specifically I/O operations, such as printing to the screen, are typical side effects. 

Because the outside world is not modelled in these semantics, side effects are informally described over the arrow of the rewrite rule. For example, a side effect |q| of a rule would be denoted as follows.

\[ 
\bigl\langle \Theta, \Gamma, S, \iota \bigr\rangle
\xRightarrow{q}
\bigl\langle \Theta', \Gamma', S', \iota' \bigr\rangle
\]

The side effect |q| occurs when the machine state is rewritten. Rules are never selected based on a side effect, i.e., the pattern match of a rewrite rule can not be influenced by the side effect of the rewrite rule.

The import statement (see Section \ref{section:import}) has no side effects, even though reading files usually has side effects. We assume however, that the imported files do not change during the execution of a program.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\subsubsection{Source of a rewrite rule} \label{section:rule-source}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The rewrite rules are written in Haskell and reformatted to yield formulas of the above form. For example, the source of Rule \ref{bw-rewrite-BlockFrame} is as follows:
\begingroup
\scriptsize
\[\vbox{%
\begin{verbatim}
rewrite           (State  heap  envs  ( stack :|: BlockFrame b )  ( BwResult  a )) = case b of
  EmptyBlock  ->  (state  heap  envs  ( stack                  )  ( BwResult  a ))
  otherwise   ->  (state  heap  envs  ( stack                  )  ( FwBlock   b ))
\end{verbatim}
}%
\]
\endgroup
